// Matteo Lulli, Massimo Bernaschi, Giorgio Parisi 2013
// Physics Dept. 'Sapienza', University of Rome
#ifndef CUDAEA3D_H
#define CUDAEA3D_H

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <cuda.h>

#define CURAND_MTGP32_HOST_H
#define CURAND_MTGP32DC_P_11213_H

#include <curand.h>
#include <curand_kernel.h>
/* include MTGP host helper functions */
#include <curand_mtgp32_host.h>
/* include MTGP pre-computed parameter sets */
#include <curand_mtgp32dc_p_11213.h>


#include "MSC.h"
#include "myRNG.h"

typedef struct{
  int L, steps, dumpN;
  int V, NJ, NS, replicas, sysN, kernelChoice;
  int cardN, blockSize, dumpF, analysisF, ram, recoverF, k;
  float beta;

  char devShm[200];
  char recoverDir[200];
  unsigned long long int *dumpTimes, startValue;

  MSC *h_reds0, *h_reds1, *h_reds2, *h_reds3;
  MSC *h_blues0, *h_blues1, *h_blues2, *h_blues3;
  MSC *d_reds0, *d_reds1, *d_reds2, *d_reds3;
  MSC *d_blues0, *d_blues1, *d_blues2, *d_blues3;

  MSC *h_Jpx, *h_Jpy, *h_Jpz;
  MSC *h_Jmx, *h_Jmy, *h_Jmz;
  MSC *d_Jpx, *d_Jpy, *d_Jpz;
  MSC *d_Jmx, *d_Jmy, *d_Jmz;

  int *realI_R, *realI_B;
  int *myI_R, *myI_B;
  int *myI_B2R, *myI_R2B;

}sysEA3D_t;

extern sysEA3D_t *sysInit(int argc, char **argv);
extern int dummyDirsMaker(sysEA3D_t *sys);
extern int SpinsRead(sysEA3D_t *sys, unsigned long long int t);
extern int SpinsReadDir(sysEA3D_t *sys, unsigned long long int t);
extern int SpinsReadSys(sysEA3D_t *sys, int t, int sysIndex);
extern int SpinsTransform(sysEA3D_t *sys);
extern int SpinsTransformRead(sysEA3D_t *sys);
extern int SpinsDump(sysEA3D_t *sys, unsigned long long int t);
extern int JRead(sysEA3D_t *sys);
extern int JReadDir(sysEA3D_t *sys);
extern int JReadSys(sysEA3D_t *sys, int sysIndex);
extern int JTransform(sysEA3D_t *sys);
extern int JTransformRead(sysEA3D_t *sys);
extern int JDump(sysEA3D_t *sys);
extern int Simulation(sysEA3D_t *sys);
extern int SimulationTest(sysEA3D_t *sys);
extern int SimulationFullRecoverTest(sysEA3D_t *sys);
extern int SimulationFullRecoverTestCongruential(sysEA3D_t *sys);
extern int testParisiRapuano(sysEA3D_t *sys);
extern int benchmark(int argc, char **argv);
extern int benchmarkSmooth(int argc, char **argv);
extern int benchmarkGrid(int argc, char **argv);
extern int benchmarkSliced_NVPROF(int argc, char **argv);
extern int benchmarkStandard_NVPROF(int argc, char **argv);
extern int benchmarkBitwise_NVPROF(int argc, char **argv);
extern int benchmarkSlicedGrid_NVPROF(int argc, char **argv);
extern int benchmarkStandardGrid_NVPROF(int argc, char **argv);
extern int benchmarkSliced_MT19937(int argc, char **argv);
extern int benchmarkSliced_PR(int argc, char **argv);
extern int benchmarkSliced_MTGP(int argc, char **argv);
extern int benchmarkSmooth_PR(int argc, char **argv);
extern int benchmarkSmooth_MT19937(int argc, char **argv);
extern int benchmarkSmooth_XORWOW(int argc, char **argv);
extern int benchmarkSmooth_MTGP32(int argc, char **argv);
extern int benchmarkGrid_PR(int argc, char **argv);

__global__ void MCStepReds_Bitwise(__restrict MSC *s0, __restrict MSC *s1,
				   __restrict MSC *s2, __restrict MSC *s3, 
				   __restrict RNGT *rand);

__global__ void MCStepBlues_Bitwise(__restrict MSC *s0, __restrict MSC *s1, 
				    __restrict MSC *s2, __restrict MSC *s3, 
				    __restrict RNGT *rand);

__global__ void MCStepReds_Sliced(__restrict MSC *s0, __restrict MSC *s1,
				  __restrict MSC *s2, __restrict MSC *s3, 
				  __restrict RNGT *rand);

__global__ void MCStepBlues_Sliced(__restrict MSC *s0, __restrict MSC *s1, 
				   __restrict MSC *s2, __restrict MSC *s3, 
				   __restrict RNGT *rand);

__global__ void MCStepReds_SlicedGrid(__restrict MSC *s0, __restrict MSC *s1, 
				      __restrict MSC *s2, __restrict MSC *s3, 
				      __restrict RNGT *rand);

__global__ void MCStepBlues_SlicedGrid(__restrict MSC *s0, __restrict MSC *s1, 
				       __restrict MSC *s2, __restrict MSC *s3, 
				       __restrict RNGT *rand);

__global__ void MCStepReds_Standard(__restrict MSC *s0, __restrict MSC *s1,
				    __restrict MSC *s2, __restrict MSC *s3, 
				    __restrict RNGT *rand);

__global__ void MCStepBlues_Standard(__restrict MSC *s0, __restrict MSC *s1, 
				     __restrict MSC *s2, __restrict MSC *s3, 
				     __restrict RNGT *rand);

__global__ void MCStepReds_StandardGrid(__restrict MSC *s0, __restrict MSC *s1, 
					__restrict MSC *s2, __restrict MSC *s3, 
					__restrict RNGT *rand);

__global__ void MCStepBlues_StandardGrid(__restrict MSC *s0, __restrict MSC *s1, 
					 __restrict MSC *s2, __restrict MSC *s3, 
					 __restrict RNGT *rand);


__global__ void testParisiRapuanoKernel(__restrict RNGT *ira, int steps,
					int ip, int ip1, int ip2, int ip3);

__global__ void testParisiRapuanoKernelMod(__restrict RNGT *ira, int steps,
					   int ip, int ip1, int ip2, int ip3);

__global__ void testMinStdKernel(__restrict RNGT *rng, int steps);

#endif
