#include <time.h>
#include "cutil.h"
#include "MersenneTwister.h"

#include "timing.h"
#include "dictionary.h"
#include "iniparser.h"
#include "util.h"

#include "msc.h"
#include "isinggpuGM.h"

#if !defined(HOSTREDU)
#include "threadFenceReduction_kernel.cu"
#endif

#define MASSIMO_RAND 0x7fffffff

#include <stdlib.h>
#define SIZEBYTE 8

char *printbit(unsigned int value) {
        int i;
        int maxbits=sizeof(unsigned int)*SIZEBYTE;
        char *p,*string;
        string=(char *)Malloc(maxbits+1);
        p=string+maxbits;
        p[0]=NULL;
        for(i=0; i<maxbits; i++) {
            *--p=value&1?'1':'0';
            value>>=1;
        }
        return string;
}

#define OUTRAND "randtest.dat"

//------------------------------------------------------------------------------------------------------
// Conguenziale: 2^31 - 1 -> Matteo: occorre controllare che non ci siano seed = 0!
// versione per variabili a 32 bits per evitare overflows; hi, low e seed devono essere unsigned int
//------------------------------------------------------------------------------------------------------
#define TWO31M1 0x7fffffff

#define RANDTWO31M1(seed)\
  lo = 16807*(seed&0xffff);			\
  hi = 16807*(seed>>16);			\
  lo += (hi&0x7fff)<<16;			\
  lo += hi>>15;					\
  seed = ((lo&0x80000000)>>31)*(lo-0x7fffffff);	\
  seed += ((lo^0x80000000)>>31)*lo;		\

#define RANDTWO31M1BIS(seed)\
  swap = (16807LL)*seed;    \
  seed = (swap&0x7fffffff) + (swap>>31);\
  if(seed&0x80000000) seed = (seed&0x7fffffff) + 1;\

#define RANDTWO32(seed)\
  seed = 1664525*seed + 1013904223;\

//------------------------------------------------------------------------------------------------------
// Nuova inizializzazione d_Rand: -> Matteo
//------------------------------------------------------------------------------------------------------

void initdRandNew(void){
  RANDOMTYPE *tmpRand, swap = 0;
  int totalThreads = numThreadsPerBlock*numBlocks;
  unsigned int i, counter = 0;
  FILE *urandom;

  urandom = fopen("/dev/urandom", "rb");
  if(urandom == NULL){
    fprintf(stderr, "\nUnable to open /dev/urandom\n");
    getchar();
    return;
  }
  tmpRand = (RANDOMTYPE *)malloc(totalThreads*sizeof(RANDOMTYPE));
  CUDA_SAFE_CALL( cudaMalloc((void **)&d_Rand, totalThreads * sizeof(RANDOMTYPE)) );
  /* inizializzo l'array temporaneo */
  for(i=0; counter<totalThreads; i++){
    //swap = (unsigned int)(lrand48()*(0xffffffff)); // meglio usare /dev/urandom
    fread(&swap, sizeof(RANDOMTYPE), 1, urandom);
    if(swap != 0 && swap%TWO31M1 != 0){ // il seme non deve essere coprimo di 0x7fffffff
      tmpRand[counter] = swap;
      counter++;
    }
  }
  /* copio l'array sul device */
  CUDA_SAFE_CALL( cudaMemcpy( d_Rand, tmpRand, totalThreads*sizeof(RANDOMTYPE), cudaMemcpyHostToDevice ) );
  fclose(urandom);

  /* dump su disco dei semi */
  FILE *dump = fopen("congruentialSeeds", "wb");
  fwrite(tmpRand, sizeof(RANDOMTYPE), totalThreads, dump);
  fclose(dump);

  return;
}

void initdRandNewRead(void){
  int totalThreads = numThreadsPerBlock*numBlocks;
  RANDOMTYPE  *tmpRand = (RANDOMTYPE *)malloc(totalThreads*sizeof(RANDOMTYPE));
  CUDA_SAFE_CALL( cudaMalloc((void **)&d_Rand, totalThreads * sizeof(RANDOMTYPE)) );
  FILE *dump = fopen("congruentialSeeds", "rb");
  fread(tmpRand, sizeof(RANDOMTYPE), totalThreads, dump);
  fclose(dump);
  CUDA_SAFE_CALL( cudaMemcpy( d_Rand, tmpRand, totalThreads*sizeof(RANDOMTYPE), cudaMemcpyHostToDevice ) );

  return;
}

//------------------------------------------------------------------------------------------------------
// newRandomMsc -> Matteo
//------------------------------------------------------------------------------------------------------

RANDOMTYPE newseed = 0;

void initSeeds(void){
  FILE *f;
  f = fopen("/dev/urandom", "rb");
  do{fread(&newseed, 4, 1, f);}while(newseed == 0);
  fclose(f);
  return;
}

MSC newRandomMsc(void){
  MSC ris = 0, hi, lo;
  for(int i=0; i<BITS; i++){
    RANDTWO31M1(newseed);
    ris <<= 1;
    MSC newBit = ((((float)(newseed))/TWO31M1) > .5) ? 0 : 1;
    ris ^= newBit;
  }
  return ris;
}

//-------------------------------------------------------------------------------------------------------
// Usage
//-------------------------------------------------------------------------------------------------------
void Usage(char *program) {
  fprintf(stderr,"Usage: %s -i inputfile [-r spin_file] [-o (over-relax only)] [-b (heat bath only)]\n",program);
}


//-------------------------------------------------------------------------------------------------------
// Init
//-------------------------------------------------------------------------------------------------------
void Init(void){

  if(sizeof(REAL)==sizeof(float)) {
    zero=0.0f;
    one=1.0f;
    two=2.0f;
    three=3.0f;
    zerofive=0.5f;
    zerotwentyfive=0.25f;
  } else if(sizeof(REAL)==sizeof(double)) {
    zero=0.0;
    one=1.0;
    two=2.0;
    three=3.0;
    zerofive=0.5;
    zerotwentyfive=0.25;
  } else {
    writelog(TRUE,APPLICATION_RC,"Invalid size of REAL: %d",sizeof(REAL));
  }


   h_Ra_S = (MSC*) Malloc(ms_s);
  h_Ba_S = (MSC*) Malloc(ms_s);
  h_Rb_S = (MSC*) Malloc(ms_s);
  h_Bb_S = (MSC*) Malloc(ms_s);
  h_Rc = (MSC*) Malloc(ms_s);
  h_Bc = (MSC*) Malloc(ms_s);
  h_Rd = (MSC*) Malloc(ms_s);
  h_Bd = (MSC*) Malloc(ms_s);
  h_Rc_S = (MSC*) Malloc(ms_s);
  h_Bc_S = (MSC*) Malloc(ms_s);
  h_Rd_S = (MSC*) Malloc(ms_s);
  h_Bd_S = (MSC*) Malloc(ms_s);
  h_Ra = (MSC*) Malloc(ms_s);
  h_Ba = (MSC*) Malloc(ms_s);
  h_Rb = (MSC*) Malloc(ms_s);
  h_Bb = (MSC*) Malloc(ms_s);


  h_Jpx = (MSC*) Malloc(ms_j);
  h_Jpy = (MSC*) Malloc(ms_j);
  h_Jpz = (MSC*) Malloc(ms_j);
  h_Jmx = (MSC*) Malloc(ms_j);
  h_Jmy = (MSC*) Malloc(ms_j);
  h_Jmz = (MSC*) Malloc(ms_j);

  h_Energy = (int*) Malloc(ms_s);

  h_sdbL = findlog2(h_side);
  h_sdbLs = findlog2(h_side*h_side);

  h_sdbLm1 = h_sdbL-1;  //log_2(L) -1 used to divide by L/2
  h_sdbLsm1 = h_sdbLs-1;  //log_2(L*L) -1 used to divide by (L*L)/2
  h_maskL = (h_side>>1)-1;
  h_maskL2 = ((h_side*h_side)>>1)-1;
  h_maskL3 = (V(h_side)>>1)-1;
  h_Ldb2 = h_side>>1;
  h_Lsdb2 = (h_side*h_side)>>1;

  h_nspin=V(h_side)/2;

  srand48(seed);

}


//-------------------------------------------------------------------------------------------------------
// CopyConstOnDevice
//-------------------------------------------------------------------------------------------------------
void CopyConstOnDevice(void){

  int tmp[2];

  //CUDA_SAFE_CALL( cudaMemcpyToSymbol("d_side",&h_side,sizeof(int),0,cudaMemcpyHostToDevice) );
  //CUDA_SAFE_CALL( cudaMemcpyToSymbol("d_nspin",&h_nspin,sizeof(int),0,cudaMemcpyHostToDevice) );

  //CUDA_SAFE_CALL( cudaMemcpyToSymbol("d_sdbL",&h_sdbL,sizeof(int),0,cudaMemcpyHostToDevice) );
  //CUDA_SAFE_CALL( cudaMemcpyToSymbol("d_sdbLs",&h_sdbLs,sizeof(int),0,cudaMemcpyHostToDevice) );
  //CUDA_SAFE_CALL( cudaMemcpyToSymbol("d_sdbLm1",&h_sdbLm1,sizeof(int),0,cudaMemcpyHostToDevice) );
  //CUDA_SAFE_CALL( cudaMemcpyToSymbol("d_sdbLsm1",&h_sdbLsm1,sizeof(int),0,cudaMemcpyHostToDevice) );
  //CUDA_SAFE_CALL( cudaMemcpyToSymbol("d_maskL",&h_maskL,sizeof(int),0,cudaMemcpyHostToDevice) );
  //CUDA_SAFE_CALL( cudaMemcpyToSymbol("d_maskL2",&h_maskL2,sizeof(int),0,cudaMemcpyHostToDevice) );
  //CUDA_SAFE_CALL( cudaMemcpyToSymbol("d_maskL3",&h_maskL3,sizeof(int),0,cudaMemcpyHostToDevice) );
  //CUDA_SAFE_CALL( cudaMemcpyToSymbol("d_Ldb2",&h_Ldb2,sizeof(int),0,cudaMemcpyHostToDevice) );
  //CUDA_SAFE_CALL( cudaMemcpyToSymbol("d_Lsdb2",&h_Lsdb2,sizeof(int),0,cudaMemcpyHostToDevice) );
  //CUDA_SAFE_CALL( cudaMemcpyToSymbol("d_h",&h_h,sizeof(REAL),0,cudaMemcpyHostToDevice) );
  //CUDA_SAFE_CALL( cudaMemcpyToSymbol("d_Beta",&h_Beta,sizeof(REAL),0,cudaMemcpyHostToDevice) );

  // CUDA v5.0 

  CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_side, &h_side, sizeof(int)) );
  CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_nspin, &h_nspin, sizeof(int)) );

  CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_sdbL, &h_sdbL, sizeof(int)) );
  CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_sdbLs, &h_sdbLs, sizeof(int)) );
  CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_sdbLm1, &h_sdbLm1, sizeof(int)) );
  CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_sdbLsm1, &h_sdbLsm1, sizeof(int)) );
  CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_maskL, &h_maskL, sizeof(int)) );
  CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_maskL2, &h_maskL2, sizeof(int)) );
  CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_maskL3, &h_maskL3, sizeof(int)) );
  CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Ldb2, &h_Ldb2, sizeof(int)) );
  CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Lsdb2, &h_Lsdb2, sizeof(int)) );
  CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_h, &h_h,sizeof(REAL)) );
  CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Beta, &h_Beta, sizeof(REAL)) );


  tmp[0]=0;
  tmp[1]=-1;

  CUDA_SAFE_CALL( cudaMemcpyToSymbol(jindred, tmp, sizeof(tmp)) );

  tmp[0]=-1;
  tmp[1]=0;

  CUDA_SAFE_CALL( cudaMemcpyToSymbol(jindblue, tmp, sizeof(tmp)) );

  CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_const_one, &one, sizeof(REAL)) );
  CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_const_two, &two, sizeof(REAL)) );
  CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_const_zerotwentyfive, &zerotwentyfive, sizeof(REAL)) );

  CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_flip_probability4, &flip_probability4, sizeof(unsigned int)) );

  CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_flip_probability2, &flip_probability2, sizeof(unsigned int)) );

  CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_flip_probability0, &flip_probability0, sizeof(unsigned int)) );

}


//-------------------------------------------------------------------------------------------------------
// findlog2
//-------------------------------------------------------------------------------------------------------
int findlog2(int n){

  int i=0;

  if(n<1 || n&1) { writelog(TRUE,FINDLOG_RC,"Invalid value %d. Must be > 0 and even\n",n); }

  while(n>0) {
    n>>=1;
    if(n>0) i++;
  }

  return i;
}


//-------------------------------------------------------------------------------------------------------
// nextPow2
//-------------------------------------------------------------------------------------------------------
unsigned int nextPow2( unsigned int x ) {
  --x;
  x |= x >> 1;
  x |= x >> 2;
  x |= x >> 4;
  x |= x >> 8;
  x |= x >> 16;
  return ++x;
}


//-------------------------------------------------------------------------------------------------------
// FreeDevice
//-------------------------------------------------------------------------------------------------------
void FreeDevice(void){

  cudaFree(d_Ra);
  cudaFree(d_Ba);
  cudaFree(d_Rb);
  cudaFree(d_Bb);
  cudaFree(d_Rc);
  cudaFree(d_Bc);
  cudaFree(d_Rd);
  cudaFree(d_Bd);


  cudaFree(d_Jpx);
  cudaFree(d_Jpy);
  cudaFree(d_Jpz);
  cudaFree(d_Jmx);
  cudaFree(d_Jmy);
  cudaFree(d_Jmz);

  cudaFree(d_Energy);

}


//-------------------------------------------------------------------------------------------------------
// FreeHost
//-------------------------------------------------------------------------------------------------------
void FreeHost(void){

  Free(h_Ra);
  Free(h_Ba);
  Free(h_Rb);
  Free(h_Bb);
  Free(h_Rc);
  Free(h_Bc);
  Free(h_Rd);
  Free(h_Bd);

  Free(h_Ra_S);
  Free(h_Ba_S);
  Free(h_Rb_S);
  Free(h_Bb_S);
  Free(h_Rc_S);
  Free(h_Bc_S);
  Free(h_Rd_S);
  Free(h_Bd_S);

  Free(h_Jpx);
  Free(h_Jpy);
  Free(h_Jpz);
  Free(h_Jmx);
  Free(h_Jmy);
  Free(h_Jmz);

  Free(h_Energy);

}

void PrintMagnetization(int istep){ 
  int magtot=0;

  for(int i = 0; i<ns_j; i++){
    magtot+=((h_Ra[i]^h_Ra_S[i])&0x2?1:-1);
    magtot+=((h_Ba[i]^h_Ba_S[i])&0x2?1:-1);
  }
  printf("Magnetization of system time %d 2=%f\n",istep, fabs(magtot)/(2.0*(double)ns_j));

}

//-------------------------------------------------------------------------------------------------------
// Dump_Spin
//-------------------------------------------------------------------------------------------------------
void Dump_Spin(int count){ // DA RIVEDERE

  FILE* tempfile;
  char namefile[MAXSTRLEN];
  
  memset(namefile,'\0',MAXSTRLEN); // NON CONTROLLO SE TORNA BENE O MENO

  //if(count>=0){

    snprintf(namefile,sizeof(namefile),"spin_dump_L%dT%dtime%d",h_side,int(h_Beta*100+.001),count);
    if(count>=0){

    CUDA_SAFE_CALL( cudaMemcpy( h_Ra, d_Ra, ms_s, cudaMemcpyDeviceToHost ) );
    CUDA_SAFE_CALL( cudaMemcpy( h_Ba, d_Ba, ms_s, cudaMemcpyDeviceToHost ) );

    CUDA_SAFE_CALL( cudaMemcpy( h_Bb, d_Bb, ms_s, cudaMemcpyDeviceToHost ) );
    CUDA_SAFE_CALL( cudaMemcpy( h_Rb, d_Rb, ms_s, cudaMemcpyDeviceToHost ) );
    CUDA_SAFE_CALL( cudaMemcpy( h_Rc, d_Rc, ms_s, cudaMemcpyDeviceToHost ) );
    CUDA_SAFE_CALL( cudaMemcpy( h_Bc, d_Bc, ms_s, cudaMemcpyDeviceToHost ) );

    CUDA_SAFE_CALL( cudaMemcpy( h_Bd, d_Bd, ms_s, cudaMemcpyDeviceToHost ) );
    CUDA_SAFE_CALL( cudaMemcpy( h_Rd, d_Rd, ms_s, cudaMemcpyDeviceToHost ) );



  }; //else snprintf(namefile,sizeof(namefile),"spin_dump_start");

  tempfile = Fopen(namefile,"wb");
/*
  for(int i = 0; i<ns_j; i++){
  char*  s1=printbit(h_Ra[i]);
  char*  s2=printbit(h_Ba[i]);
    fprintf(tempfile,"%s  \n%s  \n", s1, s2);
    Free(s1); Free(s2);
  }
*/
       fwrite(h_Ra,1,ms_s,tempfile);
       fwrite(h_Ba,1,ms_s,tempfile);
       fwrite(h_Rb,1,ms_s,tempfile);
       fwrite(h_Bb,1,ms_s,tempfile);
       fwrite(h_Rc,1,ms_s,tempfile);
       fwrite(h_Bc,1,ms_s,tempfile);
       fwrite(h_Rd,1,ms_s,tempfile);
       fwrite(h_Bd,1,ms_s,tempfile);


  fflush(tempfile);

  fclose(tempfile);

}
//-------------------------------------------------------------------------------------------------------
// Dump_J
//-------------------------------------------------------------------------------------------------------
void Dump_J(void){

  FILE* tempfile;

  tempfile = Fopen("j_dump_start","wb");

/*    for(int i = 0; i<ns_j; i++) {
      fprintf(tempfile,"%d %d %d\n",h_Jpx[i],h_Jpy[i],h_Jpz[i]);
      fprintf(tempfile,"%d %d %d\n",h_Jmx[i],h_Jmy[i],h_Jmz[i]);
        }
*/
       fwrite(h_Jpx,1,ns_j*sizeof(int),tempfile);
       fwrite(h_Jpy,1,ns_j*sizeof(int),tempfile);
       fwrite(h_Jpz,1,ns_j*sizeof(int),tempfile);
       fwrite(h_Jmx,1,ns_j*sizeof(int),tempfile);
       fwrite(h_Jmy,1,ns_j*sizeof(int),tempfile);
       fwrite(h_Jmz,1,ns_j*sizeof(int),tempfile);


  fflush(tempfile);

  fclose(tempfile);

}

//-------------------------------------------------------------------------------------------------------
// Read_J: -> Matteo
//-------------------------------------------------------------------------------------------------------

void Read_J(void){
  FILE *jdump = fopen("j_dump_start", "rb");
  int chkcp = fread(h_Jpx, 4, ns_j, jdump);
  chkcp += fread(h_Jpy, 4, ns_j, jdump);
  chkcp += fread(h_Jpz, 4, ns_j, jdump);
  chkcp += fread(h_Jmx, 4, ns_j, jdump);
  chkcp += fread(h_Jmy, 4, ns_j, jdump);
  chkcp += fread(h_Jmz, 4, ns_j, jdump);
  fclose(jdump);
  if(chkcp^(ns_j*6)) fprintf(stderr, "\nError Copying couplings\n");

  return;
}

//-------------------------------------------------------------------------------------------------------
// Initialize_Spin
//-------------------------------------------------------------------------------------------------------
void Initialize_Spin(MSC* sa_a,MSC* sa_a_S){


  for(int i = 0; i<h_nspin; i++){

    sa_a_S[i]=sa_a[i]=newRandomMsc();
    /* sa_a_S[i]=sa_a[i]=1;
       sa_a_S[i] = sa_a_S[i]^(~sa_a_S[i]);
       sa_a[i] = sa_a[i]^(~sa_a[i]); -> Matteo: all aligned*/

  }

}


//-------------------------------------------------------------------------------------------------------
// Read_Spin
//-------------------------------------------------------------------------------------------------------
/*void Read_Spin(char *inputfile,
               MSC* sr_a, MSC* sb_a) { // DA RIVEDERE

  MSC a;
  int j=0, k=0;
  FILE *fpi;

  fpi=Fopen(inputfile,"r");

  fprintf(stderr,"Reading spin from file %s\n",inputfile);

  for(int i = 0; i< V(h_side); i++){

         fscanf(fpi,"%d",&a);


    if(i&1) {
      sb_a[j]=a;
      j++;
    } else {
      sr_a[k]=a;
      k++;
    }

  }

  fclose(fpi);


  }*/

void Read_Spin(char *spindump){
  FILE *f = fopen(spindump, "rb");
  int chkcp = fread(h_Ra, 4, ns_j, f);
  chkcp += fread(h_Ba, 4, ns_j, f);

  chkcp += fread(h_Rb, 4, ns_j, f);
  chkcp += fread(h_Bb, 4, ns_j, f);

  chkcp += fread(h_Rc, 4, ns_j, f);
  chkcp += fread(h_Bc, 4, ns_j, f);

  chkcp += fread(h_Rd, 4, ns_j, f);
  chkcp += fread(h_Bd, 4, ns_j, f);

  fclose(f);
  return;
}


//-------------------------------------------------------------------------------------------------------
// CopyDataOnDevice
//-------------------------------------------------------------------------------------------------------
void CopyDataOnDevice(void){

  CUDA_SAFE_CALL( cudaMalloc( (void **) &d_Ra, ms_s ) );
  CUDA_SAFE_CALL( cudaMalloc( (void **) &d_Ba, ms_s ) );
  CUDA_SAFE_CALL( cudaMalloc( (void **) &d_Rb, ms_s ) );
  CUDA_SAFE_CALL( cudaMalloc( (void **) &d_Bb, ms_s ) );
  CUDA_SAFE_CALL( cudaMalloc( (void **) &d_Rc, ms_s ) );
  CUDA_SAFE_CALL( cudaMalloc( (void **) &d_Bc, ms_s ) );
  CUDA_SAFE_CALL( cudaMalloc( (void **) &d_Rd, ms_s ) );
  CUDA_SAFE_CALL( cudaMalloc( (void **) &d_Bd, ms_s ) );

  CUDA_SAFE_CALL( cudaMemcpy( d_Ra, h_Ra, ms_s, cudaMemcpyHostToDevice ) );
  CUDA_SAFE_CALL( cudaMemcpy( d_Ba, h_Ba, ms_s, cudaMemcpyHostToDevice ) );
  CUDA_SAFE_CALL( cudaMemcpy( d_Rb, h_Rb, ms_s, cudaMemcpyHostToDevice ) );
  CUDA_SAFE_CALL( cudaMemcpy( d_Bb, h_Bb, ms_s, cudaMemcpyHostToDevice ) );

  CUDA_SAFE_CALL( cudaMemcpy( d_Rc, h_Rc, ms_s, cudaMemcpyHostToDevice ) );
  CUDA_SAFE_CALL( cudaMemcpy( d_Bc, h_Bc, ms_s, cudaMemcpyHostToDevice ) );
  CUDA_SAFE_CALL( cudaMemcpy( d_Rd, h_Rd, ms_s, cudaMemcpyHostToDevice ) );
  CUDA_SAFE_CALL( cudaMemcpy( d_Bd, h_Bd, ms_s, cudaMemcpyHostToDevice ) );


  CUDA_SAFE_CALL( cudaMalloc( (void **) &d_Jpx, ms_j ) );
  CUDA_SAFE_CALL( cudaMalloc( (void **) &d_Jpy, ms_j ) );
  CUDA_SAFE_CALL( cudaMalloc( (void **) &d_Jpz, ms_j ) );

  CUDA_SAFE_CALL( cudaMalloc( (void **) &d_Jmx, ms_j ) );
  CUDA_SAFE_CALL( cudaMalloc( (void **) &d_Jmy, ms_j ) );
  CUDA_SAFE_CALL( cudaMalloc( (void **) &d_Jmz, ms_j ) );

  CUDA_SAFE_CALL( cudaMemcpy( d_Jpx, h_Jpx, ms_j, cudaMemcpyHostToDevice ) );
  CUDA_SAFE_CALL( cudaMemcpy( d_Jpy, h_Jpy, ms_j, cudaMemcpyHostToDevice ) );
  CUDA_SAFE_CALL( cudaMemcpy( d_Jpz, h_Jpz, ms_j, cudaMemcpyHostToDevice ) );

  CUDA_SAFE_CALL( cudaMemcpy( d_Jmx, h_Jmx, ms_j, cudaMemcpyHostToDevice ) );
  CUDA_SAFE_CALL( cudaMemcpy( d_Jmy, h_Jmy, ms_j, cudaMemcpyHostToDevice ) );
  CUDA_SAFE_CALL( cudaMemcpy( d_Jmz, h_Jmz, ms_j, cudaMemcpyHostToDevice ) );

  CUDA_SAFE_CALL( cudaMalloc( (void **) &d_Energy, ms_s ) );
  CUDA_SAFE_CALL( cudaMemset( d_Energy, 0, ms_s ) );
}


//-------------------------------------------------------------------------------------------------------
// Initialize_J
//-------------------------------------------------------------------------------------------------------
void Initialize_J(void){

  int j;


        for (j=0;j<ns_j;j++) {
	     switch(init_type_j) {
		case 1:
		  h_Jpx[j] = newRandomMsc(); // per ripristinare basta rimuovere new: -> Matteo
     		 h_Jpy[j] = newRandomMsc();
     		 h_Jpz[j] = newRandomMsc();
     		 h_Jmx[j] = newRandomMsc();
      		 h_Jmy[j] = newRandomMsc();
      		 h_Jmz[j] = newRandomMsc();
		break;
		case 0:
                 h_Jpx[j]=h_Jpy[j]=h_Jpz[j]=h_Jmx[j]=h_Jmy[j]=h_Jmz[j]=0;
		break;
		case -1:
                 h_Jpx[j]=h_Jpy[j]=h_Jpz[j]=h_Jmx[j]=h_Jmy[j]=h_Jmz[j]=(unsigned)-1;
		break;
		default:
		 writelog(TRUE,APPLICATION_RC,"Invalid initialization type for J: %d",init_type_j);
		break;
	}
}

}




//-------------------------------------------------------------------------------------------------------
// Parameters_Reading
//-------------------------------------------------------------------------------------------------------
void Parameters_Reading(char *inputfile){

  dictionary *ini;
  char key[MAXSTRLEN];

  if(inputfile==NULL) { writelog(TRUE,APPLICATION_RC,"No inputfile"); }

  ini = iniparser_load(inputfile);

  if(ini==NULL) { writelog(TRUE,APPLICATION_RC,"Cannot parse file: %s\n", inputfile); }

  READINTFI(h_side,"Side");
  READINTFI(init_type_j,"Initialization type for J");
  READINTFI(nsteps,"Number of steps");
  READINTFI(overrelaxiter,"Number of overrelaxation iterations");
  READINTFI(outfreq,"Output frequency");
  READINTFI(dumpfreq,"Dump configuration frequency");
  READREALFI(h_h,"Magnetic field");
  READREALFI(h_Beta,"Beta (1/T)");
  READINTFI(seed,"Random seed");

  DEVICEREADINTFI(numBlocks,"Number of blocks");
  DEVICEREADINTFI(numThreadsPerBlock,"Number of threads for block");

  DEVICEREADINTFI(RandNumBlocks,"RandGene num of blocks");
  DEVICEREADINTFI(RandNumThreadsPerBlock,"RandGene num of threads for block");

}


//-------------------------------------------------------------------------------------------------------
// Calculate_Sizes
//-------------------------------------------------------------------------------------------------------
void Calculate_Sizes(void){

  ns_j = (V(h_side)/2);

  ms_s = ns_j*sizeof(REAL);

  ms_j = ns_j*sizeof(REAL);

}


//-------------------------------------------------------------------------------------------------------
// getNumBlocksAndThreads
//-------------------------------------------------------------------------------------------------------
void getNumBlocksAndThreads(int n, int maxBlocks, int maxThreads, int &blocks, int &threads) {

  if (n == 1){
    threads = 1;
    blocks = 1;
  } else{
    threads = (n < maxThreads*2) ? nextPow2(n / 2) : maxThreads;
    blocks = max(1, n / (threads * 2));
  }

  blocks = min(maxBlocks, blocks);
}



REAL Media(int vec[BITS]){
        REAL ris=0;
        for (int i=0;i<BITS;i++) {
          ris+=vec[i];
        }
        return ris/BITS;
}
//-------------------------------------------------------------------------------------------------------
// PrintEnergy
//-------------------------------------------------------------------------------------------------------

void PrintEnergy(int tempo, MSC * h_R,MSC * h_B){
  LongMsc newEnergy;
  AllocateLongMsc(&newEnergy,findlog2(12*h_nspin));
//  PrintLongMsc(& newEnergy);
  NewComputeHEnergy(h_R,  h_B ,h_Jpx, h_Jpy, h_Jpz, h_Jmx, h_Jmy, h_Jmz, & newEnergy,TRUE);
  //PrintClearLongMsc(& newEnergy);
  int EnergyVec[BITS];
  ConvertiArrayIntLongMsc(EnergyVec,&newEnergy);
  for (int i=1;i<BITS ;i++) {
    printf("Energy at time %d per spin on host:%d %f\n",tempo, i,((REAL)(EnergyVec[i]-3*h_nspin))/(h_nspin));
  }
  FreeLongMsc(&newEnergy);
}

//-------------------------------------------------------------------------------------------------------
unsigned long long lista_tiempos[200];
void genera_tiempos(){
  int nt,j;
  unsigned long long t;

  nt=0;
  lista_tiempos[0]=0;
  for( j=0;j<=4*nsteps;j++){
    t=rint(pow(2.,j/4.));
    if(t>lista_tiempos[nt]){
      nt++;
      lista_tiempos[nt]=t;
    }
  }
}
// 
//-------------------------------------------------------------------------------------------------------

//------------------------------------------------------------------------------------------------------
// Chi -> Matteo
//------------------------------------------------------------------------------------------------------

#define WINDOW 4
#define MEASURE 2048
#define BLOCKS 1024
#define THERM  10000
#define OUT1 "partials.dat"
#define OUT2 "blocks.dat"



double chi[32];
FILE *un, *deux;

void initChi(void){
  int i;
  for(i=0; i<32; i++) chi[i] = 0.;
  return;
}

MSC pedantic(MSC *colour, int level){
  MSC sum = 0;
  int i;
  for(i=0; i<ns_j; i++) sum += ((colour[i]>>level)&1U);
  return sum;
}

void Chi(void){
  int sum = 0, i;
  MSC overlapR[ns_j], overlapB[ns_j];

  for(i=0; i<ns_j; i++){
    overlapR[i] = ~(h_Ra[i]^h_Rb[i]);
    overlapB[i] = ~(h_Ba[i]^h_Bb[i]);
  }

  for(i=0; i<32; i++){
    sum = 0;
    sum += 2*pedantic(overlapR, i) - ns_j;
    sum += 2*pedantic(overlapB, i) - ns_j;
    fprintf(un, "%d %.10f\n", i, ((double)sum)*((double)sum)/(2.*ns_j));
    chi[i] += ((double)sum)*((double)sum)/(2.*ns_j); 
  }

  return;
}

void printChi(int current){
  int i;
  for(i=0; i<32; i++) fprintf(deux, "%d %d %.10f\n", current, i, chi[i]/MEASURE);
  return;
}

//------------------------------------------------------------------------------------------------------
// Parisi-Rapuano -> Matteo
//------------------------------------------------------------------------------------------------------

#define RAND1(ira)  ((ira[threadIndex*260 + ip++]=ira[threadIndex*260 + ip1++]+ira[threadIndex*260 + ip2++]) ^ira[threadIndex*260 + ip3++] )

RANDOMTYPE *h_ira, *d_ira;

void initIra(void){
  int totalThreads = numThreadsPerBlock*numBlocks;
  unsigned char ip, ip1, ip2, ip3;
  int i, j, threadIndex;
  FILE *urandom;

  urandom = fopen("/dev/urandom", "rb");
  if(urandom == NULL){
    fprintf(stderr, "\nUnable to open /dev/urandom\n");
    getchar();
    return;
  }

  /* 260: 256 sono per implementare le condizioni periodiche tramite la postincrementazione
     degli unsigned char; gli altri quattro servono in maniera tale che ogni thread mantenga
     in memoria il valore degli indici
   */

  h_ira = (RANDOMTYPE *)malloc(260*totalThreads*sizeof(RANDOMTYPE));
  CUDA_SAFE_CALL( cudaMalloc( (void **) &d_ira, 260*totalThreads*sizeof(RANDOMTYPE)) );

  ip=61;         // -> 256
  ip1=ip-24;     // -> 257
  ip2=ip-55;     // -> 258
  ip3=ip-61;     // -> 259

  /* copio il valore iniziale degli indici alla fine di ogni blocco da 260 */

  for(i=0; i<totalThreads; i++){
    h_ira[i*260 + 256] = ip;
    h_ira[i*260 + 257] = ip1; 
    h_ira[i*260 + 258] = ip2; 
    h_ira[i*260 + 259] = ip3; 
  }

  /* inizializzo i primi 61 valori */

  for(i=0; i<totalThreads; i++){
    /*for(j=ip3; j<ip; j++){
      h_ira[i*260 + j] = (RANDOMTYPE)(lrand48()*(0xffffffff));
      }*/
    // Meglio usare urandom; a meno di non voler riprodurre il risultato usando lo stesso seme
    fread((h_ira + i*260), sizeof(RANDOMTYPE), ip, urandom); 
  }

  /* finisco di inizializzare ogni blocco e copio gli indici */

  for(threadIndex=0; threadIndex<totalThreads; threadIndex++){
    ip = h_ira[threadIndex*260 + 256];
    ip1 = h_ira[threadIndex*260 + 257];
    ip2 = h_ira[threadIndex*260 + 258];
    ip3 = h_ira[threadIndex*260 + 259];
    //printf("Prima: ip: %d, ip1: %d, ip2: %d, ip3: %d\n", ip, ip1, ip2, ip3);
    for(j=0; j<1111; j++) RAND1(h_ira);
    h_ira[threadIndex*260 + 256] = ip;
    h_ira[threadIndex*260 + 257] = ip1; 
    h_ira[threadIndex*260 + 258] = ip2; 
    h_ira[threadIndex*260 + 259] = ip3; 

    //printf("Dopo: ip: %d, ip1: %d, ip2: %d, ip3: %d\n", ip, ip1, ip2, ip3);
  }

  /*for(i=0; i<totalThreads; i++){
    for(j=0; j<260; j++){
      printf("Block: (%d,%d), h_ira[%d]: %d\n", i, j, i*260 + j, h_ira[i*260 + j]);
    }
    }*/

  /* copio l'array sul device */
  CUDA_SAFE_CALL( cudaMemcpy( d_ira, h_ira, 260*totalThreads*sizeof(RANDOMTYPE), cudaMemcpyHostToDevice ) );
  fclose(urandom);

  return;
}

/* controlla che il risultato dei passi eseguiti su GPU sia uguale a quello eseguito su Host */

void ParisiRapuanoTester(int n){
  int totalThreads = numThreadsPerBlock*numBlocks;
  RANDOMTYPE *swap;
  int i, threadIndex, counter;
  unsigned char ip, ip1, ip2, ip3;

  swap = (RANDOMTYPE *)malloc(totalThreads*260*sizeof(RANDOMTYPE));
  CUDA_SAFE_CALL( cudaMemcpy(swap, d_ira, totalThreads*260*sizeof(RANDOMTYPE), cudaMemcpyDeviceToHost) );

  for(threadIndex=0; threadIndex<totalThreads; threadIndex++){
    counter = 0;
    ip = h_ira[threadIndex*260 + 256];
    ip1 = h_ira[threadIndex*260 + 257];
    ip2 = h_ira[threadIndex*260 + 258];
    ip3 = h_ira[threadIndex*260 + 259];
    for(i=0; i<n; i++) RAND1(h_ira);
    h_ira[threadIndex*260 + 256] = ip;
    h_ira[threadIndex*260 + 257] = ip1; 
    h_ira[threadIndex*260 + 258] = ip2; 
    h_ira[threadIndex*260 + 259] = ip3;
    for(i=0; i<260; i++){
      if(h_ira[threadIndex*260 + i] != swap[threadIndex*260 + i]){
	fprintf(stderr, "%d %u %u\n", threadIndex*260 + i, h_ira[threadIndex*260 + i], swap[threadIndex*260 + i]);
	counter++;
      }
    }
    if(counter == 0){
      printf("YES~, threadIndex: %d\n", threadIndex);
      getchar();
    } 
    if(counter > 0){
      printf("NO~, threadIndex: %d\n", threadIndex);
      getchar();
    }    
  }

  free(swap);
  return;
}

//------------------------------------------------------------------------------------------------------
//------------------------------------------------------------------------------------------------------
//------------------------------------------------------------------------------------------------------

int main(int argc, char **argv){
  //  TestMacro();
  char *inputfile = NULL;
  char *po;
  int istep;
  int nblocks_red, nthreads_red;
  double ttimeover=0.;
  //REAL normene;
  //int j;
  char *readspin=NULL;
#if !defined(HOSTREDU)
  unsigned int retCnt;
#endif

  cudaSetDevice(0);

  TIMER_DEF;

  for(int i = 1; i < argc; i++) {
    po = argv[i];
    if (*po++ == '-') {
      switch (*po++) {
      case '?':
        SKIPBLANK
          Usage(argv[0]);
        exit(OK);
        break;
      case 'v':
        SKIPBLANK
          verbose=TRUE;
        break;
      case 'i':
        SKIPBLANK
          inputfile=Strdup(po);
        break;
      case 'o':
        SKIPBLANK
          onlyOverrelaxation=true;
        break;
      case 'r':
        SKIPBLANK
          readspin=Strdup(po);
        break;
      case 'b':
        SKIPBLANK
          onlyHeatBath=true;
        break;

      default:
        Usage(argv[0]);
        exit(OK);
        break;
      }
    }
  }


  Parameters_Reading(inputfile);
  genera_tiempos();
  // printf("Beta: %f\n",h_Beta);
 
  /*  flip_probability4=((REAL)MASSIMO_RAND)*exp(-4.*h_Beta);
  flip_probability2=MASSIMO_RAND*exp(-8.*h_Beta);
  flip_probability0=MASSIMO_RAND*exp(-12.*h_Beta); -> Matteo */

  /* i valori delle flip probability non sembrano essere precisi -> Matteo */

  flip_probability4=(unsigned int)((double)MASSIMO_RAND*exp(-4.*h_Beta)); 
  flip_probability2=(unsigned int)((double)MASSIMO_RAND*exp(-8.*h_Beta));
  flip_probability0=(unsigned int)((double)MASSIMO_RAND*exp(-12.*h_Beta));
  
  // printf("flipsprobabilities %u %u %u\n %.10f %.10f\n",flip_probability4,flip_probability2,flip_probability0, 
  //	 (double)MASSIMO_RAND*exp(-4.*h_Beta), exp(-4.*h_Beta)); // -> Matteo

  if (h_side<=0 || !pow_of_two(h_side)) {
    writelog(TRUE,APPLICATION_RC,"Invalid side: must be > 0 and power of two\n");
  }

  //  return 0;
  initSeeds();

  Calculate_Sizes();

  Init();

  dim3 dimGrid(numBlocks,1,1);
  dim3 dimBlock(numThreadsPerBlock,1,1);
  int firstIndex = 0;

  if(readspin==NULL) {
    //  printf("Initialize_Spin\n");
    //    Initialize_Spin(h_Ra,h_Ra_S);
    //    Initialize_Spin(h_Ba,h_Ba_S);
    //    Initialize_Spin(h_Rb,h_Rb_S);
    //    Initialize_Spin(h_Bb,h_Bb_S);
    // printf("Initialize_Spin4\n");
    //    Initialize_Spin(h_Rc,h_Rc_S);
    // printf("Initialize_Spin5\n");
    //    Initialize_Spin(h_Bc,h_Bc_S);
    // printf("Initialize_Spin6\n");
    //    Initialize_Spin(h_Rd,h_Rd_S);
    // printf("Initialize_Spin7\n");
    //    Initialize_Spin(h_Bd,h_Bd_S);
    // printf("Initialized_Spin8\n");
    //    Dump_Spin(-1);
    initdRandNew();
    //    Initialize_J();
    //    Dump_J();   
  } else {
    //    Read_Spin(readspin,h_Ra,h_Ba);
    Read_Spin(readspin);
    initdRandNewRead();
    Read_J();
    int dummyBeta = 0, dummyL = 0, dummyRun = 0;
    sscanf(readspin, "spin_dump_L%dT%2dtime%d", &dummyL, &dummyBeta, &firstIndex);
    printf("dummyRun: %d, dummyL: %d, dummyBeta: %d, firstIndex: %d\n", 
	   dummyRun, dummyL, dummyBeta, firstIndex);
    if(dummyL != h_side || dummyBeta != (int)(h_Beta*100 + 0.001)){
      fprintf(stderr, "Resume data do not match input file data!\n");
      return 0;
    }
    firstIndex += 1;
  }

  //initRand(numThreadsPerBlock*numBlocks);

  CopyConstOnDevice();

  CopyDataOnDevice();

  getNumBlocksAndThreads(h_nspin, numBlocks, numThreadsPerBlock,nblocks_red, nthreads_red); // -> Matteo; sembrerebbe non fare nulla

  //normene=one/(three*V(h_side));

  // PrintEnergy(0);
  // PrintMagnetization(0);

  CUDA_SAFE_CALL( cudaMalloc((void **)&r_odata,numBlocks*sizeof(REAL)) );


  int current=0, counterAve = 0;

  un = fopen(OUT1, "w");
  //deux = fopen(OUT2, "w");

  // -> Matteo
  initChi(); 

  // texture binding
  cudaBindTexture(NULL, texture_Ra, d_Ra, ms_s);
  cudaBindTexture(NULL, texture_Ba, d_Ba, ms_s);
  cudaBindTexture(NULL, texture_Rb, d_Rb, ms_s);
  cudaBindTexture(NULL, texture_Bb, d_Bb, ms_s);
  cudaBindTexture(NULL, texture_Rc, d_Rc, ms_s);
  cudaBindTexture(NULL, texture_Bc, d_Bc, ms_s);
  cudaBindTexture(NULL, texture_Rd, d_Rd, ms_s);
  cudaBindTexture(NULL, texture_Bd, d_Bd, ms_s);

  cudaBindTexture(NULL, texture_Jpx, d_Jpx, ms_j);
  cudaBindTexture(NULL, texture_Jpy, d_Jpy, ms_j);
  cudaBindTexture(NULL, texture_Jpz, d_Jpz, ms_j);
  cudaBindTexture(NULL, texture_Jmx, d_Jmx, ms_j);
  cudaBindTexture(NULL, texture_Jmy, d_Jmy, ms_j);
  cudaBindTexture(NULL, texture_Jmz, d_Jmz, ms_j);

  //ComputeRandomGPU(RandNumBlocks,RandNumThreadsPerBlock,ZEROONE,d_Rand);

  /* Questa funzione è stata rimossa perche' causava problemi se invertito
     l'ordine di chiamata con initIra: a seconda dell'ordine di chiamata
     si puo' passare da \chi = 19.29(1) a \chi = 19.89(1) con seed = 10316;
     usando la nuova inizializzazione il risultato e' stabile sotto inversione
     dell'ordine di chiamata -> Matteo
   */


  //  printRand(randNum, d_Rand, OUTRAND); -> Controllo numeri Random di inizializzazione

  // prova per settare la cache -> Matteo
  cudaFuncSetCacheConfig(OverrelaxRed, cudaFuncCachePreferL1);
  cudaFuncSetCacheConfig(OverrelaxBlue, cudaFuncCachePreferL1);

  for(int myIndex=0; myIndex<firstIndex; myIndex++){
    if(myIndex == lista_tiempos[current]) current++; 
  }
  

  for(istep=firstIndex; istep<=pow(2.,nsteps)  /*current<BLOCKS*/ ; istep++) {

      TIMER_START;
      OverrelaxRed<<< dimGrid,dimBlock >>>(d_Ra, d_Rb, d_Rc, d_Rd,
					   d_Rand);

      CUDA_SAFE_CALL( cudaThreadSynchronize() );
      
      OverrelaxBlue<<< dimGrid,dimBlock>>>(d_Ba, d_Bb, d_Bc, d_Bd,
					   d_Rand);
      
      CUDA_SAFE_CALL( cudaThreadSynchronize() );

      //      ParisiRapuanoTester(4);

      TIMER_STOP;
      ttimeover+=TIMER_ELAPSED;

      /*      if(lista_tiempos[current]==istep && istep < THERM){
	current++;
	printf("current: %d\n", current);
	}
      if(lista_tiempos[current]==istep && istep >= THERM) {
	current++;
	printf("Yes: %d\n", current);*/ 
      if(lista_tiempos[current]==istep){
	current++;
	//printf("Yes: %d\n", current); 
		      
	/*CUDA_SAFE_CALL( cudaMemcpy( h_Ra, d_Ra, ms_s, cudaMemcpyDeviceToHost ) );
	CUDA_SAFE_CALL( cudaMemcpy( h_Rb, d_Rb, ms_s, cudaMemcpyDeviceToHost ) );
	
	CUDA_SAFE_CALL( cudaMemcpy( h_Ba, d_Ba, ms_s, cudaMemcpyDeviceToHost ) );
	CUDA_SAFE_CALL( cudaMemcpy( h_Bb, d_Bb, ms_s, cudaMemcpyDeviceToHost ) );
	
	CUDA_SAFE_CALL( cudaMemcpy( h_Rc, d_Rc, ms_s, cudaMemcpyDeviceToHost ) );
	CUDA_SAFE_CALL( cudaMemcpy( h_Rd, d_Rd, ms_s, cudaMemcpyDeviceToHost ) );
	
	CUDA_SAFE_CALL( cudaMemcpy( h_Bc, d_Bc, ms_s, cudaMemcpyDeviceToHost ) );
	CUDA_SAFE_CALL( cudaMemcpy( h_Bd, d_Bd, ms_s, cudaMemcpyDeviceToHost ) );*/
	
	//          PrintEnergy(istep,h_Rd,h_Bd);
	//          PrintEnergy(istep,h_Rc,h_Bc);
	//          PrintEnergy(istep,h_Rb,h_Bb);
	//          PrintEnergy(istep,h_Ra,h_Ba);
	
	//	  PrintMagnetization(istep);
	
	// la copia da device a host degli spin è già in Dump_Spin

	//Dump_Spin(istep); 
	
	//Chi();
	// counterAve++;
	//printf("chi: %.10f\n", chi);
	//getchar();
	//printf("counterAve: %d\n", counterAve);
      } 

      /*if(counterAve == MEASURE){
	printChi(current);
	initChi();
	current++;
	counterAve = 0;
	}*/
      
  }

  // -> Matteo
  fclose(un);
  //fclose(deux);

  //    printf("Total time for %d iterations of overrelaxation on a %d^3 cube: %f seconds\n", nsteps*overrelaxiter, h_side, ttimeover*0.000001);

  //    printf("Time for spin: %f microseconds\n",ttimeover/(overrelaxiter*pow(2.,nsteps)*V(h_side)));

  //Dump_Spin(nsteps);

  FreeDevice();

  FreeHost();

//  FreeRand();

  return 0;

}


//-------------------------------------------------------------------------------------------------------
// Overrelaxation
//-------------------------------------------------------------------------------------------------------

#define UPDATE_RED(a,an) \
	{SUM_111_3(j0,j1,\
	a[i]^tex1Dfetch(an, spx)^tex1Dfetch(texture_Jpx, i),\
	a[i]^tex1Dfetch(an, spy)^tex1Dfetch(texture_Jpy, i),\
	a[i]^tex1Dfetch(an, spz)^tex1Dfetch(texture_Jpz, i));\
	SUM_111_3(j2,j3,\
	a[i]^tex1Dfetch(an, smx)^tex1Dfetch(texture_Jmx, i),\
	a[i]^tex1Dfetch(an, smy)^tex1Dfetch(texture_Jmy, i),\
	a[i]^tex1Dfetch(an, smz)^tex1Dfetch(texture_Jmz, i));\
	SUM_33_6(j0,j1,j2,j3,sigma0,sigma1,sigma2);\
	RANDTWO31M1(random);\
	int condition4= -(random<d_flip_probability4);\
	int condition2=-(random<d_flip_probability2);\
	int condition0=-(random<d_flip_probability0);\
	MSC flip=	((~condition4) & LT_4_3BITS(sigma0,sigma1,sigma2) )| \
				(condition4 & (~condition2) &  LT_5_3BITS(sigma0,sigma1,sigma2) )|\
				(condition2 & (~condition0)& LT_6_3BITS(sigma0,sigma1,sigma2) )|\
				(condition0 & ALWAYS_3BITS(sigma0,sigma1,sigma2));\
	a[i]^=flip;\
	}
//	printf("condition0=%x,condition2=%x,condidion4=%x,LT_4_3BITS=%x ",condition0,condition2,condition4,LT_4_3BITS(sigma0,sigma1,sigma2));\
//	printf("a[%d]=%x,FLIP=%x\n",i,a[i],flip);\

#define UPDATE_BLUE(a,an) \
	{SUM_111_3(j0,j1,\
        a[i]^tex1Dfetch(an, spx)^tex1Dfetch(texture_Jmx, spx),\
	a[i]^tex1Dfetch(an, spy)^tex1Dfetch(texture_Jmy, spy),\
	a[i]^tex1Dfetch(an, spz)^tex1Dfetch(texture_Jmz, spz));\
	SUM_111_3(j2,j3,\
	a[i]^tex1Dfetch(an, smx)^tex1Dfetch(texture_Jpx, smx),\
	a[i]^tex1Dfetch(an, smy)^tex1Dfetch(texture_Jpy, smy),\
	a[i]^tex1Dfetch(an, smz)^tex1Dfetch(texture_Jpz, smz));\
	SUM_33_6(j0,j1,j2,j3,sigma0,sigma1,sigma2);\
	RANDTWO31M1(random);\
        int condition4= -(random<d_flip_probability4);\
	int condition2=-(random<d_flip_probability2);\
	int condition0=-(random<d_flip_probability0);\
	MSC flip=	((~condition4) & LT_4_3BITS(sigma0,sigma1,sigma2) )| \
				(condition4 & (~condition2) &  LT_5_3BITS(sigma0,sigma1,sigma2) )|\
				(condition2 & (~condition0)& LT_6_3BITS(sigma0,sigma1,sigma2) )|\
				(condition0 & ALWAYS_3BITS(sigma0,sigma1,sigma2));\
	a[i]^=flip;\
	}

// OverrelaxRed
//-------------------------------------------------------------------------------------------------------
/*__global__ void OverrelaxRed(MSC *a, MSC *an, MSC *b, MSC *bn,MSC *c, MSC *cn, MSC *d, MSC *dn,
                             MSC *Jpx, MSC *Jpy, MSC *Jpz,
                             MSC *Jmx, MSC *Jmy, MSC *Jmz, REAL* Rand) { -> Matteo */

/*__global__ void OverrelaxRed(__restrict MSC *a, __restrict const MSC *an, __restrict MSC *b, __restrict const MSC *bn, __restrict MSC *c, __restrict const MSC *cn, __restrict MSC *d, __restrict const MSC *dn,
                             __restrict const MSC *Jpx, __restrict const MSC *Jpy, __restrict const MSC *Jpz,
                             __restrict const MSC *Jmx, __restrict const MSC *Jmy, __restrict const MSC *Jmz, __restrict RANDOMTYPE* Rand) {
*/

__global__ void OverrelaxRed(__restrict MSC *a, __restrict MSC *b, __restrict MSC *c, __restrict MSC *d,
			     RANDOMTYPE *Rand) {


  unsigned int iy, iz;
  int diff;

  int spx, smx, spy, smy, spz, smz;
  int i;

  const unsigned int tid = threadIdx.x;

  const unsigned int num_threads = gridDim.x*blockDim.x;
  int istart;

  istart=blockDim.x*blockIdx.x;
  int threadIndex = tid + istart;

  //  unsigned int random=Rand[threadIndex]*((REAL)MASSIMO_RAND); -> Matteo

  unsigned int random=Rand[threadIndex];

  // generatore 2^31 - 1 -> Matteo
  unsigned int lo, hi;
  //  unsigned long long int swap;

  for(i=istart+tid; i<d_nspin; i+=num_threads) {

    iz=i>>d_sdbLsm1;
    iy=(i - (iz<<d_sdbLsm1))>>d_sdbLm1;

    //do the same thing if both $iz and $iy are even or odd
    diff=(iz&1)^(iy&1);

    spx=(i >> (-diff)) + (((i-(i&d_maskL)) + ((i+1)&d_maskL)) >> (diff-1));
    smx=(((i-(i&d_maskL)) + ((i-1)&d_maskL)) >> (-diff)) + (i>>(diff-1));
    spy=(i-(i&d_maskL2))+((i+d_Ldb2)&d_maskL2);
    smy=(i-(i&d_maskL2))+((i-d_Ldb2)&d_maskL2);
    spz=(i+d_Lsdb2) & d_maskL3;
    smz=(i-d_Lsdb2) & d_maskL3;

    UPDATE_RED(a, texture_Ba);
    UPDATE_RED(b, texture_Bb);
    UPDATE_RED(c, texture_Bc);
    UPDATE_RED(d, texture_Bd);
  }
  //  Rand[threadIndex]=double(random)/double(MASSIMO_RAND); -> Matteo

  Rand[threadIndex]=random;
  return;
}


//-------------------------------------------------------------------------------------------------------
// OverrelaxBlue
//-------------------------------------------------------------------------------------------------------
/*__global__ void OverrelaxBlue(MSC *a, MSC *an, MSC *b, MSC *bn,MSC *c, MSC *cn, MSC *d, MSC *dn,
                              MSC *Jpx, MSC *Jpy, MSC *Jpz,
                              MSC *Jmx, MSC *Jmy, MSC *Jmz, REAL* Rand) { -> Matteo */

__global__ void OverrelaxBlue(__restrict MSC *a, __restrict MSC *b, __restrict MSC *c, __restrict MSC *d,
			      RANDOMTYPE* Rand) {

/*
__global__ void OverrelaxBlue(MSC *a, const MSC *an, MSC *b, const MSC *bn, MSC *c, const MSC *cn, MSC *d, const MSC *dn,
			      const MSC *Jpx, const MSC *Jpy, const MSC *Jpz,
                              const MSC *Jmx, const MSC *Jmy, const MSC *Jmz, RANDOMTYPE* Rand) {
*/

//  int threadIndex = threadIdx.x + (gridDim.x*blockDim.x);


  unsigned int iy, iz;
  int diff;

  int spx, smx, spy, smy, spz, smz;
  int i;

  const unsigned int tid = threadIdx.x;

  const unsigned int num_threads = gridDim.x*blockDim.x;
  int istart;

  istart=blockDim.x*blockIdx.x;
  int threadIndex = tid + istart;

  //  unsigned int random=Rand[threadIndex]*((REAL)MASSIMO_RAND); -> Matteo

  unsigned int random=Rand[threadIndex];

  // generatore 2^31 - 1 -> Matteo
  unsigned int lo, hi; 
  //  unsigned long long int swap;

  for(i=istart+tid; i<d_nspin; i+=num_threads) {

    iz=i>>d_sdbLsm1;
    iy=(i - (iz<<d_sdbLsm1))>>d_sdbLm1;

    //do the same thing if both $iz and $iy are even or odd
    diff=(iz&1)^(iy&1);

    spx=(i >> (diff-1)) + (((i-(i&d_maskL)) + ((i+1)&d_maskL)) >> (-diff));
    smx=(((i-(i&d_maskL)) + ((i-1)&d_maskL)) >> (diff-1)) + (i>>(-diff));
    spy=(i-(i&d_maskL2))+((i+d_Ldb2)&d_maskL2);
    smy=(i-(i&d_maskL2))+((i-d_Ldb2)&d_maskL2);
    spz=(i+d_Lsdb2) & d_maskL3;
    smz=(i-d_Lsdb2) & d_maskL3;

    UPDATE_BLUE(a, texture_Ra);
    UPDATE_BLUE(b, texture_Rb);
    UPDATE_BLUE(c, texture_Rc);
    UPDATE_BLUE(d, texture_Rd); 
   
  }
  //  Rand[threadIndex]=double(random)/double(MASSIMO_RAND); -> Matteo
  Rand[threadIndex]=random;

  return;
}




//-------------------------------------------------------------------------------------------------------
// ComputeEnergy
//-------------------------------------------------------------------------------------------------------
__global__ void ComputeEnergy(MSC *a,  MSC *an,
                              MSC *Jpx, MSC *Jpy, MSC *Jpz,
                              MSC *Jmx, MSC *Jmy, MSC *Jmz,
                              int *energy,
                              int even) {
}


//-------------------------------------------------------------------------------------------------------
// ComputeHEnergy
//-------------------------------------------------------------------------------------------------------
void ComputeHEnergy(MSC *a,MSC *an,
                    MSC *Jpx, MSC *Jpy, MSC *Jpz,
                    MSC *Jmx, MSC *Jmy, MSC *Jmz,
                    MSC *energy,
                    int even) {
        return;
}

void NewComputeHEnergy(MSC *a,MSC *an,
                    MSC *Jpx, MSC *Jpy, MSC *Jpz,
                    MSC *Jmx, MSC *Jmy, MSC *Jmz,
                    LongMsc  *energy,
                    int even) {
  unsigned int iy, iz;
  int diff;
  int spx, smx, spy, smy, spz, smz;

  int i;
 // int EnergyVec[BITS];


  for(i=0; i<h_nspin; i++) {
//        LongMsc local;
//        AllocateLongMsc(&local,3);
    iz=i>>h_sdbLsm1;
    iy=(i - (iz<<h_sdbLsm1))>>h_sdbLm1;

    //do the same thing if both $iz and $iy are even or odd
    diff=(((iz&1)^(iy&1))^even);

    spx=(i >> (diff-1)) + (((i-(i&h_maskL)) + ((i+1)&h_maskL)) >> (-diff));
    smx=(((i-(i&h_maskL)) + ((i-1)&h_maskL)) >> (diff-1)) + (i>>(-diff));
    spy=(i-(i&h_maskL2))+((i+h_Ldb2)&h_maskL2);
    smy=(i-(i&h_maskL2))+((i-h_Ldb2)&h_maskL2);
    spz=(i+h_Lsdb2) & h_maskL3;
    smz=(i-h_Lsdb2) & h_maskL3;
        AddOneBitH(energy,a[i]^Jpx[i]^an[spx]);
        AddOneBitH(energy,a[i]^Jpy[i]^an[spy]);
        AddOneBitH(energy,a[i]^Jpz[i]^an[spz]);
        AddOneBitH(energy,a[i]^Jmx[i]^an[smx]);
        AddOneBitH(energy,a[i]^Jmy[i]^an[smy]);
        AddOneBitH(energy,a[i]^Jmz[i]^an[smz]);
  }

}


//-------------------------------------------------------------------------------------------------------
// ComputeRandomGPU
//-------------------------------------------------------------------------------------------------------
//void ComputeRandomGPU(int numBlocksRand, int numThreadsPerBlockRand, int mode, REAL* array){ -> Matteo

void ComputeRandomGPU(int numBlocksRand, int numThreadsPerBlockRand, int mode, RANDOMTYPE* array){
 
  seedRand(0);

  CUDA_SAFE_CALL( cudaThreadSynchronize() );

  dim3 dimGrid(numBlocksRand,1,1);

  dim3 dimBlock(numThreadsPerBlockRand,1,1);

  RandomGPU<<<dimGrid, numThreadsPerBlockRand>>>(array,mode);

  //cutilCheckMsg("RandomGPU() execution failed\n");

  CUDA_SAFE_CALL( cudaThreadSynchronize() );

}


//-------------------------------------------------------------------------------------------------------
// FreeRand
//-------------------------------------------------------------------------------------------------------
void FreeRand(void){

//  CUDA_SAFE_CALL( cudaFree(d_aRand) );
//  CUDA_SAFE_CALL( cudaFree(d_bRand) );
//  CUDA_SAFE_CALL( cudaFree(d_cRand) );
  CUDA_SAFE_CALL( cudaFree(d_Rand) );

}


//-------------------------------------------------------------------------------------------------------
// initRand
//-------------------------------------------------------------------------------------------------------
int initRand(int randomNumber){
  //printf("--->%d, %d\n",randomNumber, abs(randomNumber));

fflush(stdout);

  int n_per_rng;
  // int randNum; DA DISCOMMENTARE DOPO PROVA MESSO GLOBAL

  n_per_rng = iAlignUp(iDivUp(randomNumber, MT_RNG_COUNT), 2);

  randNum = MT_RNG_COUNT * n_per_rng;

  printf("--->%d\n",randNum); 

  //  printf("--->%d, %d\n",randNum, n_per_rng); 

  //CUDA_SAFE_CALL( cudaMalloc((void **)&d_aRand, randNum * sizeof(REAL)) );
  //CUDA_SAFE_CALL( cudaMalloc((void **)&d_bRand, randNum * sizeof(REAL)) );
//  CUDA_SAFE_CALL( cudaMalloc((void **)&d_cRand, randNum * sizeof(REAL)) );

//  CUDA_SAFE_CALL( cudaMalloc((void **)&d_Rand, randNum * sizeof(REAL)) ); -> Matteo
 CUDA_SAFE_CALL( cudaMalloc((void **)&d_Rand, randNum * sizeof(RANDOMTYPE)) );

 CUDA_SAFE_CALL( cudaMemcpyToSymbol("NPerRng",&n_per_rng,sizeof(int),0,cudaMemcpyHostToDevice) );

 return randNum;

}


//-------------------------------------------------------------------------------------------------------
// seedRand
//-------------------------------------------------------------------------------------------------------
void seedRand(int seed){

  if(seed==0) seed=(int)(drand48()*1023); // UNA SCHIFEZZA

  static mt_struct_stripped h_MT[MT_RNG_COUNT];
  static mt_struct MT[MT_RNG_COUNT];

  FILE *fd = Fopen("/home/matteo/Desktop/simulazioni/data/MersenneTwister.raw", "rb"); /* -> Matteo: per mettere l'eseguibile dove voglio */

  for (int i = 0; i < MT_RNG_COUNT; i++){
    //Inline structure size for compatibility,
    //since pointer types are 8-byte on 64-bit systems (unused *state variable)
    Fread(MT + i, 16 /* sizeof(mt_struct) */ * sizeof(int), 1, fd);

  }

  fclose(fd);

  fd = Fopen("/home/matteo/Desktop/simulazioni/data/MersenneTwister.dat", "rb"); /* -> Matteo: per mettere l'eseguibile dove voglio */

  Fread(h_MT, sizeof(h_MT), 1, fd);

  fclose(fd);

  //Need to be thread-safe
  mt_struct_stripped *MT_temp = (mt_struct_stripped *)Malloc(MT_RNG_COUNT * sizeof(mt_struct_stripped));

  for(int i = 0; i < MT_RNG_COUNT; i++){
    MT_temp[i] = h_MT[i];
    MT_temp[i].seed = seed;
  }

  CUDA_SAFE_CALL( cudaMemcpyToSymbol("ds_MT",MT_temp,sizeof(h_MT),0,cudaMemcpyHostToDevice) );

  free(MT_temp);

}


//-------------------------------------------------------------------------------------------------------
// printRand
//-------------------------------------------------------------------------------------------------------
//void printRand(int randomNumber,REAL* array,char* namefile){ -> Matteo

void printRand(int randomNumber,RANDOMTYPE* array,char* namefile){

  // REAL * h_Rand  = (REAL *)Malloc(randomNumber * sizeof(REAL)); -> Matteo

  RANDOMTYPE * h_Rand  = (RANDOMTYPE *)Malloc(randomNumber * sizeof(RANDOMTYPE));

  // CUDA_SAFE_CALL( cudaMemcpy(h_Rand, array, randomNumber * sizeof(REAL), cudaMemcpyDeviceToHost) ); -> Matteo

  CUDA_SAFE_CALL( cudaMemcpy(h_Rand, array, randomNumber * sizeof(RANDOMTYPE), cudaMemcpyDeviceToHost) );

  FILE* out = Fopen(namefile, "w");

  //for(int x=0;x<randomNumber;x++) { fprintf(out,"%f\n",h_Rand[x]); }

  for(int x=0;x<randomNumber;x++) { fprintf(out,"h_Rand[%d]: %u\n", x, h_Rand[x]); }

  fclose(out);
  Free(h_Rand);

}


////////////////////////////////////////////////////////////////////////////////
// Write MT_RNG_COUNT vertical lanes of NPerRng random numbers to *d_Random.
// For coalesced global writes MT_RNG_COUNT should be a multiple of warp size.
// Initial states for each generator are the same, since the states are
// initialized from the global seed. In order to improve distribution properties
// on small NPerRng supply dedicated (local) seed to each twister.
// The local seeds, in their turn, can be extracted from global seed
// by means of any simple random number generator, like LCG.
////////////////////////////////////////////////////////////////////////////////

//__global__ void RandomGPU( REAL *d_Random, int mode ){ -> Matteo

__global__ void RandomGPU(RANDOMTYPE *d_Random, int mode ){

  const int tid = blockDim.x * blockIdx.x + threadIdx.x;
  const int THREAD_N = blockDim.x * gridDim.x;

  int iState, iState1, iStateM, iOut;
  unsigned int mti, mti1, mtiM, x;
  unsigned int mt[MT_NN];

  REAL temp_rand;

  for(int iRng = tid; iRng < MT_RNG_COUNT; iRng += THREAD_N){
    //Load bit-vector Mersenne Twister parameters
    mt_struct_stripped config = ds_MT[iRng];

    //Initialize current state
    mt[0] = config.seed;
    for(iState = 1; iState < MT_NN; iState++)
      mt[iState] = (1812433253U * (mt[iState - 1] ^ (mt[iState - 1] >> 30)) + iState) & MT_WMASK;

    iState = 0;
    mti1 = mt[0];

    for(iOut = 0; iOut < NPerRng; iOut++){
      //iState1 = (iState +     1) % MT_NN
      //iStateM = (iState + MT_MM) % MT_NN
      do{
      iState1 = iState + 1;
      iStateM = iState + MT_MM;
      if(iState1 >= MT_NN) iState1 -= MT_NN;
      if(iStateM >= MT_NN) iStateM -= MT_NN;
      mti  = mti1;
      mti1 = mt[iState1];
      mtiM = mt[iStateM];
	
      x = (mti & MT_UMASK) | (mti1 & MT_LMASK);
      x = mtiM ^ (x >> 1) ^ ((x & 1) ? config.matrix_a : 0);
      mt[iState] = x;
      iState = iState1;
      
      //Tempering transformation
      x ^= (x >> MT_SHIFT0);
      x ^= (x << MT_SHIFTB) & config.mask_b;
      x ^= (x << MT_SHIFTC) & config.mask_c;
      x ^= (x >> MT_SHIFT1);
      }while(x == 0);
   
      temp_rand = ((REAL)x + 1.0f) / 4294967296.0f; //Convert to (0, 1] float
      // da non considerare per avere d_Random intero 

      if(mode==ZEROONE) { // write to global memory
	//        d_Random[iRng + iOut * MT_RNG_COUNT] = temp_rand; -> Matteo
        d_Random[iRng + iOut * MT_RNG_COUNT] = x; // che e' l'unica che ci interessa

      } else if(mode==ONEONE) {
        d_Random[iRng + iOut * MT_RNG_COUNT] = 2.0f*temp_rand-1.0f;
      }else {
        d_Random[iRng + iOut * MT_RNG_COUNT] = temp_rand - 0.5f;
      }

    }
  }
}


