// Matteo Lulli, Massimo Bernaschi, Giorgio Parisi 2013
// Physics Dept. 'Sapienza', University of Rome
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <cuda.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>

#include <curand.h>
#include <curand_kernel.h>
/* include MTGP host helper functions */
#include <curand_mtgp32_host.h>
/* include MTGP pre-computed parameter sets */
#include <curand_mtgp32dc_p_11213.h>

#include "cudaEA3D.h"
#include "MSC.h"
#include "myRNG.h"

#define MYMAXRAND TWO32M1
#define GPURAND(seed) RANDTWO31M1IMP(seed)
#define CPURAND(seed) RANDTWO31M1BIS(seed)

#define SP(a, m) (a&(~(-(a >= m))))
#define SM(a, m) (a+((-(a < 0))&m))
#define SMM(a, m) (a - 1 +((-(a < 1))&m))

#define KB1  (1<<10)
#define GB1  (1<<30)

#define MYCUDA_ERROR(functionCall)\
  {cudaError_t error_id = functionCall;\
  if(error_id != cudaSuccess){\
  fprintf(stderr, "\nError %s at line %d of %s\n",\
	  cudaGetErrorString(error_id), __LINE__, __FILE__);\
  return 1;\
  }\
  }

#define MYCUDA_LAST(myString)\
  {cudaError_t error_id = cudaGetLastError();	\
  if(error_id != cudaSuccess){\
  fprintf(stderr, "\nError %s at line %d of %s for %s\n",\
          cudaGetErrorString(error_id), __LINE__, __FILE__, myString);	\
  return 1;\
  }\
  }

#define MYCURAND_ERROR(functionCall)\
  {if(functionCall != CURAND_STATUS_SUCCESS){\
  fprintf(stderr, "\nCurand Error at line %d of %s\n",\
	  __LINE__, __FILE__);			      \
  return EXIT_FAILURE;\
  }\
  }

#define SUM_111_3(j0,j1,i0,i1,i2)\
MSC j1 = (i0) ^ (i1) ; \
MSC j0 = j1 ^ (i2) ;\
j1 = ((i0)&(i1)) ^ (j1&(i2)) ;\

//j0 e' il bit meno significativo: 2*j1+j0

#define SUM_33_6(j0,j1,j2,j3,sigma0,sigma1,sigma2)\
MSC sigma0 = j0 & j2 ;\
MSC sigma2 = j1 ^ j3 ;\
MSC sigma1 = sigma2 ^ sigma0 ;\
sigma2 = (j1 & j3) ^ (sigma2 & sigma0) ;\
sigma0 = j0 ^ j2 ;\
//sigma0 e' il bit meno significativo: 4*sigma2+2sigma1+sigma0

#define UPDATE_RED(a,an) \
	{SUM_111_3(j0,j1,\
	a[i]^tex1Dfetch(an, spx)^tex1Dfetch(texture_Jpx, i),\
	a[i]^tex1Dfetch(an, spy)^tex1Dfetch(texture_Jpy, i),\
	a[i]^tex1Dfetch(an, spz)^tex1Dfetch(texture_Jpz, i));\
	SUM_111_3(j2,j3,\
	a[i]^tex1Dfetch(an, smx)^tex1Dfetch(texture_Jmx, i),\
	a[i]^tex1Dfetch(an, smy)^tex1Dfetch(texture_Jmy, i),\
	a[i]^tex1Dfetch(an, smz)^tex1Dfetch(texture_Jmz, i));\
	SUM_33_6(j0,j1,j2,j3,sigma0,sigma1,sigma2);\
	GPURAND(seed);\
	int condition4 =-(seed<d_betaExp4);\
	int condition8 =-(seed<d_betaExp8);\
	int condition12=-(seed<d_betaExp12);\
	MSC flip = (((condition12 & sigma2) & sigma1) |			\
		    ((condition8 & sigma2) & sigma0) |			\
		    (((condition4 & sigma2) & (~sigma1)) & (~sigma0)) |	\
		    (~sigma2));						\
	a[i]^=flip;\
	}

#define UPDATE_RED_PRNG(a,an) \
	{SUM_111_3(j0,j1,\
	a[i]^tex1Dfetch(an, spx)^tex1Dfetch(texture_Jpx, i),\
	a[i]^tex1Dfetch(an, spy)^tex1Dfetch(texture_Jpy, i),\
	a[i]^tex1Dfetch(an, spz)^tex1Dfetch(texture_Jpz, i));\
	SUM_111_3(j2,j3,\
	a[i]^tex1Dfetch(an, smx)^tex1Dfetch(texture_Jmx, i),\
	a[i]^tex1Dfetch(an, smy)^tex1Dfetch(texture_Jmy, i),\
	a[i]^tex1Dfetch(an, smz)^tex1Dfetch(texture_Jmz, i));\
	SUM_33_6(j0,j1,j2,j3,sigma0,sigma1,sigma2);\
	int condition4 =-(seed<d_betaExp4);\
	int condition8 =-(seed<d_betaExp8);\
	int condition12=-(seed<d_betaExp12);\
	MSC flip = (((condition12 & sigma2) & sigma1) |			\
		    ((condition8 & sigma2) & sigma0) |			\
		    (((condition4 & sigma2) & (~sigma1)) & (~sigma0)) |	\
		    (~sigma2));						\
	a[i]^=flip;\
	}

#define UPDATE_RED_SLICED(a,an) \
	{SUM_111_3(j0,j1,\
	a[i]^tex1Dfetch(an, spx)^tex1Dfetch(texture_Jpx, i),\
	a[i]^tex1Dfetch(an, spy)^tex1Dfetch(texture_Jpy, i),\
	a[i]^tex1Dfetch(an, i)^tex1Dfetch(texture_Jpz, i));\
	SUM_111_3(j2,j3,\
	a[i]^tex1Dfetch(an, smx)^tex1Dfetch(texture_Jmx, i),\
	a[i]^tex1Dfetch(an, smy)^tex1Dfetch(texture_Jmy, i),\
	a[i]^tex1Dfetch(an, smz)^tex1Dfetch(texture_Jmz, i));\
	SUM_33_6(j0,j1,j2,j3,sigma0,sigma1,sigma2);\
	GPURAND(seed);\
	int condition4 =-(seed<d_betaExp4);\
	int condition8 =-(seed<d_betaExp8);\
	int condition12=-(seed<d_betaExp12);\
	MSC flip = (((condition12 & sigma2) & sigma1) |			\
		    ((condition8 & sigma2) & sigma0) |			\
		    (((condition4 & sigma2) & (~sigma1)) & (~sigma0)) |	\
		    (~sigma2));						\
	a[i]^=flip;\
	}

#define UPDATE_RED_SLICED_PRNG(a,an) \
	{SUM_111_3(j0,j1,\
	a[i]^tex1Dfetch(an, spx)^tex1Dfetch(texture_Jpx, i),\
	a[i]^tex1Dfetch(an, spy)^tex1Dfetch(texture_Jpy, i),\
	a[i]^tex1Dfetch(an, i)^tex1Dfetch(texture_Jpz, i));\
	SUM_111_3(j2,j3,\
	a[i]^tex1Dfetch(an, smx)^tex1Dfetch(texture_Jmx, i),\
	a[i]^tex1Dfetch(an, smy)^tex1Dfetch(texture_Jmy, i),\
	a[i]^tex1Dfetch(an, smz)^tex1Dfetch(texture_Jmz, i));\
	SUM_33_6(j0,j1,j2,j3,sigma0,sigma1,sigma2);\
	int condition4 =-(seed<d_betaExp4);\
	int condition8 =-(seed<d_betaExp8);\
	int condition12=-(seed<d_betaExp12);\
	MSC flip = (((condition12 & sigma2) & sigma1) |			\
		    ((condition8 & sigma2) & sigma0) |			\
		    (((condition4 & sigma2) & (~sigma1)) & (~sigma0)) |	\
		    (~sigma2));						\
	a[i]^=flip;\
	}


#define UPDATE_BLUE(a,an) \
	{SUM_111_3(j0,j1,\
        a[i]^tex1Dfetch(an, spx)^tex1Dfetch(texture_Jmx, spx),\
	a[i]^tex1Dfetch(an, spy)^tex1Dfetch(texture_Jmy, spy),\
	a[i]^tex1Dfetch(an, spz)^tex1Dfetch(texture_Jmz, spz));\
	SUM_111_3(j2,j3,\
	a[i]^tex1Dfetch(an, smx)^tex1Dfetch(texture_Jpx, smx),\
	a[i]^tex1Dfetch(an, smy)^tex1Dfetch(texture_Jpy, smy),\
	a[i]^tex1Dfetch(an, smz)^tex1Dfetch(texture_Jpz, smz));\
	SUM_33_6(j0,j1,j2,j3,sigma0,sigma1,sigma2);\
	GPURAND(seed);\
	int condition4 =-(seed<d_betaExp4);\
	int condition8 =-(seed<d_betaExp8);\
	int condition12=-(seed<d_betaExp12);\
	MSC flip = (((condition12 & sigma2) & sigma1) |			\
		    ((condition8 & sigma2) & sigma0) |			\
		    (((condition4 & sigma2) & (~sigma1)) & (~sigma0)) |	\
		    (~sigma2));						\
	a[i]^=flip;\
	}

#define UPDATE_BLUE_PRNG(a,an) \
	{SUM_111_3(j0,j1,\
        a[i]^tex1Dfetch(an, spx)^tex1Dfetch(texture_Jmx, spx),\
	a[i]^tex1Dfetch(an, spy)^tex1Dfetch(texture_Jmy, spy),\
	a[i]^tex1Dfetch(an, spz)^tex1Dfetch(texture_Jmz, spz));\
	SUM_111_3(j2,j3,\
	a[i]^tex1Dfetch(an, smx)^tex1Dfetch(texture_Jpx, smx),\
	a[i]^tex1Dfetch(an, smy)^tex1Dfetch(texture_Jpy, smy),\
	a[i]^tex1Dfetch(an, smz)^tex1Dfetch(texture_Jpz, smz));\
	SUM_33_6(j0,j1,j2,j3,sigma0,sigma1,sigma2);\
	int condition4 =-(seed<d_betaExp4);\
	int condition8 =-(seed<d_betaExp8);\
	int condition12=-(seed<d_betaExp12);\
	MSC flip = (((condition12 & sigma2) & sigma1) |			\
		    ((condition8 & sigma2) & sigma0) |			\
		    (((condition4 & sigma2) & (~sigma1)) & (~sigma0)) |	\
		    (~sigma2));						\
	a[i]^=flip;\
	}

#define UPDATE_BLUE_SLICED(a,an) \
	{SUM_111_3(j0,j1,\
        a[i]^tex1Dfetch(an, spx)^tex1Dfetch(texture_Jmx, spx),\
	a[i]^tex1Dfetch(an, spy)^tex1Dfetch(texture_Jmy, spy),\
	a[i]^tex1Dfetch(an, spz)^tex1Dfetch(texture_Jmz, spz));\
	SUM_111_3(j2,j3,\
	a[i]^tex1Dfetch(an, smx)^tex1Dfetch(texture_Jpx, smx),\
	a[i]^tex1Dfetch(an, smy)^tex1Dfetch(texture_Jpy, smy),\
	a[i]^tex1Dfetch(an, i)^tex1Dfetch(texture_Jpz, i));\
	SUM_33_6(j0,j1,j2,j3,sigma0,sigma1,sigma2);\
	GPURAND(seed);\
        int condition4 =-(seed<d_betaExp4);\
	int condition8 =-(seed<d_betaExp8);\
	int condition12=-(seed<d_betaExp12);\
	MSC flip = (((condition12 & sigma2) & sigma1) |			\
		    ((condition8 & sigma2) & sigma0) |			\
		    (((condition4 & sigma2) & (~sigma1)) & (~sigma0)) |	\
		    (~sigma2));						\
	a[i]^=flip;\
	}

#define UPDATE_BLUE_SLICED_PRNG(a,an) \
	{SUM_111_3(j0,j1,\
        a[i]^tex1Dfetch(an, spx)^tex1Dfetch(texture_Jmx, spx),\
	a[i]^tex1Dfetch(an, spy)^tex1Dfetch(texture_Jmy, spy),\
	a[i]^tex1Dfetch(an, spz)^tex1Dfetch(texture_Jmz, spz));\
	SUM_111_3(j2,j3,\
	a[i]^tex1Dfetch(an, smx)^tex1Dfetch(texture_Jpx, smx),\
	a[i]^tex1Dfetch(an, smy)^tex1Dfetch(texture_Jpy, smy),\
	a[i]^tex1Dfetch(an, i)^tex1Dfetch(texture_Jpz, i));\
	SUM_33_6(j0,j1,j2,j3,sigma0,sigma1,sigma2);\
        int condition4 =-(seed<d_betaExp4);\
	int condition8 =-(seed<d_betaExp8);\
	int condition12=-(seed<d_betaExp12);\
	MSC flip = (((condition12 & sigma2) & sigma1) |			\
		    ((condition8 & sigma2) & sigma0) |			\
		    (((condition4 & sigma2) & (~sigma1)) & (~sigma0)) |	\
		    (~sigma2));						\
	a[i]^=flip;\
	}



/* Controllo bit-wise per la trasformazione di coordinate*/
#define RED 0
#define BLU 1

typedef struct{
  int x, y, z, zz, k;
}posType;

static void myK2Vec(posType *pos, int k, int L, int col);
static void my2Real(posType *my, posType *real, int L);
static void realVec2K(posType *pos, int L);
static int my2RealIndex(int i, int L, int col);
static void mapsDefine(int *realI_R, int *realI_B, int *myI_R, int *myI_B, int *myI_B2R, int *myI_R2B, int L); 

static void myK2Vec(posType *pos, int k, int L, int col){
  int A = L*L;
  pos->k = k;
  pos->x = k%L;
  pos->y = (k/L)%L;
  pos->z = k/A;
  if(col == RED) pos->zz = 2*pos->z;
  else  pos->zz = 2*pos->z + 1;
  return;
}

static void my2Real(posType *my, posType *real, int L){
  real->x = my->x;
  real->y = (my->x + my->y)%L;
  real->zz = real->z = (my->y + my->zz)%L;  
  return;
}

static void realVec2K(posType *pos, int L){
  int A = L*L;
  pos->k = (pos->x + pos->y*L + pos->z*A)/2; 
  return;
}

static int my2RealIndex(int i, int L, int col){
  posType myPos, realPos;
  myK2Vec(&myPos, i, L, col);
  my2Real(&myPos, &realPos, L);
  realVec2K(&realPos, L);
  return realPos.k;
}

static void mapsDefine(int *realI_R, int *realI_B, int *myI_R, int *myI_B, int *myI_B2R, int *myI_R2B, int L){
  int hV = L*L*L/2;
  for(int i=0; i<hV; i++){
    realI_R[i] = my2RealIndex(i, L, RED);
    realI_B[i] = my2RealIndex(i, L, BLU);
  }

  for(int i=0; i<hV; i++){
    myI_R[realI_R[i]] = i; 
    myI_B[realI_B[i]] = i; 
  }

  for(int i=0; i<hV; i++){
    myI_B2R[myI_B[i]] = myI_R[i]; 
    myI_R2B[myI_R[i]] = myI_B[i]; 
  }
  return;
}

static void arrayTransformMSC(MSC *array, int *map, int size){
  MSC *swap = (MSC *)malloc(size*sizeof(MSC));
  for(int i=0; i<size; i++) swap[i] = 0;
  for(int i=0; i<size; i++) swap[map[i]] = array[i];
  for(int i=0; i<size; i++) array[i] = swap[i];
  free(swap);
  return;
}

static void arrayTransformRNGT(RNGT *array, int *map, int size){
  RNGT *swap = (RNGT *)malloc(size*sizeof(RNGT));
  for(int i=0; i<size; i++) swap[i] = 0;
  for(int i=0; i<size; i++) swap[map[i]] = array[i];
  for(int i=0; i<size; i++) array[i] = swap[i];
  free(swap);
  return;
}

/* fine controllo trasformazione di coordinate */

__device__ __constant__ MSC d_sdbLs;
__device__ __constant__ MSC d_sdbL;
__device__ __constant__ MSC d_sdbLm1;
__device__ __constant__ MSC d_sdbLsm1; 
__device__ __constant__ MSC d_maskL;
__device__ __constant__ MSC d_Ldb2;
__device__ __constant__ MSC d_Lsdb2;
__device__ __constant__ MSC d_maskL2; 
__device__ __constant__ MSC d_maskL3;

__device__ __constant__ RNGT d_betaExp4;
__device__ __constant__ RNGT d_betaExp8;
__device__ __constant__ RNGT d_betaExp12;

__device__ __constant__ int d_replicas;
__device__ __constant__ int d_sysN;
__device__ __constant__ int d_hV;
__device__ __constant__ int d_A;
__device__ __constant__ int d_L;
__device__ __constant__ int d_hL;
__device__ __constant__ int d_hA;

__device__ __constant__ int d_threads;
__device__ __constant__ int d_iraLen;

__device__ __constant__ int d_block_xStride;
__device__ __constant__ int d_block_yStride;
__device__ __constant__ int d_block_xStrideS;
__device__ __constant__ int d_block_yStrideS;
__device__ __constant__ int d_loopStride;
__device__ __constant__ int d_rngOffset;

MSC hh_sdbLs;
MSC hh_sdbL;
MSC hh_sdbLm1;
MSC hh_sdbLsm1; 
MSC hh_maskL;
MSC hh_Ldb2;
MSC hh_Lsdb2;
MSC hh_maskL2; 
MSC hh_maskL3; 

RNGT h_betaExp4;
RNGT h_betaExp8;
RNGT h_betaExp12;

texture<MSC> texture_Jpx;
texture<MSC> texture_Jpy;
texture<MSC> texture_Jpz;
texture<MSC> texture_Jmx;
texture<MSC> texture_Jmy;
texture<MSC> texture_Jmz;

texture<MSC> texture_R0;
texture<MSC> texture_B0;
texture<MSC> texture_R1;
texture<MSC> texture_B1;
texture<MSC> texture_R2;
texture<MSC> texture_B2;
texture<MSC> texture_R3;
texture<MSC> texture_B3;

RNGT globalSeed;

static int masksInit(sysEA3D_t *sys);
static int MCWeightsInit(sysEA3D_t *sys);
static int dMemInit(sysEA3D_t *sys);
static int hMemInit(sysEA3D_t *sys);

static int H2DSpins(sysEA3D_t *sys);
static int H2DJ(sysEA3D_t *sys);
static int H2DSysCopy(sysEA3D_t *sys);

static int D2HSpins(sysEA3D_t *sys);

static void usage(char **argv);

static RNGT initSeed31M1(void);
static MSC flatMSCbitDist31M1(void);
static int JInit(sysEA3D_t *sys);
static int SpinsInit(sysEA3D_t *sys);

static int masksInit(sysEA3D_t *sys){
  int L = sys->L;
  int V = sys->V;
  hh_sdbLs = findLog2(L*L); 
  hh_sdbL = findLog2(L);
  hh_sdbLm1 = hh_sdbL-1;
  hh_sdbLsm1 = hh_sdbLs-1;
  hh_maskL = (L>>1)-1;
  hh_Ldb2 = L>>1;
  hh_Lsdb2 = (L*L)>>1;
  hh_maskL2 = ((L*L)>>1)-1;
  hh_maskL3 = (V>>1)-1;

  //  spz=(i+d_Lsdb2) & d_maskL3;
  //  smz=(i-d_Lsdb2) & d_maskL3;

  MYCUDA_ERROR( cudaMemcpyToSymbol(d_sdbLs, &hh_sdbLs, sizeof(MSC)) );
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_sdbL, &hh_sdbL, sizeof(MSC)) );
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_sdbLm1, &hh_sdbLm1, sizeof(MSC)) );
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_sdbLsm1, &hh_sdbLsm1, sizeof(MSC)) );
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_maskL, &hh_maskL, sizeof(MSC)) );
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_Ldb2, &hh_Ldb2, sizeof(MSC)) );
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_Lsdb2, &hh_Lsdb2, sizeof(MSC)) );
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_maskL2, &hh_maskL2, sizeof(MSC)) );
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_maskL3, &hh_maskL3, sizeof(MSC)) );

  int hV = sys->V/2;
  int A = L*L;
  int hL = L/2;
  int hA = A/2;
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_replicas, &sys->replicas, sizeof(int)) );
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_sysN, &sys->sysN, sizeof(int)) );
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_hV, &hV, sizeof(int)) );
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_A, &A, sizeof(int)) );
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_L, &L, sizeof(int)) );
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_hL, &hL, sizeof(int)) );
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_hA, &hA, sizeof(int)) );

  return 0;
}

static int MCWeightsInit(sysEA3D_t *sys){
  h_betaExp4 = ((RNGT)((double)MYMAXRAND*exp(-4./sys->beta)));
  h_betaExp8 = ((RNGT)((double)MYMAXRAND*exp(-8./sys->beta)));
  h_betaExp12 = ((RNGT)((double)MYMAXRAND*exp(-12./sys->beta)));

  MYCUDA_ERROR( cudaMemcpyToSymbol(d_betaExp4, &h_betaExp4, sizeof(RNGT)) );
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_betaExp8, &h_betaExp8, sizeof(RNGT)) );
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_betaExp12,&h_betaExp12,sizeof(RNGT)) );

  return 0;
}

static int dMemInit(sysEA3D_t *sys){
  MYCUDA_ERROR( cudaMalloc((void **) &sys->d_Jpx, sys->NJ*sizeof(MSC)) );
  MYCUDA_ERROR( cudaMalloc((void **) &sys->d_Jpy, sys->NJ*sizeof(MSC)) );
  MYCUDA_ERROR( cudaMalloc((void **) &sys->d_Jpz, sys->NJ*sizeof(MSC)) );
  MYCUDA_ERROR( cudaMalloc((void **) &sys->d_Jmx, sys->NJ*sizeof(MSC)) );
  MYCUDA_ERROR( cudaMalloc((void **) &sys->d_Jmy, sys->NJ*sizeof(MSC)) );
  MYCUDA_ERROR( cudaMalloc((void **) &sys->d_Jmz, sys->NJ*sizeof(MSC)) );

  MYCUDA_ERROR( cudaBindTexture(NULL, texture_Jpx, sys->d_Jpx, sys->NJ*sizeof(MSC)) );
  MYCUDA_ERROR( cudaBindTexture(NULL, texture_Jpy, sys->d_Jpy, sys->NJ*sizeof(MSC)) );
  MYCUDA_ERROR( cudaBindTexture(NULL, texture_Jpz, sys->d_Jpz, sys->NJ*sizeof(MSC)) );
  MYCUDA_ERROR( cudaBindTexture(NULL, texture_Jmx, sys->d_Jmx, sys->NJ*sizeof(MSC)) );
  MYCUDA_ERROR( cudaBindTexture(NULL, texture_Jmy, sys->d_Jmy, sys->NJ*sizeof(MSC)) );
  MYCUDA_ERROR( cudaBindTexture(NULL, texture_Jmz, sys->d_Jmz, sys->NJ*sizeof(MSC)) );

  MYCUDA_ERROR( cudaMalloc((void **) &sys->d_reds0, sys->NS*sizeof(MSC)) );
  MYCUDA_ERROR( cudaMalloc((void **) &sys->d_reds1, sys->NS*sizeof(MSC)) );
  MYCUDA_ERROR( cudaMalloc((void **) &sys->d_reds2, sys->NS*sizeof(MSC)) );
  MYCUDA_ERROR( cudaMalloc((void **) &sys->d_reds3, sys->NS*sizeof(MSC)) );
  
  MYCUDA_ERROR( cudaMalloc((void **) &sys->d_blues0, sys->NS*sizeof(MSC)) );
  MYCUDA_ERROR( cudaMalloc((void **) &sys->d_blues1, sys->NS*sizeof(MSC)) );
  MYCUDA_ERROR( cudaMalloc((void **) &sys->d_blues2, sys->NS*sizeof(MSC)) );
  MYCUDA_ERROR( cudaMalloc((void **) &sys->d_blues3, sys->NS*sizeof(MSC)) );
  
  MYCUDA_ERROR( cudaBindTexture(NULL, texture_R0, sys->d_reds0, sys->NS*sizeof(MSC)) );
  MYCUDA_ERROR( cudaBindTexture(NULL, texture_R1, sys->d_reds1, sys->NS*sizeof(MSC)) );
  MYCUDA_ERROR( cudaBindTexture(NULL, texture_R2, sys->d_reds2, sys->NS*sizeof(MSC)) );
  MYCUDA_ERROR( cudaBindTexture(NULL, texture_R3, sys->d_reds3, sys->NS*sizeof(MSC)) );
  
  MYCUDA_ERROR( cudaBindTexture(NULL, texture_B0, sys->d_blues0, sys->NS*sizeof(MSC)) );
  MYCUDA_ERROR( cudaBindTexture(NULL, texture_B1, sys->d_blues1, sys->NS*sizeof(MSC)) );
  MYCUDA_ERROR( cudaBindTexture(NULL, texture_B2, sys->d_blues2, sys->NS*sizeof(MSC)) );
  MYCUDA_ERROR( cudaBindTexture(NULL, texture_B3, sys->d_blues3, sys->NS*sizeof(MSC)) );

  return 0;
}

static int dMemFree(sysEA3D_t *sys){
  MYCUDA_ERROR( cudaFree(sys->d_Jpx) );
  MYCUDA_ERROR( cudaFree(sys->d_Jpy) );
  MYCUDA_ERROR( cudaFree(sys->d_Jpz) );
  MYCUDA_ERROR( cudaFree(sys->d_Jmx) );
  MYCUDA_ERROR( cudaFree(sys->d_Jmy) );
  MYCUDA_ERROR( cudaFree(sys->d_Jmz) );

  MYCUDA_ERROR( cudaUnbindTexture(texture_Jpx) );
  MYCUDA_ERROR( cudaUnbindTexture(texture_Jpy) );
  MYCUDA_ERROR( cudaUnbindTexture(texture_Jpz) );
  MYCUDA_ERROR( cudaUnbindTexture(texture_Jmx) );
  MYCUDA_ERROR( cudaUnbindTexture(texture_Jmy) );
  MYCUDA_ERROR( cudaUnbindTexture(texture_Jmz) );

  MYCUDA_ERROR( cudaFree(sys->d_reds0) );
  MYCUDA_ERROR( cudaFree(sys->d_reds1) );
  MYCUDA_ERROR( cudaFree(sys->d_reds2) );
  MYCUDA_ERROR( cudaFree(sys->d_reds3) );
  
  MYCUDA_ERROR( cudaFree(sys->d_blues0) );
  MYCUDA_ERROR( cudaFree(sys->d_blues1) );
  MYCUDA_ERROR( cudaFree(sys->d_blues2) );
  MYCUDA_ERROR( cudaFree(sys->d_blues3) );
  
  MYCUDA_ERROR( cudaUnbindTexture(texture_R0) );
  MYCUDA_ERROR( cudaUnbindTexture(texture_R1) );
  MYCUDA_ERROR( cudaUnbindTexture(texture_R2) );
  MYCUDA_ERROR( cudaUnbindTexture(texture_R3) );
  
  MYCUDA_ERROR( cudaUnbindTexture(texture_B0) );
  MYCUDA_ERROR( cudaUnbindTexture(texture_B1) );
  MYCUDA_ERROR( cudaUnbindTexture(texture_B2) );
  MYCUDA_ERROR( cudaUnbindTexture(texture_B3) );

  return 0;
}

static int hMemInit2(sysEA3D_t *sys){
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_Jpx, sys->NJ*sizeof(MSC), cudaHostAllocDefault) );
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_Jpy, sys->NJ*sizeof(MSC), cudaHostAllocDefault) );
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_Jpz, sys->NJ*sizeof(MSC), cudaHostAllocDefault) );
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_Jmx, sys->NJ*sizeof(MSC), cudaHostAllocDefault) );
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_Jmy, sys->NJ*sizeof(MSC), cudaHostAllocDefault) );
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_Jmz, sys->NJ*sizeof(MSC), cudaHostAllocDefault) );

  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_reds0, sys->NS*sizeof(MSC), cudaHostAllocDefault) );
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_reds1, sys->NS*sizeof(MSC), cudaHostAllocDefault) );
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_reds2, sys->NS*sizeof(MSC), cudaHostAllocDefault) );
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_reds3, sys->NS*sizeof(MSC), cudaHostAllocDefault) );
  
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_blues0, sys->NS*sizeof(MSC), cudaHostAllocDefault) );
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_blues1, sys->NS*sizeof(MSC), cudaHostAllocDefault) );
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_blues2, sys->NS*sizeof(MSC), cudaHostAllocDefault) );
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_blues3, sys->NS*sizeof(MSC), cudaHostAllocDefault) );

  return 0;
}

static int hMemFree(sysEA3D_t *sys){
  MYCUDA_ERROR( cudaFreeHost(sys->h_Jpx) );
  MYCUDA_ERROR( cudaFreeHost(sys->h_Jpy) );
  MYCUDA_ERROR( cudaFreeHost(sys->h_Jpz) );
  MYCUDA_ERROR( cudaFreeHost(sys->h_Jmx) );
  MYCUDA_ERROR( cudaFreeHost(sys->h_Jmy) );
  MYCUDA_ERROR( cudaFreeHost(sys->h_Jmz) );

  MYCUDA_ERROR( cudaFreeHost(sys->h_reds0) );
  MYCUDA_ERROR( cudaFreeHost(sys->h_reds1) );
  MYCUDA_ERROR( cudaFreeHost(sys->h_reds2) );
  MYCUDA_ERROR( cudaFreeHost(sys->h_reds3) );
  
  MYCUDA_ERROR( cudaFreeHost(sys->h_blues0) );
  MYCUDA_ERROR( cudaFreeHost(sys->h_blues1) );
  MYCUDA_ERROR( cudaFreeHost(sys->h_blues2) );
  MYCUDA_ERROR( cudaFreeHost(sys->h_blues3) );

  return 0;
}

static int hMemInit(sysEA3D_t *sys){
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_Jpx, sys->NJ*sizeof(MSC), cudaHostAllocDefault) );
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_Jpy, sys->NJ*sizeof(MSC), cudaHostAllocDefault) );
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_Jpz, sys->NJ*sizeof(MSC), cudaHostAllocDefault) );
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_Jmx, sys->NJ*sizeof(MSC), cudaHostAllocDefault) );
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_Jmy, sys->NJ*sizeof(MSC), cudaHostAllocDefault) );
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_Jmz, sys->NJ*sizeof(MSC), cudaHostAllocDefault) );

  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_reds0, sys->NS*sizeof(MSC), cudaHostAllocDefault) );
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_reds1, sys->NS*sizeof(MSC), cudaHostAllocDefault) );
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_reds2, sys->NS*sizeof(MSC), cudaHostAllocDefault) );
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_reds3, sys->NS*sizeof(MSC), cudaHostAllocDefault) );
  
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_blues0, sys->NS*sizeof(MSC), cudaHostAllocDefault) );
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_blues1, sys->NS*sizeof(MSC), cudaHostAllocDefault) );
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_blues2, sys->NS*sizeof(MSC), cudaHostAllocDefault) );
  MYCUDA_ERROR( cudaHostAlloc((void **) &sys->h_blues3, sys->NS*sizeof(MSC), cudaHostAllocDefault) );

  unsigned long long int newValue = 0, oldValue = 0;
  int newIndex = 0;
  int steps = sys->steps;
  unsigned long long int tMax = (unsigned long long int)rint(pow(2.,steps));

  if(sys->recoverF == 0){
    int counter = 0;
    
    for(int i=0; i<=4*steps; i++){
      newValue = (unsigned long long int)rint(pow(2.,i/4.));
      if(newValue > oldValue){ 
	counter++; 
	oldValue = newValue; 
      }
    }
    
    int dumpN = (4*steps + 1)*(4*steps + 1) + counter;
    
    sys->dumpTimes = (unsigned long long int *)malloc(dumpN*sizeof(unsigned long long int));
    if(sys->dumpTimes == NULL){
      fprintf(stderr,"\nNo Alloc dumpTimes hMemInit\n");
      return 1;
    }
    
    int k = 0;
    
    sys->dumpTimes[0] = 1; newValue = 0; k = 0;
    for(int i=0; i<=4*steps; i++){
      newValue = (unsigned long long int)rint(pow(2.,i/4.));
      if(newValue > sys->dumpTimes[k]){ 
	k++;
	sys->dumpTimes[k] = newValue;
      }
    }
    
    k++;
    
    for(int i=0; i<=4*steps; i++){
      for(int j=0; j<=4*steps; j++){
	unsigned long long int swap = (unsigned long long int)rint(pow(2.,i/4.)) + (unsigned long long int)rint(pow(2.,j/4.));
	if(swap <= tMax){
	  sys->dumpTimes[k] = swap; 
	  k++;
	}
      }
    }
    
    // sorting
    for(int i=0; i<k; i++){
      for(int j=0; j<k - i - 1; j++){
	if(sys->dumpTimes[j] > sys->dumpTimes[j + 1]){
	  unsigned long long int swap = sys->dumpTimes[j];
	  sys->dumpTimes[j] = sys->dumpTimes[j + 1];
	  sys->dumpTimes[j + 1] = swap;
	}
      }
    }
    
    // deleting replicas
    newIndex = 0;
    for(int i=0; i<k; i++){
      if(sys->dumpTimes[newIndex] != sys->dumpTimes[i]){
	newIndex++;
	sys->dumpTimes[newIndex] = sys->dumpTimes[i];
      }
    }

    /*for(int i=0; i<=newIndex; i++){
      printf("%llu\n", sys->dumpTimes[i]);
      }*/

    sys->dumpN = newIndex + 1;

  } 

  if(sys->recoverF == 1){
    int counter = 0;
    
    for(int i=0; i<=4*steps; i++){
      newValue = (unsigned long long int)rint(pow(2.,i/4.));
      if(newValue > oldValue && (newValue + sys->startValue) <= tMax){ 
	counter++; 
	oldValue = newValue; 
      }
    }

    int dumpN = counter;
    
    sys->dumpTimes = (unsigned long long int *)malloc(dumpN*sizeof(unsigned long long int));
    if(sys->dumpTimes == NULL){
      fprintf(stderr,"\nNo Alloc dumpTimes hMemInit\n");
      return 1;
    }

    int k = 0;
    
    sys->dumpTimes[0] = sys->startValue + 1; newValue = 0; k = 0;
    for(int i=0; i<=4*steps; i++){
      newValue = (unsigned long long int)rint(pow(2.,i/4.));
      if((newValue + sys->startValue) > sys->dumpTimes[k] && (newValue + sys->startValue) <= tMax){ 
	k++;
	sys->dumpTimes[k] = newValue + sys->startValue;
      }
    }

    /*for(int i=0; i<dumpN; i++){
      printf("%llu\n", sys->dumpTimes[i]);
      }*/

    sys->dumpN = dumpN;  
  }
 
  return 0;
}

static int H2DSpins(sysEA3D_t *sys){
  MYCUDA_ERROR( cudaMemcpy(sys->d_reds0, sys->h_reds0, sys->NS*sizeof(MSC), cudaMemcpyHostToDevice) );
  MYCUDA_ERROR( cudaMemcpy(sys->d_reds1, sys->h_reds1, sys->NS*sizeof(MSC), cudaMemcpyHostToDevice) );
  MYCUDA_ERROR( cudaMemcpy(sys->d_reds2, sys->h_reds2, sys->NS*sizeof(MSC), cudaMemcpyHostToDevice) );
  MYCUDA_ERROR( cudaMemcpy(sys->d_reds3, sys->h_reds3, sys->NS*sizeof(MSC), cudaMemcpyHostToDevice) );
  
  MYCUDA_ERROR( cudaMemcpy(sys->d_blues0, sys->h_blues0, sys->NS*sizeof(MSC), cudaMemcpyHostToDevice) );
  MYCUDA_ERROR( cudaMemcpy(sys->d_blues1, sys->h_blues1, sys->NS*sizeof(MSC), cudaMemcpyHostToDevice) );
  MYCUDA_ERROR( cudaMemcpy(sys->d_blues2, sys->h_blues2, sys->NS*sizeof(MSC), cudaMemcpyHostToDevice) );
  MYCUDA_ERROR( cudaMemcpy(sys->d_blues3, sys->h_blues3, sys->NS*sizeof(MSC), cudaMemcpyHostToDevice) );

  return 0;
}

static int H2DJ(sysEA3D_t *sys){
  
  MYCUDA_ERROR( cudaMemcpy(sys->d_Jpx, sys->h_Jpx, sys->NJ*sizeof(MSC), cudaMemcpyHostToDevice) );
  MYCUDA_ERROR( cudaMemcpy(sys->d_Jpy, sys->h_Jpy, sys->NJ*sizeof(MSC), cudaMemcpyHostToDevice) );
  MYCUDA_ERROR( cudaMemcpy(sys->d_Jpz, sys->h_Jpz, sys->NJ*sizeof(MSC), cudaMemcpyHostToDevice) );
  MYCUDA_ERROR( cudaMemcpy(sys->d_Jmx, sys->h_Jmx, sys->NJ*sizeof(MSC), cudaMemcpyHostToDevice) );
  MYCUDA_ERROR( cudaMemcpy(sys->d_Jmy, sys->h_Jmy, sys->NJ*sizeof(MSC), cudaMemcpyHostToDevice) );
  MYCUDA_ERROR( cudaMemcpy(sys->d_Jmz, sys->h_Jmz, sys->NJ*sizeof(MSC), cudaMemcpyHostToDevice) );

  return 0;
}

static int H2DSysCopy(sysEA3D_t *sys){
  if(H2DSpins(sys)) return 1;
  if(H2DJ(sys)) return 1;
  return 0;
}

static int D2HSpins(sysEA3D_t *sys){
  MYCUDA_ERROR( cudaMemcpy(sys->h_reds0, sys->d_reds0, sys->NS*sizeof(MSC), cudaMemcpyDeviceToHost) );
  MYCUDA_ERROR( cudaMemcpy(sys->h_reds1, sys->d_reds1, sys->NS*sizeof(MSC), cudaMemcpyDeviceToHost) );
  MYCUDA_ERROR( cudaMemcpy(sys->h_reds2, sys->d_reds2, sys->NS*sizeof(MSC), cudaMemcpyDeviceToHost) );
  MYCUDA_ERROR( cudaMemcpy(sys->h_reds3, sys->d_reds3, sys->NS*sizeof(MSC), cudaMemcpyDeviceToHost) );
  
  MYCUDA_ERROR( cudaMemcpy(sys->h_blues0, sys->d_blues0, sys->NS*sizeof(MSC), cudaMemcpyDeviceToHost) );
  MYCUDA_ERROR( cudaMemcpy(sys->h_blues1, sys->d_blues1, sys->NS*sizeof(MSC), cudaMemcpyDeviceToHost) );
  MYCUDA_ERROR( cudaMemcpy(sys->h_blues2, sys->d_blues2, sys->NS*sizeof(MSC), cudaMemcpyDeviceToHost) );
  MYCUDA_ERROR( cudaMemcpy(sys->h_blues3, sys->d_blues3, sys->NS*sizeof(MSC), cudaMemcpyDeviceToHost) );

  return 0;
}

static void usage(char **argv){
  fprintf(stderr, "usage: %s <L> <T> <steps> <replicas> <sysN> <blockSize> <k> <kernelChoice> <gpu> <dump> <analysis> <ram> <recover> <recoverTime>\n", argv[0]);
  return;
}


static RNGT initSeed31M1(void){
  RNGT seed = 0;
  size_t chkcp = 0;
  FILE *urandom = fopen("/dev/urandom", "rb");
  if(urandom == NULL){
    fprintf(stderr, "\nError opening /dev/urandom\n");
    return 0;
  }
  while(seed == 0){
    chkcp = fread(&seed, sizeof(RNGT), 1, urandom);
    seed = seed&TWO31M1;
    if(chkcp^1) seed = 0;
  }
  fclose(urandom);
  return seed;
}

static MSC flatMSCbitDist31M1(void){
  MSC number = 0;

  for(int i=0; i<MSCBITS; i++){
    number <<= 1;
    RANDTWO31M1BIS(globalSeed);
    MSC bit = (((float)globalSeed/TWO31M1) > 0.5 ? 0 : 1);
    number ^= bit;
  }
  return number;
}

static int JInit(sysEA3D_t *sys){
  for(int i=0; i<sys->NJ; i++){
    if(i%(sys->V/2) == 0){
      globalSeed = initSeed31M1();
      if(globalSeed == 0) return 1;
    }
    sys->h_Jpx[i] = flatMSCbitDist31M1();
    sys->h_Jpy[i] = flatMSCbitDist31M1();
    sys->h_Jpz[i] = flatMSCbitDist31M1();
    sys->h_Jmx[i] = flatMSCbitDist31M1();
    sys->h_Jmy[i] = flatMSCbitDist31M1();
    sys->h_Jmz[i] = flatMSCbitDist31M1();
  }
  
  return 0;
}

static int SpinsInit(sysEA3D_t *sys){

  for(int i=0; i<sys->NS; i++){
    if(i%(sys->V/2) == 0){
      globalSeed = initSeed31M1();
      if(globalSeed == 0) return 1;
    }
    
    sys->h_reds0[i] = flatMSCbitDist31M1();
    sys->h_reds1[i] = flatMSCbitDist31M1();
    sys->h_reds2[i] = flatMSCbitDist31M1();
    sys->h_reds3[i] = flatMSCbitDist31M1();
    
    sys->h_blues0[i] = flatMSCbitDist31M1();
    sys->h_blues1[i] = flatMSCbitDist31M1();
    sys->h_blues2[i] = flatMSCbitDist31M1();
    sys->h_blues3[i] = flatMSCbitDist31M1();
  }

  return 0;
}

extern int dummyDirsMaker(sysEA3D_t *sys){
  int sysRet = 0;
  char cmd[200];

  if(sys->analysisF == 1){
    snprintf(cmd, sizeof(cmd), "mkdir %s", sys->devShm);
    sysRet = system(cmd);
    if(sysRet) return 1;
  }


  if(sys->dumpF == 1 && sys->recoverF == 0){
    for(int i=0; i<sys->sysN; i++){
      snprintf(cmd, sizeof(cmd), "mkdir ./dummyRun%d", i);
      sysRet = system(cmd);
      if(sysRet) return 1;
    }
  }

  if(sys->recoverF == 1){
    snprintf(cmd, sizeof(cmd), "mkdir ./recover%016llX/", sys->startValue);
    sysRet = system(cmd);
    if(sysRet) return 1;

    for(int i=0; i<sys->sysN; i++){
      snprintf(cmd, sizeof(cmd), "mkdir ./recover%016llX/dummyRun%d", sys->startValue, i);
      sysRet = system(cmd);
      if(sysRet) return 1;
    }
  }

  return 0;
}

extern int JRead(sysEA3D_t *sys){
  FILE *jdump = fopen("j_dump_start", "rb");
  if(jdump == NULL){
    fprintf(stderr, "\nError opening j_dump_start\n");
    return 1;
  }
  int chkcp = 0;

  for(int i=0; i<sys->sysN; i++){
    int offset = i*(sys->V/2)*6;
    chkcp += fread(sys->h_Jpx + offset, sizeof(MSC), sys->V/2, jdump);
    chkcp += fread(sys->h_Jpy + offset, sizeof(MSC), sys->V/2, jdump);
    chkcp += fread(sys->h_Jpz + offset, sizeof(MSC), sys->V/2, jdump);
    chkcp += fread(sys->h_Jmx + offset, sizeof(MSC), sys->V/2, jdump);
    chkcp += fread(sys->h_Jmy + offset, sizeof(MSC), sys->V/2, jdump);
    chkcp += fread(sys->h_Jmz + offset, sizeof(MSC), sys->V/2, jdump);
  }

  fclose(jdump);
  if(chkcp^(sys->NJ*6)){
    fprintf(stderr, "\nError Copying couplings: wrong size\n");
    return 1;
  }

  return 0;
}

extern int JReadSys(sysEA3D_t *sys, int sysIndex){
  FILE *jdump = fopen("j_dump_start", "rb");
  if(jdump == NULL){
    fprintf(stderr, "\nError opening j_dump_start\n");
    return 1;
  }
  int chkcp = 0;

  int offset = sysIndex*(sys->V/2);
  chkcp += fread(sys->h_Jpx + offset, sizeof(MSC), sys->V/2, jdump);
  chkcp += fread(sys->h_Jpy + offset, sizeof(MSC), sys->V/2, jdump);
  chkcp += fread(sys->h_Jpz + offset, sizeof(MSC), sys->V/2, jdump);
  chkcp += fread(sys->h_Jmx + offset, sizeof(MSC), sys->V/2, jdump);
  chkcp += fread(sys->h_Jmy + offset, sizeof(MSC), sys->V/2, jdump);
  chkcp += fread(sys->h_Jmz + offset, sizeof(MSC), sys->V/2, jdump);

  fclose(jdump);
  if(chkcp^((sys->V/2)*6)){
    fprintf(stderr, "\nError Copying couplings: wrong size\n");
    return 1;
  }

  return 0;
}

extern int SpinsRead(sysEA3D_t *sys, int t){
  char spinFileName[200];
  snprintf(spinFileName, sizeof(spinFileName), "spin_dump_L%dT%dtime%d",sys->L,(int)(sys->beta*100+.001), t);
  FILE *spindump = fopen(spinFileName, "rb");
  if(spindump == NULL){
    fprintf(stderr, "\nError opening %s\n", spinFileName);
    return 1;
  }

  int chkcp = 0;
  for(int i=0; i<sys->sysN; i++){
    int offset = i*(sys->V/2)*4;

    chkcp += fread(sys->h_reds0 + offset, sizeof(MSC), sys->V/2, spindump);
    chkcp += fread(sys->h_blues0 + offset, sizeof(MSC), sys->V/2, spindump);
    chkcp += fread(sys->h_reds1 + offset, sizeof(MSC), sys->V/2, spindump);
    chkcp += fread(sys->h_blues1 + offset, sizeof(MSC), sys->V/2, spindump);
    chkcp += fread(sys->h_reds2 + offset, sizeof(MSC), sys->V/2, spindump);
    chkcp += fread(sys->h_blues2 + offset, sizeof(MSC), sys->V/2, spindump);
    chkcp += fread(sys->h_reds3 + offset, sizeof(MSC), sys->V/2, spindump);
    chkcp += fread(sys->h_blues3 + offset, sizeof(MSC), sys->V/2, spindump);      
  }
  fclose(spindump);

  if(chkcp^(sys->NS*2*4)){
    fprintf(stderr, "\nError Copying spins: wrong size\n");
    return 1;
  }
  return 0;
}

extern int SpinsReadSys(sysEA3D_t *sys, int t, int sysIndex){
  char spinFileName[200];
  snprintf(spinFileName, sizeof(spinFileName), "spin_dump_L%dT%dtime%d",sys->L,(int)(sys->beta*100+.001), t);
  FILE *spindump = fopen(spinFileName, "rb");
  if(spindump == NULL){
    fprintf(stderr, "\nError opening %s\n", spinFileName);
    return 1;
  }

  int chkcp = 0;
  int offset = sysIndex*(sys->V/2);

  chkcp += fread(sys->h_reds0 + offset, sizeof(MSC), sys->V/2, spindump);
  chkcp += fread(sys->h_blues0 + offset, sizeof(MSC), sys->V/2, spindump);
  chkcp += fread(sys->h_reds1 + offset, sizeof(MSC), sys->V/2, spindump);
  chkcp += fread(sys->h_blues1 + offset, sizeof(MSC), sys->V/2, spindump);
  chkcp += fread(sys->h_reds2 + offset, sizeof(MSC), sys->V/2, spindump);
  chkcp += fread(sys->h_blues2 + offset, sizeof(MSC), sys->V/2, spindump);
  chkcp += fread(sys->h_reds3 + offset, sizeof(MSC), sys->V/2, spindump);
  chkcp += fread(sys->h_blues3 + offset, sizeof(MSC), sys->V/2, spindump);      
  
  fclose(spindump);

  if(chkcp^((sys->V/2)*4*2)){
    fprintf(stderr, "\nError Copying spins: wrong size\n");
    return 1;
  }
  return 0;
}

extern int SpinsTransform(sysEA3D_t *sys){
  int hV = sys->V/2;
  for(int i=0; i<sys->sysN; i++){
    int offset = i*hV;

    arrayTransformMSC(sys->h_reds0 + offset, sys->realI_R, hV);
    arrayTransformMSC(sys->h_reds1 + offset, sys->realI_R, hV);
    arrayTransformMSC(sys->h_reds2 + offset, sys->realI_R, hV);
    arrayTransformMSC(sys->h_reds3 + offset, sys->realI_R, hV);
    
    arrayTransformMSC(sys->h_blues0 + offset, sys->realI_B, hV);
    arrayTransformMSC(sys->h_blues1 + offset, sys->realI_B, hV);
    arrayTransformMSC(sys->h_blues2 + offset, sys->realI_B, hV);
    arrayTransformMSC(sys->h_blues3 + offset, sys->realI_B, hV);

  }

  return 0;
}

extern int SpinsTransformRead(sysEA3D_t *sys){
  int hV = sys->V/2;
  for(int i=0; i<sys->sysN; i++){
    int offset = i*hV;

    arrayTransformMSC(sys->h_reds0 + offset, sys->myI_R, hV);
    arrayTransformMSC(sys->h_reds1 + offset, sys->myI_R, hV);
    arrayTransformMSC(sys->h_reds2 + offset, sys->myI_R, hV);
    arrayTransformMSC(sys->h_reds3 + offset, sys->myI_R, hV);
    
    arrayTransformMSC(sys->h_blues0 + offset, sys->myI_B, hV);
    arrayTransformMSC(sys->h_blues1 + offset, sys->myI_B, hV);
    arrayTransformMSC(sys->h_blues2 + offset, sys->myI_B, hV);
    arrayTransformMSC(sys->h_blues3 + offset, sys->myI_B, hV);

  }

  return 0;
}

extern int SpinsDump(sysEA3D_t *sys, unsigned long long int t){
  char spinFileName[200];

  int chkwr = 0;
  for(int i=0; i<sys->sysN; i++){
    int offset = i*(sys->V/2);

    if(sys->recoverF == 0) snprintf(spinFileName, sizeof(spinFileName), "./dummyRun%d/spin_dump_L%dT%dtime%llu", 
				    i, sys->L,(int)(sys->beta*100+.001), t);
    if(sys->recoverF == 1) snprintf(spinFileName, sizeof(spinFileName), "./recover%016llX/dummyRun%d/spin_dump_L%dT%dtime%llu", 
				    sys->startValue, i, sys->L,(int)(sys->beta*100+.001), t);

    FILE *out = fopen(spinFileName, "wb");
    if(out == NULL){
      fprintf(stderr, "\nError opening pointer for file %s\n", spinFileName);
      return 1;
    }

    chkwr += fwrite(sys->h_reds0 + offset, sizeof(MSC), sys->V/2, out);
    chkwr += fwrite(sys->h_blues0 + offset, sizeof(MSC), sys->V/2, out);
    chkwr += fwrite(sys->h_reds1 + offset, sizeof(MSC), sys->V/2, out);
    chkwr += fwrite(sys->h_blues1 + offset, sizeof(MSC), sys->V/2, out);
    chkwr += fwrite(sys->h_reds2 + offset, sizeof(MSC), sys->V/2, out);
    chkwr += fwrite(sys->h_blues2 + offset, sizeof(MSC), sys->V/2, out);
    chkwr += fwrite(sys->h_reds3 + offset, sizeof(MSC), sys->V/2, out);
    chkwr += fwrite(sys->h_blues3 + offset, sizeof(MSC), sys->V/2, out);

    fflush(out);
    fclose(out);

  }

  if(chkwr^(sys->NS*2*sys->replicas)){
    fprintf(stderr, "\nError writing spins SpinsDump\n");
    return 1;
  }
  return 0;
}

extern int SpinsReadDir(sysEA3D_t *sys, unsigned long long int t){
  char spinFileName[200];

  int chkwr = 0;
  for(int i=0; i<sys->sysN; i++){
    int offset = i*(sys->V/2);

    if(sys->recoverF == 1){
      snprintf(spinFileName, sizeof(spinFileName), "./run%d/spin_dump_L%dT%dtime%llu", 
	       i, sys->L,(int)(sys->beta*100+.001), t);
    }

    FILE *out = fopen(spinFileName, "rb");
    if(out == NULL){
      fprintf(stderr, "\nError opening pointer for file %s\n", spinFileName);
      return 1;
    }

    chkwr += fread(sys->h_reds0 + offset, sizeof(MSC), sys->V/2, out);
    chkwr += fread(sys->h_blues0 + offset, sizeof(MSC), sys->V/2, out);
    chkwr += fread(sys->h_reds1 + offset, sizeof(MSC), sys->V/2, out);
    chkwr += fread(sys->h_blues1 + offset, sizeof(MSC), sys->V/2, out);
    chkwr += fread(sys->h_reds2 + offset, sizeof(MSC), sys->V/2, out);
    chkwr += fread(sys->h_blues2 + offset, sizeof(MSC), sys->V/2, out);
    chkwr += fread(sys->h_reds3 + offset, sizeof(MSC), sys->V/2, out);
    chkwr += fread(sys->h_blues3 + offset, sizeof(MSC), sys->V/2, out);
    
    fflush(out);
    fclose(out);

  }

  if(chkwr^(sys->NS*2*sys->replicas)){
    fprintf(stderr, "\nError reading spins SpinsReadDir %d %d\n", chkwr, sys->NS*2*4);
    return 1;
  }
  return 0;
}

extern int JTransform(sysEA3D_t *sys){
  int hV = sys->V/2;
  for(int i=0; i<sys->sysN; i++){
    int offset = i*hV;      
    arrayTransformMSC(sys->h_Jpx + offset, sys->realI_R, hV);
    arrayTransformMSC(sys->h_Jpy + offset, sys->realI_R, hV);
    arrayTransformMSC(sys->h_Jpz + offset, sys->realI_R, hV);
    arrayTransformMSC(sys->h_Jmx + offset, sys->realI_R, hV);
    arrayTransformMSC(sys->h_Jmy + offset, sys->realI_R, hV);
    arrayTransformMSC(sys->h_Jmz + offset, sys->realI_R, hV);
  }      
  return 0;
}

extern int JTransformRead(sysEA3D_t *sys){
  int hV = sys->V/2;
  for(int i=0; i<sys->sysN; i++){
    int offset = i*hV;      
    arrayTransformMSC(sys->h_Jpx + offset, sys->myI_R, hV);
    arrayTransformMSC(sys->h_Jpy + offset, sys->myI_R, hV);
    arrayTransformMSC(sys->h_Jpz + offset, sys->myI_R, hV);
    arrayTransformMSC(sys->h_Jmx + offset, sys->myI_R, hV);
    arrayTransformMSC(sys->h_Jmy + offset, sys->myI_R, hV);
    arrayTransformMSC(sys->h_Jmz + offset, sys->myI_R, hV);
  }      
  return 0;
}

extern int JDump(sysEA3D_t *sys){
  char JFileName[200];

  int chkwr = 0;
  for(int i=0; i<sys->sysN; i++){
    int offset = i*(sys->V/2);

    if(sys->recoverF == 0) snprintf(JFileName, sizeof(JFileName), "./dummyRun%d/j_dump_start", i);
    if(sys->recoverF == 1) snprintf(JFileName, sizeof(JFileName), "./recover%016llX/dummyRun%d/j_dump_start", 
				    sys->startValue, i);    

    FILE *out = fopen(JFileName, "wb");
    if(out == NULL){
      fprintf(stderr, "\nError opening pointer for file %s\n", JFileName);
      return 1;
    }

    chkwr += fwrite(sys->h_Jpx + offset, sizeof(MSC), sys->V/2, out);
    chkwr += fwrite(sys->h_Jpy + offset, sizeof(MSC), sys->V/2, out);
    chkwr += fwrite(sys->h_Jpz + offset, sizeof(MSC), sys->V/2, out);
    chkwr += fwrite(sys->h_Jmx + offset, sizeof(MSC), sys->V/2, out);
    chkwr += fwrite(sys->h_Jmy + offset, sizeof(MSC), sys->V/2, out);
    chkwr += fwrite(sys->h_Jmz + offset, sizeof(MSC), sys->V/2, out);

    fflush(out);
    fclose(out);

  }

  if(chkwr^(sys->NJ*6)){
    fprintf(stderr, "\nError writing j_dump_start\n");
    return 1;
  }

  return 0;
} 

extern int JReadDir(sysEA3D_t *sys){
  char JFileName[41];

  int chkwr = 0;
  for(int i=0; i<sys->sysN; i++){
    int offset = i*(sys->V/2);

    snprintf(JFileName, sizeof(JFileName), "./run%d/j_dump_start", i);
    FILE *out = fopen(JFileName, "rb");
    if(out == NULL){
      fprintf(stderr, "\nError opening pointer for file %s\n", JFileName);
      return 1;
    }

    chkwr += fread(sys->h_Jpx + offset, sizeof(MSC), sys->V/2, out);
    chkwr += fread(sys->h_Jpy + offset, sizeof(MSC), sys->V/2, out);
    chkwr += fread(sys->h_Jpz + offset, sizeof(MSC), sys->V/2, out);
    chkwr += fread(sys->h_Jmx + offset, sizeof(MSC), sys->V/2, out);
    chkwr += fread(sys->h_Jmy + offset, sizeof(MSC), sys->V/2, out);
    chkwr += fread(sys->h_Jmz + offset, sizeof(MSC), sys->V/2, out);

    fflush(out);
    fclose(out);

  }

  if(chkwr^(sys->NJ*6)){
    fprintf(stderr, "\nError writing j_dump_start\n");
    return 1;
  }

  return 0;
}

extern sysEA3D_t *sysInit(int argc, char **argv){
  if(argc != 15){ usage(argv); return NULL; }
  sysEA3D_t *ptr = (sysEA3D_t *)malloc(sizeof(sysEA3D_t));
  if(ptr == NULL){
    fprintf(stderr, "\nNo Alloc ptr in sysInit\n");
    return NULL;
  }
  sscanf(argv[1], "%d", &ptr->L);
  sscanf(argv[2], "%f", &ptr->beta);
  sscanf(argv[3], "%d", &ptr->steps);
  sscanf(argv[4], "%d", &ptr->replicas);
  sscanf(argv[5], "%d", &ptr->sysN);
  sscanf(argv[6], "%d", &ptr->blockSize);
  sscanf(argv[7], "%d", &ptr->k);
  sscanf(argv[8], "%d", &ptr->kernelChoice);
  sscanf(argv[9], "%d", &ptr->cardN);
  sscanf(argv[10], "%d", &ptr->dumpF);
  sscanf(argv[11], "%d", &ptr->analysisF);
  sscanf(argv[12], "%d", &ptr->ram);
  sscanf(argv[13], "%d", &ptr->recoverF);
  sscanf(argv[14], "%llu", &ptr->startValue);

  // GPU selection
  int numberOfGPUs = 0;
  cudaGetDeviceCount(&numberOfGPUs);
  if(ptr->cardN >= numberOfGPUs){
    fprintf(stderr, "gpu %d >= gpuN %d\n", ptr->cardN, numberOfGPUs);
    return NULL;
  }
  cudaSetDevice(ptr->cardN);
  cudaDeviceProp prop;
  cudaGetDeviceProperties(&prop, ptr->cardN);
  printf("GPU %d: %s\n", ptr->cardN, prop.name);

  // ram option
  if(ptr->ram == 1 && ptr->analysisF == 1){
    snprintf(ptr->devShm, sizeof(ptr->devShm), "/dev/shm/dev%dL%dB%.3f/",
	     ptr->cardN, ptr->L, ptr->beta);
  }else{
    snprintf(ptr->devShm, sizeof(ptr->devShm), "./dev%dL%dB%.3f/",
	     ptr->cardN, ptr->L, ptr->beta);
  }

  // recover
  if(ptr->recoverF == 0) ptr->startValue = 0;

  ptr->V = ptr->L*ptr->L*ptr->L;
  ptr->NJ = ptr->sysN*ptr->V/2;
  ptr->NS = ptr->sysN*ptr->V/2;

  int hV = ptr->V/2;
  ptr->realI_R = (int *)malloc(hV*sizeof(int));
  ptr->realI_B = (int *)malloc(hV*sizeof(int));
  ptr->myI_R = (int *)malloc(hV*sizeof(int));
  ptr->myI_B = (int *)malloc(hV*sizeof(int));
  ptr->myI_B2R = (int *)malloc(hV*sizeof(int));
  ptr->myI_R2B = (int *)malloc(hV*sizeof(int)); 

  mapsDefine(ptr->realI_R, ptr->realI_B, ptr->myI_R, ptr->myI_B, ptr->myI_B2R, ptr->myI_R2B, ptr->L); 

  if(masksInit(ptr)){
    fprintf(stderr,"\nError in masksInit\n");
    return NULL;
  }
  if(MCWeightsInit(ptr)){
    fprintf(stderr,"\nError in MCWeightsInit\n");
    return NULL;
  }
  if(dMemInit(ptr)){
    fprintf(stderr,"\nError in dMemInit\n");
    return NULL;
  }
  if(hMemInit(ptr)){
    fprintf(stderr,"\nError in hMemInit\n");
    return NULL;
  }

  if(ptr->recoverF == 0){
    if(JInit(ptr)){
      fprintf(stderr,"\nError in JInit\n");
      return NULL;
    }

    if(SpinsInit(ptr)){
      fprintf(stderr,"\nError in SpinsInit\n");
      return NULL;
    }
    if(H2DSysCopy(ptr)){
      fprintf(stderr,"\nError in H2DSysCopy\n");
      return NULL;
    }
  }

  if(ptr->recoverF == 1){
    if(SpinsReadDir(ptr, ptr->startValue)){
      fprintf(stderr,"\nError in SpinsReadDir\n");
      return NULL;
    }

    if(JReadDir(ptr)){
      fprintf(stderr,"\nError in JRead\n");
      return NULL;
    }

    if(ptr->kernelChoice == 1 || ptr->kernelChoice == 2){
      SpinsTransformRead(ptr);
      JTransformRead(ptr);
    }

    if(H2DSysCopy(ptr)){
      fprintf(stderr,"\nError in H2DSysCopy\n");
      return NULL;
    }
  }


  if(ptr->dumpF == 1 || ptr->ram == 1){

    if(dummyDirsMaker(ptr)){
      fprintf(stderr,"\nError in dummyDirsMaker\n");
      return NULL;
    }

  }

  if(ptr->kernelChoice == 1 || ptr->kernelChoice == 2){
    SpinsTransform(ptr);
    JTransform(ptr);
  }

  if(ptr->dumpF == 1){
    if(SpinsDump(ptr, ptr->startValue)){
      fprintf(stderr,"\nError in SpinsDump\n");
      return NULL;
    }

    if(JDump(ptr)){
      fprintf(stderr,"\nError in JDump\n");
      return NULL;
    }
  }

  //if(ptr->recoverF == 1) return NULL;
  return ptr;
  //return NULL;
  // fin qui funziona il recovery con le repliche
}

// BEGINNING KERNELS

//////////////////////////////////////////////////////////////////////
// KERNEL BITWISE
//////////////////////////////////////////////////////////////////////

__global__ void MCStepReds_Bitwise(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *rand){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = kk + blockIdx.y*d_loopStride;
  RNGT seed = rand[rngIndex];

  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    unsigned int iz=k>>d_sdbLsm1;
    unsigned int iy=(k - (iz<<d_sdbLsm1))>>d_sdbLm1;
    int diff=(iz&1)^(iy&1);
    
    int spx=(k >> (-diff)) + (((k-(k&d_maskL)) + ((k+1)&d_maskL)) >> (diff-1)) + off;
    int smx=(((k-(k&d_maskL)) + ((k-1)&d_maskL)) >> (-diff)) + (k>>(diff-1)) + off;
    int spy=(k-(k&d_maskL2))+((k+d_Ldb2)&d_maskL2) + off;
    int smy=(k-(k&d_maskL2))+((k-d_Ldb2)&d_maskL2) + off;
    int spz=((k+d_Lsdb2) & d_maskL3) + off;
    int smz=((k-d_Lsdb2) & d_maskL3) + off;
    
    UPDATE_RED(s0, texture_B0);
    UPDATE_RED(s1, texture_B1);
    UPDATE_RED(s2, texture_B2);
    UPDATE_RED(s3, texture_B3);
  }
  rand[rngIndex] = seed;

  return;
}

__global__ void MCStepReds_Bitwise_PR(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *ira, int ip0, int ip1, int ip2, int ip3, int spins){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = kk + blockIdx.y*d_loopStride;

  int count = 0;
  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    unsigned int iz=k>>d_sdbLsm1;
    unsigned int iy=(k - (iz<<d_sdbLsm1))>>d_sdbLm1;
    int diff=(iz&1)^(iy&1);
    
    int spx=(k >> (-diff)) + (((k-(k&d_maskL)) + ((k+1)&d_maskL)) >> (diff-1)) + off;
    int smx=(((k-(k&d_maskL)) + ((k-1)&d_maskL)) >> (-diff)) + (k>>(diff-1)) + off;
    int spy=(k-(k&d_maskL2))+((k+d_Ldb2)&d_maskL2) + off;
    int smy=(k-(k&d_maskL2))+((k-d_Ldb2)&d_maskL2) + off;
    int spz=((k+d_Lsdb2) & d_maskL3) + off;
    int smz=((k-d_Lsdb2) & d_maskL3) + off;

    RNGT seed = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];
    ira[ip0*d_threads + rngIndex] = seed;
    seed ^= ira[ip3*d_threads + rngIndex];
    
    ip0 = ((ip0 + 1)&PRWHEELMASK);
    ip1 = ((ip1 + 1)&PRWHEELMASK);
    ip2 = ((ip2 + 1)&PRWHEELMASK);
    ip3 = ((ip3 + 1)&PRWHEELMASK);
    
    UPDATE_RED_PRNG(s0, texture_B0);

    seed = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];
    ira[ip0*d_threads + rngIndex] = seed;
    seed ^= ira[ip3*d_threads + rngIndex];   

    ip0 = ((ip0 + 1)&PRWHEELMASK);
    ip1 = ((ip1 + 1)&PRWHEELMASK);
    ip2 = ((ip2 + 1)&PRWHEELMASK);
    ip3 = ((ip3 + 1)&PRWHEELMASK);

    UPDATE_RED_PRNG(s1, texture_B1);

    seed = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];
    ira[ip0*d_threads + rngIndex] = seed;
    seed ^= ira[ip3*d_threads + rngIndex];   

    ip0 = ((ip0 + 1)&PRWHEELMASK);
    ip1 = ((ip1 + 1)&PRWHEELMASK);
    ip2 = ((ip2 + 1)&PRWHEELMASK);
    ip3 = ((ip3 + 1)&PRWHEELMASK);

    UPDATE_RED_PRNG(s2, texture_B2);

    seed = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];
    ira[ip0*d_threads + rngIndex] = seed;
    seed ^= ira[ip3*d_threads + rngIndex];   

    ip0 = ((ip0 + 1)&PRWHEELMASK);
    ip1 = ((ip1 + 1)&PRWHEELMASK);
    ip2 = ((ip2 + 1)&PRWHEELMASK);
    ip3 = ((ip3 + 1)&PRWHEELMASK);

    UPDATE_RED_PRNG(s3, texture_B3);

    count += 4;
  }

  if(count < spins && count > 0){
    for(int i = count; i < spins; i++){
      ira[ip0*d_threads + rngIndex] = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];

      ip0 = ((ip0 + 1)&PRWHEELMASK);
      ip1 = ((ip1 + 1)&PRWHEELMASK);
      ip2 = ((ip2 + 1)&PRWHEELMASK);
      ip3 = ((ip3 + 1)&PRWHEELMASK);
    }
  }

  return;
}

__global__ void MCStepReds_Bitwise_MT19937(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *rand, int k0, int k1, int k2, int spins){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = kk + blockIdx.y*d_loopStride;

  int count = 0;
  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    unsigned int iz=k>>d_sdbLsm1;
    unsigned int iy=(k - (iz<<d_sdbLsm1))>>d_sdbLm1;
    int diff=(iz&1)^(iy&1);
    
    int spx=(k >> (-diff)) + (((k-(k&d_maskL)) + ((k+1)&d_maskL)) >> (diff-1)) + off;
    int smx=(((k-(k&d_maskL)) + ((k-1)&d_maskL)) >> (-diff)) + (k>>(diff-1)) + off;
    int spy=(k-(k&d_maskL2))+((k+d_Ldb2)&d_maskL2) + off;
    int smy=(k-(k&d_maskL2))+((k-d_Ldb2)&d_maskL2) + off;
    int spz=((k+d_Lsdb2) & d_maskL3) + off;
    int smz=((k-d_Lsdb2) & d_maskL3) + off;

    RNGT seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK); 
    seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
    rand[k0*d_threads + rngIndex] = seed;
    
    k0 = SP(k0 + 1, N_MT);
    k1 = SP(k1 + 1, N_MT);
    k2 = SP(k2 + 1, N_MT);
    
    seed ^= (seed >> 11);
    seed ^= (seed << 7) & 0x9d2c5680U;
    seed ^= (seed << 15) & 0xefc60000U;
    seed ^= (seed >> 18);
    
    UPDATE_RED_PRNG(s0, texture_B0);

    seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK); 
    seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
    rand[k0*d_threads + rngIndex] = seed;
    
    k0 = SP(k0 + 1, N_MT);
    k1 = SP(k1 + 1, N_MT);
    k2 = SP(k2 + 1, N_MT);
    
    seed ^= (seed >> 11);
    seed ^= (seed << 7) & 0x9d2c5680U;
    seed ^= (seed << 15) & 0xefc60000U;
    seed ^= (seed >> 18);

    UPDATE_RED_PRNG(s1, texture_B1);

    seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK); 
    seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
    rand[k0*d_threads + rngIndex] = seed;
    
    k0 = SP(k0 + 1, N_MT);
    k1 = SP(k1 + 1, N_MT);
    k2 = SP(k2 + 1, N_MT);
    
    seed ^= (seed >> 11);
    seed ^= (seed << 7) & 0x9d2c5680U;
    seed ^= (seed << 15) & 0xefc60000U;
    seed ^= (seed >> 18);

    UPDATE_RED_PRNG(s2, texture_B2);

    seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK); 
    seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
    rand[k0*d_threads + rngIndex] = seed;
    
    k0 = SP(k0 + 1, N_MT);
    k1 = SP(k1 + 1, N_MT);
    k2 = SP(k2 + 1, N_MT);
    
    seed ^= (seed >> 11);
    seed ^= (seed << 7) & 0x9d2c5680U;
    seed ^= (seed << 15) & 0xefc60000U;
    seed ^= (seed >> 18);

    UPDATE_RED_PRNG(s3, texture_B3);

    count += 4;
  }

  if(count < spins && count > 0){
    for(int i = count; i < spins; i++){
      RNGT seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK);
      seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
      rand[k0*d_threads + rngIndex] = seed;
      
      k0 = SP(k0 + 1, N_MT);
      k1 = SP(k1 + 1, N_MT);
      k2 = SP(k2 + 1, N_MT);      
    }
  }

  return;
}


__global__ void MCStepReds_Bitwise_MTGP32(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, curandStateMtgp32 *state, int spins){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = blockIdx.x + gridDim.x*blockIdx.y;

  int count = 0;
  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    unsigned int iz=k>>d_sdbLsm1;
    unsigned int iy=(k - (iz<<d_sdbLsm1))>>d_sdbLm1;
    int diff=(iz&1)^(iy&1);
    
    int spx=(k >> (-diff)) + (((k-(k&d_maskL)) + ((k+1)&d_maskL)) >> (diff-1)) + off;
    int smx=(((k-(k&d_maskL)) + ((k-1)&d_maskL)) >> (-diff)) + (k>>(diff-1)) + off;
    int spy=(k-(k&d_maskL2))+((k+d_Ldb2)&d_maskL2) + off;
    int smy=(k-(k&d_maskL2))+((k-d_Ldb2)&d_maskL2) + off;
    int spz=((k+d_Lsdb2) & d_maskL3) + off;
    int smz=((k-d_Lsdb2) & d_maskL3) + off;

    RNGT seed = curand(&state[rngIndex]);
    
    UPDATE_RED_PRNG(s0, texture_B0);

    seed = curand(&state[rngIndex]);

    UPDATE_RED_PRNG(s1, texture_B1);

    seed = curand(&state[rngIndex]);

    UPDATE_RED_PRNG(s2, texture_B2);

    seed = curand(&state[rngIndex]);

    UPDATE_RED_PRNG(s3, texture_B3);

  }

  if(count < spins && count > 0){
    for(int i = count; i < spins; i++){
      curand(&state[rngIndex]);
    }
  }

  return;
}


__global__ void MCStepReds_Bitwise_XORWOW(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, curandState *state){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = kk + blockIdx.y*d_loopStride;
  curandState localPrng = state[rngIndex];

  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    unsigned int iz=k>>d_sdbLsm1;
    unsigned int iy=(k - (iz<<d_sdbLsm1))>>d_sdbLm1;
    int diff=(iz&1)^(iy&1);
    
    int spx=(k >> (-diff)) + (((k-(k&d_maskL)) + ((k+1)&d_maskL)) >> (diff-1)) + off;
    int smx=(((k-(k&d_maskL)) + ((k-1)&d_maskL)) >> (-diff)) + (k>>(diff-1)) + off;
    int spy=(k-(k&d_maskL2))+((k+d_Ldb2)&d_maskL2) + off;
    int smy=(k-(k&d_maskL2))+((k-d_Ldb2)&d_maskL2) + off;
    int spz=((k+d_Lsdb2) & d_maskL3) + off;
    int smz=((k-d_Lsdb2) & d_maskL3) + off;

    RNGT seed = curand(&localPrng);
    
    UPDATE_RED_PRNG(s0, texture_B0);

    seed = curand(&localPrng);

    UPDATE_RED_PRNG(s1, texture_B1);

    seed = curand(&localPrng);

    UPDATE_RED_PRNG(s2, texture_B2);

    seed = curand(&localPrng);

    UPDATE_RED_PRNG(s3, texture_B3);

  }

  state[rngIndex] = localPrng;
  return;
}

__global__ void MCStepBlues_Bitwise(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *rand){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = kk + blockIdx.y*d_loopStride;
  RNGT seed = rand[rngIndex];

  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    unsigned int iz=k>>d_sdbLsm1;
    unsigned int iy=(k - (iz<<d_sdbLsm1))>>d_sdbLm1;
    int diff=(iz&1)^(iy&1);
    
    int spx=(k >> (diff-1)) + (((k-(k&d_maskL)) + ((k+1)&d_maskL)) >> (-diff)) + off;
    int smx=(((k-(k&d_maskL)) + ((k-1)&d_maskL)) >> (diff-1)) + (k>>(-diff)) + off;
    int spy=(k-(k&d_maskL2))+((k+d_Ldb2)&d_maskL2) + off;
    int smy=(k-(k&d_maskL2))+((k-d_Ldb2)&d_maskL2) + off;
    int spz=((k+d_Lsdb2) & d_maskL3) + off;
    int smz=((k-d_Lsdb2) & d_maskL3) + off;
    
    UPDATE_BLUE(s0, texture_R0);
    UPDATE_BLUE(s1, texture_R1);
    UPDATE_BLUE(s2, texture_R2);
    UPDATE_BLUE(s3, texture_R3);
  }
  rand[rngIndex] = seed;

  return;
}

__global__ void MCStepBlues_Bitwise_PR(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *ira, int ip0, int ip1, int ip2, int ip3, int spins){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = kk + blockIdx.y*d_loopStride;

  int count = 0;
  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    unsigned int iz=k>>d_sdbLsm1;
    unsigned int iy=(k - (iz<<d_sdbLsm1))>>d_sdbLm1;
    int diff=(iz&1)^(iy&1);
    
    int spx=(k >> (diff-1)) + (((k-(k&d_maskL)) + ((k+1)&d_maskL)) >> (-diff)) + off;
    int smx=(((k-(k&d_maskL)) + ((k-1)&d_maskL)) >> (diff-1)) + (k>>(-diff)) + off;
    int spy=(k-(k&d_maskL2))+((k+d_Ldb2)&d_maskL2) + off;
    int smy=(k-(k&d_maskL2))+((k-d_Ldb2)&d_maskL2) + off;
    int spz=((k+d_Lsdb2) & d_maskL3) + off;
    int smz=((k-d_Lsdb2) & d_maskL3) + off;

    RNGT seed = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];
    ira[ip0*d_threads + rngIndex] = seed;
    seed ^= ira[ip3*d_threads + rngIndex];
    
    ip0 = ((ip0 + 1)&PRWHEELMASK);
    ip1 = ((ip1 + 1)&PRWHEELMASK);
    ip2 = ((ip2 + 1)&PRWHEELMASK);
    ip3 = ((ip3 + 1)&PRWHEELMASK);
    
    UPDATE_BLUE_PRNG(s0, texture_R0);

    seed = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];
    ira[ip0*d_threads + rngIndex] = seed;
    seed ^= ira[ip3*d_threads + rngIndex];   

    ip0 = ((ip0 + 1)&PRWHEELMASK);
    ip1 = ((ip1 + 1)&PRWHEELMASK);
    ip2 = ((ip2 + 1)&PRWHEELMASK);
    ip3 = ((ip3 + 1)&PRWHEELMASK);

    UPDATE_BLUE_PRNG(s1, texture_R1);

    seed = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];
    ira[ip0*d_threads + rngIndex] = seed;
    seed ^= ira[ip3*d_threads + rngIndex];   

    ip0 = ((ip0 + 1)&PRWHEELMASK);
    ip1 = ((ip1 + 1)&PRWHEELMASK);
    ip2 = ((ip2 + 1)&PRWHEELMASK);
    ip3 = ((ip3 + 1)&PRWHEELMASK);

    UPDATE_BLUE_PRNG(s2, texture_R2);

    seed = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];
    ira[ip0*d_threads + rngIndex] = seed;
    seed ^= ira[ip3*d_threads + rngIndex];   

    ip0 = ((ip0 + 1)&PRWHEELMASK);
    ip1 = ((ip1 + 1)&PRWHEELMASK);
    ip2 = ((ip2 + 1)&PRWHEELMASK);
    ip3 = ((ip3 + 1)&PRWHEELMASK);

    UPDATE_BLUE_PRNG(s3, texture_R3);

    count += 4;
  }

  if(count < spins && count > 0){
    for(int i = count; i < spins; i++){
      ira[ip0*d_threads + rngIndex] = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];

      ip0 = ((ip0 + 1)&PRWHEELMASK);
      ip1 = ((ip1 + 1)&PRWHEELMASK);
      ip2 = ((ip2 + 1)&PRWHEELMASK);
      ip3 = ((ip3 + 1)&PRWHEELMASK);
    }
  }

  return;
}

__global__ void MCStepBlues_Bitwise_MT19937(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *rand, int k0, int k1, int k2, int spins){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = kk + blockIdx.y*d_loopStride;

  int count = 0;
  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    unsigned int iz=k>>d_sdbLsm1;
    unsigned int iy=(k - (iz<<d_sdbLsm1))>>d_sdbLm1;
    int diff=(iz&1)^(iy&1);
    
    int spx=(k >> (diff-1)) + (((k-(k&d_maskL)) + ((k+1)&d_maskL)) >> (-diff)) + off;
    int smx=(((k-(k&d_maskL)) + ((k-1)&d_maskL)) >> (diff-1)) + (k>>(-diff)) + off;
    int spy=(k-(k&d_maskL2))+((k+d_Ldb2)&d_maskL2) + off;
    int smy=(k-(k&d_maskL2))+((k-d_Ldb2)&d_maskL2) + off;
    int spz=((k+d_Lsdb2) & d_maskL3) + off;
    int smz=((k-d_Lsdb2) & d_maskL3) + off;

    RNGT seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK); 
    seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
    rand[k0*d_threads + rngIndex] = seed;
    
    k0 = SP(k0 + 1, N_MT);
    k1 = SP(k1 + 1, N_MT);
    k2 = SP(k2 + 1, N_MT);
    
    seed ^= (seed >> 11);
    seed ^= (seed << 7) & 0x9d2c5680U;
    seed ^= (seed << 15) & 0xefc60000U;
    seed ^= (seed >> 18);
    
    UPDATE_BLUE_PRNG(s0, texture_R0);

    seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK); 
    seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
    rand[k0*d_threads + rngIndex] = seed;
    
    k0 = SP(k0 + 1, N_MT);
    k1 = SP(k1 + 1, N_MT);
    k2 = SP(k2 + 1, N_MT);
    
    seed ^= (seed >> 11);
    seed ^= (seed << 7) & 0x9d2c5680U;
    seed ^= (seed << 15) & 0xefc60000U;
    seed ^= (seed >> 18);

    UPDATE_BLUE_PRNG(s1, texture_R1);

    seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK); 
    seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
    rand[k0*d_threads + rngIndex] = seed;
    
    k0 = SP(k0 + 1, N_MT);
    k1 = SP(k1 + 1, N_MT);
    k2 = SP(k2 + 1, N_MT);
    
    seed ^= (seed >> 11);
    seed ^= (seed << 7) & 0x9d2c5680U;
    seed ^= (seed << 15) & 0xefc60000U;
    seed ^= (seed >> 18);

    UPDATE_BLUE_PRNG(s2, texture_R2);

    seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK); 
    seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
    rand[k0*d_threads + rngIndex] = seed;
    
    k0 = SP(k0 + 1, N_MT);
    k1 = SP(k1 + 1, N_MT);
    k2 = SP(k2 + 1, N_MT);
    
    seed ^= (seed >> 11);
    seed ^= (seed << 7) & 0x9d2c5680U;
    seed ^= (seed << 15) & 0xefc60000U;
    seed ^= (seed >> 18);

    UPDATE_BLUE_PRNG(s3, texture_R3);

    count += 4;
  }

  if(count < spins && count > 0){
    for(int i = count; i < spins; i++){
      RNGT seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK);
      seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
      rand[k0*d_threads + rngIndex] = seed;
      
      k0 = SP(k0 + 1, N_MT);
      k1 = SP(k1 + 1, N_MT);
      k2 = SP(k2 + 1, N_MT);      
    }
  }

  return;
}

__global__ void MCStepBlues_Bitwise_MTGP32(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, curandStateMtgp32 *state, int spins){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = blockIdx.x + gridDim.x*blockIdx.y;

  int count = 0;
  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    unsigned int iz=k>>d_sdbLsm1;
    unsigned int iy=(k - (iz<<d_sdbLsm1))>>d_sdbLm1;
    int diff=(iz&1)^(iy&1);
    
    int spx=(k >> (diff-1)) + (((k-(k&d_maskL)) + ((k+1)&d_maskL)) >> (-diff)) + off;
    int smx=(((k-(k&d_maskL)) + ((k-1)&d_maskL)) >> (diff-1)) + (k>>(-diff)) + off;
    int spy=(k-(k&d_maskL2))+((k+d_Ldb2)&d_maskL2) + off;
    int smy=(k-(k&d_maskL2))+((k-d_Ldb2)&d_maskL2) + off;
    int spz=((k+d_Lsdb2) & d_maskL3) + off;
    int smz=((k-d_Lsdb2) & d_maskL3) + off;

    RNGT seed = curand(&state[rngIndex]);
    
    UPDATE_BLUE_PRNG(s0, texture_R0);

    seed = curand(&state[rngIndex]);

    UPDATE_BLUE_PRNG(s1, texture_R1);

    seed = curand(&state[rngIndex]);

    UPDATE_BLUE_PRNG(s2, texture_R2);

    seed = curand(&state[rngIndex]);

    UPDATE_BLUE_PRNG(s3, texture_R3);

    count += 4;
  }

  if(count < spins && count > 0){
    for(int i = count; i < spins; i++){
      curand(&state[rngIndex]);
    }
  }

  return;
}

__global__ void MCStepBlues_Bitwise_XORWOW(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, curandState *state){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = kk + blockIdx.y*d_loopStride;
  curandState localPrng = state[rngIndex];

  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    unsigned int iz=k>>d_sdbLsm1;
    unsigned int iy=(k - (iz<<d_sdbLsm1))>>d_sdbLm1;
    int diff=(iz&1)^(iy&1);
    
    int spx=(k >> (diff-1)) + (((k-(k&d_maskL)) + ((k+1)&d_maskL)) >> (-diff)) + off;
    int smx=(((k-(k&d_maskL)) + ((k-1)&d_maskL)) >> (diff-1)) + (k>>(-diff)) + off;
    int spy=(k-(k&d_maskL2))+((k+d_Ldb2)&d_maskL2) + off;
    int smy=(k-(k&d_maskL2))+((k-d_Ldb2)&d_maskL2) + off;
    int spz=((k+d_Lsdb2) & d_maskL3) + off;
    int smz=((k-d_Lsdb2) & d_maskL3) + off;

    RNGT seed = curand(&localPrng);
    
    UPDATE_BLUE_PRNG(s0, texture_R0);

    seed = curand(&localPrng);

    UPDATE_BLUE_PRNG(s1, texture_R1);

    seed = curand(&localPrng);

    UPDATE_BLUE_PRNG(s2, texture_R2);

    seed = curand(&localPrng);

    UPDATE_BLUE_PRNG(s3, texture_R3);
  }

  state[rngIndex] = localPrng;
  return;
}

//////////////////////////////////////////////////////////////////////
// KERNEL STANDARD
//////////////////////////////////////////////////////////////////////


__global__ void MCStepReds_Standard(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *rand){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = kk + blockIdx.y*d_loopStride;
  RNGT seed = rand[rngIndex];

  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    int x = k%d_hL; int y = (k/d_hL)%d_L; int z = k/d_hA;
    int par = (y^z)&1;
    
    int spx = i - x + SP(x + 1 - (par^1), d_hL);
    int smx = i - x + SM(x - 1 + par, d_hL);
    int spy = i + (SP(y + 1, d_L) - y)*d_hL;
    int smy = i + (SM(y - 1, d_L) - y)*d_hL;
    int spz = i + (SP(z + 1, d_L) - z)*d_hA;
    int smz = i + (SM(z - 1, d_L) - z)*d_hA;
    
    UPDATE_RED(s0, texture_B0);
    UPDATE_RED(s1, texture_B1);
    UPDATE_RED(s2, texture_B2);
    UPDATE_RED(s3, texture_B3);
  }

  rand[rngIndex] = seed;

  return;
}

__global__ void MCStepReds_Standard_PR(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *ira, int ip0, int ip1, int ip2, int ip3, int spins){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = kk + blockIdx.y*d_loopStride;

  int count = 0;
  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    int x = k%d_hL; int y = (k/d_hL)%d_L; int z = k/d_hA;
    int par = (y^z)&1;
    
    int spx = i - x + SP(x + 1 - (par^1), d_hL);
    int smx = i - x + SM(x - 1 + par, d_hL);
    int spy = i + (SP(y + 1, d_L) - y)*d_hL;
    int smy = i + (SM(y - 1, d_L) - y)*d_hL;
    int spz = i + (SP(z + 1, d_L) - z)*d_hA;
    int smz = i + (SM(z - 1, d_L) - z)*d_hA;

    RNGT seed = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];
    ira[ip0*d_threads + rngIndex] = seed;
    seed ^= ira[ip3*d_threads + rngIndex];
    
    ip0 = ((ip0 + 1)&PRWHEELMASK);
    ip1 = ((ip1 + 1)&PRWHEELMASK);
    ip2 = ((ip2 + 1)&PRWHEELMASK);
    ip3 = ((ip3 + 1)&PRWHEELMASK);
    
    UPDATE_RED_PRNG(s0, texture_B0);

    seed = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];
    ira[ip0*d_threads + rngIndex] = seed;
    seed ^= ira[ip3*d_threads + rngIndex];   

    ip0 = ((ip0 + 1)&PRWHEELMASK);
    ip1 = ((ip1 + 1)&PRWHEELMASK);
    ip2 = ((ip2 + 1)&PRWHEELMASK);
    ip3 = ((ip3 + 1)&PRWHEELMASK);

    UPDATE_RED_PRNG(s1, texture_B1);

    seed = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];
    ira[ip0*d_threads + rngIndex] = seed;
    seed ^= ira[ip3*d_threads + rngIndex];   

    ip0 = ((ip0 + 1)&PRWHEELMASK);
    ip1 = ((ip1 + 1)&PRWHEELMASK);
    ip2 = ((ip2 + 1)&PRWHEELMASK);
    ip3 = ((ip3 + 1)&PRWHEELMASK);

    UPDATE_RED_PRNG(s2, texture_B2);

    seed = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];
    ira[ip0*d_threads + rngIndex] = seed;
    seed ^= ira[ip3*d_threads + rngIndex];   

    ip0 = ((ip0 + 1)&PRWHEELMASK);
    ip1 = ((ip1 + 1)&PRWHEELMASK);
    ip2 = ((ip2 + 1)&PRWHEELMASK);
    ip3 = ((ip3 + 1)&PRWHEELMASK);

    UPDATE_RED_PRNG(s3, texture_B3);

    count += 4;
  }

  if(count < spins && count > 0){
    for(int i = count; i < spins; i++){
      ira[ip0*d_threads + rngIndex] = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];

      ip0 = ((ip0 + 1)&PRWHEELMASK);
      ip1 = ((ip1 + 1)&PRWHEELMASK);
      ip2 = ((ip2 + 1)&PRWHEELMASK);
      ip3 = ((ip3 + 1)&PRWHEELMASK);
    }
  }

  return;
}

__global__ void MCStepReds_Standard_MT19937(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *rand, int k0, int k1, int k2, int spins){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = kk + blockIdx.y*d_loopStride;

  int count = 0;
  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    int x = k%d_hL; int y = (k/d_hL)%d_L; int z = k/d_hA;
    int par = (y^z)&1;
    
    int spx = i - x + SP(x + 1 - (par^1), d_hL);
    int smx = i - x + SM(x - 1 + par, d_hL);
    int spy = i + (SP(y + 1, d_L) - y)*d_hL;
    int smy = i + (SM(y - 1, d_L) - y)*d_hL;
    int spz = i + (SP(z + 1, d_L) - z)*d_hA;
    int smz = i + (SM(z - 1, d_L) - z)*d_hA;

    RNGT seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK); 
    seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
    rand[k0*d_threads + rngIndex] = seed;
    
    k0 = SP(k0 + 1, N_MT);
    k1 = SP(k1 + 1, N_MT);
    k2 = SP(k2 + 1, N_MT);
    
    seed ^= (seed >> 11);
    seed ^= (seed << 7) & 0x9d2c5680U;
    seed ^= (seed << 15) & 0xefc60000U;
    seed ^= (seed >> 18);
    
    UPDATE_RED_PRNG(s0, texture_B0);

    seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK); 
    seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
    rand[k0*d_threads + rngIndex] = seed;
    
    k0 = SP(k0 + 1, N_MT);
    k1 = SP(k1 + 1, N_MT);
    k2 = SP(k2 + 1, N_MT);
    
    seed ^= (seed >> 11);
    seed ^= (seed << 7) & 0x9d2c5680U;
    seed ^= (seed << 15) & 0xefc60000U;
    seed ^= (seed >> 18);

    UPDATE_RED_PRNG(s1, texture_B1);

    seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK); 
    seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
    rand[k0*d_threads + rngIndex] = seed;
    
    k0 = SP(k0 + 1, N_MT);
    k1 = SP(k1 + 1, N_MT);
    k2 = SP(k2 + 1, N_MT);
    
    seed ^= (seed >> 11);
    seed ^= (seed << 7) & 0x9d2c5680U;
    seed ^= (seed << 15) & 0xefc60000U;
    seed ^= (seed >> 18);

    UPDATE_RED_PRNG(s2, texture_B2);

    seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK); 
    seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
    rand[k0*d_threads + rngIndex] = seed;
    
    k0 = SP(k0 + 1, N_MT);
    k1 = SP(k1 + 1, N_MT);
    k2 = SP(k2 + 1, N_MT);
    
    seed ^= (seed >> 11);
    seed ^= (seed << 7) & 0x9d2c5680U;
    seed ^= (seed << 15) & 0xefc60000U;
    seed ^= (seed >> 18);

    UPDATE_RED_PRNG(s3, texture_B3);

    count += 4;
  }

  if(count < spins && count > 0){
    for(int i = count; i < spins; i++){
      RNGT seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK);
      seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
      rand[k0*d_threads + rngIndex] = seed;
      
      k0 = SP(k0 + 1, N_MT);
      k1 = SP(k1 + 1, N_MT);
      k2 = SP(k2 + 1, N_MT);      
    }
  }

  return;
}

__global__ void MCStepReds_Standard_MTGP32(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, curandStateMtgp32 *state, int spins){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = blockIdx.x + gridDim.x*blockIdx.y;

  int count = 0;
  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    int x = k%d_hL; int y = (k/d_hL)%d_L; int z = k/d_hA;
    int par = (y^z)&1;
    
    int spx = i - x + SP(x + 1 - (par^1), d_hL);
    int smx = i - x + SM(x - 1 + par, d_hL);
    int spy = i + (SP(y + 1, d_L) - y)*d_hL;
    int smy = i + (SM(y - 1, d_L) - y)*d_hL;
    int spz = i + (SP(z + 1, d_L) - z)*d_hA;
    int smz = i + (SM(z - 1, d_L) - z)*d_hA;

    RNGT seed = curand(&state[rngIndex]);
    
    UPDATE_RED_PRNG(s0, texture_B0);

    seed = curand(&state[rngIndex]);

    UPDATE_RED_PRNG(s1, texture_B1);

    seed = curand(&state[rngIndex]);

    UPDATE_RED_PRNG(s2, texture_B2);

    seed = curand(&state[rngIndex]);

    UPDATE_RED_PRNG(s3, texture_B3);

    count += 4;
  }

  if(count < spins && count > 0){
    for(int i = count; i < spins; i++){
      curand(&state[rngIndex]);
    }
  }

  return;
}

__global__ void MCStepReds_Standard_XORWOW(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, curandState *state){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = kk + blockIdx.y*d_loopStride;
  curandState localPrng = state[rngIndex];

  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    int x = k%d_hL; int y = (k/d_hL)%d_L; int z = k/d_hA;
    int par = (y^z)&1;
    
    int spx = i - x + SP(x + 1 - (par^1), d_hL);
    int smx = i - x + SM(x - 1 + par, d_hL);
    int spy = i + (SP(y + 1, d_L) - y)*d_hL;
    int smy = i + (SM(y - 1, d_L) - y)*d_hL;
    int spz = i + (SP(z + 1, d_L) - z)*d_hA;
    int smz = i + (SM(z - 1, d_L) - z)*d_hA;

    RNGT seed = curand(&localPrng);
    
    UPDATE_RED_PRNG(s0, texture_B0);

    seed = curand(&localPrng);

    UPDATE_RED_PRNG(s1, texture_B1);

    seed = curand(&localPrng);

    UPDATE_RED_PRNG(s2, texture_B2);

    seed = curand(&localPrng);

    UPDATE_RED_PRNG(s3, texture_B3);
  }

  state[rngIndex] = localPrng;

  return;
}

__global__ void MCStepBlues_Standard(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *rand){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = kk + blockIdx.y*d_loopStride;
  RNGT seed = rand[rngIndex];

  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    int x = k%d_hL; int y = (k/d_hL)%d_L; int z = k/d_hA;
    int par = (y^z)&1;
    
    int spx = i - x + SP((x + 1 - par), d_hL);
    int smx = i - x + SM((x - 1 + (par^1)), d_hL);
    int spy = i + (SP(y + 1, d_L) - y)*d_hL;
    int smy = i + (SM(y - 1, d_L) - y)*d_hL;
    int spz = i + (SP(z + 1, d_L) - z)*d_hA;
    int smz = i + (SM(z - 1, d_L) - z)*d_hA;
    
    UPDATE_BLUE(s0, texture_R0);
    UPDATE_BLUE(s1, texture_R1);
    UPDATE_BLUE(s2, texture_R2);
    UPDATE_BLUE(s3, texture_R3);
  }

  rand[rngIndex] = seed;

  return;
}

__global__ void MCStepBlues_Standard_PR(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *ira, int ip0, int ip1, int ip2, int ip3, int spins){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = kk + blockIdx.y*d_loopStride;

  int count = 0;
  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    int x = k%d_hL; int y = (k/d_hL)%d_L; int z = k/d_hA;
    int par = (y^z)&1;
    
    int spx = i - x + SP((x + 1 - par), d_hL);
    int smx = i - x + SM((x - 1 + (par^1)), d_hL);
    int spy = i + (SP(y + 1, d_L) - y)*d_hL;
    int smy = i + (SM(y - 1, d_L) - y)*d_hL;
    int spz = i + (SP(z + 1, d_L) - z)*d_hA;
    int smz = i + (SM(z - 1, d_L) - z)*d_hA;
    
    RNGT seed = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];
    ira[ip0*d_threads + rngIndex] = seed;
    seed ^= ira[ip3*d_threads + rngIndex];
    
    ip0 = ((ip0 + 1)&PRWHEELMASK);
    ip1 = ((ip1 + 1)&PRWHEELMASK);
    ip2 = ((ip2 + 1)&PRWHEELMASK);
    ip3 = ((ip3 + 1)&PRWHEELMASK);

    UPDATE_BLUE_PRNG(s0, texture_R0);

    seed = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];
    ira[ip0*d_threads + rngIndex] = seed;
    seed ^= ira[ip3*d_threads + rngIndex];   

    ip0 = ((ip0 + 1)&PRWHEELMASK);
    ip1 = ((ip1 + 1)&PRWHEELMASK);
    ip2 = ((ip2 + 1)&PRWHEELMASK);
    ip3 = ((ip3 + 1)&PRWHEELMASK);

    UPDATE_BLUE_PRNG(s1, texture_R1);

    seed = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];
    ira[ip0*d_threads + rngIndex] = seed;
    seed ^= ira[ip3*d_threads + rngIndex];   

    ip0 = ((ip0 + 1)&PRWHEELMASK);
    ip1 = ((ip1 + 1)&PRWHEELMASK);
    ip2 = ((ip2 + 1)&PRWHEELMASK);
    ip3 = ((ip3 + 1)&PRWHEELMASK);

    UPDATE_BLUE_PRNG(s2, texture_R2);

    seed = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];
    ira[ip0*d_threads + rngIndex] = seed;
    seed ^= ira[ip3*d_threads + rngIndex];   

    ip0 = ((ip0 + 1)&PRWHEELMASK);
    ip1 = ((ip1 + 1)&PRWHEELMASK);
    ip2 = ((ip2 + 1)&PRWHEELMASK);
    ip3 = ((ip3 + 1)&PRWHEELMASK);

    UPDATE_BLUE_PRNG(s3, texture_R3);

    count += 4;
  }

  if(count < spins && count > 0){
    for(int i = count; i < spins; i++){
      ira[ip0*d_threads + rngIndex] = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];

      ip0 = ((ip0 + 1)&PRWHEELMASK);
      ip1 = ((ip1 + 1)&PRWHEELMASK);
      ip2 = ((ip2 + 1)&PRWHEELMASK);
      ip3 = ((ip3 + 1)&PRWHEELMASK);
    }
  }

  return;
}

__global__ void MCStepBlues_Standard_MT19937(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *rand, int k0, int k1, int k2, int spins){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = kk + blockIdx.y*d_loopStride;

  int count = 0;
  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    int x = k%d_hL; int y = (k/d_hL)%d_L; int z = k/d_hA;
    int par = (y^z)&1;
    
    int spx = i - x + SP((x + 1 - par), d_hL);
    int smx = i - x + SM((x - 1 + (par^1)), d_hL);
    int spy = i + (SP(y + 1, d_L) - y)*d_hL;
    int smy = i + (SM(y - 1, d_L) - y)*d_hL;
    int spz = i + (SP(z + 1, d_L) - z)*d_hA;
    int smz = i + (SM(z - 1, d_L) - z)*d_hA;
    
    RNGT seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK); 
    seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
    rand[k0*d_threads + rngIndex] = seed;
    
    k0 = SP(k0 + 1, N_MT);
    k1 = SP(k1 + 1, N_MT);
    k2 = SP(k2 + 1, N_MT);
    
    seed ^= (seed >> 11);
    seed ^= (seed << 7) & 0x9d2c5680U;
    seed ^= (seed << 15) & 0xefc60000U;
    seed ^= (seed >> 18);

    UPDATE_BLUE_PRNG(s0, texture_R0);

    seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK); 
    seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
    rand[k0*d_threads + rngIndex] = seed;
    
    k0 = SP(k0 + 1, N_MT);
    k1 = SP(k1 + 1, N_MT);
    k2 = SP(k2 + 1, N_MT);
    
    seed ^= (seed >> 11);
    seed ^= (seed << 7) & 0x9d2c5680U;
    seed ^= (seed << 15) & 0xefc60000U;
    seed ^= (seed >> 18);

    UPDATE_BLUE_PRNG(s1, texture_R1);

    seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK); 
    seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
    rand[k0*d_threads + rngIndex] = seed;
    
    k0 = SP(k0 + 1, N_MT);
    k1 = SP(k1 + 1, N_MT);
    k2 = SP(k2 + 1, N_MT);
    
    seed ^= (seed >> 11);
    seed ^= (seed << 7) & 0x9d2c5680U;
    seed ^= (seed << 15) & 0xefc60000U;
    seed ^= (seed >> 18);

    UPDATE_BLUE_PRNG(s2, texture_R2);

    seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK); 
    seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
    rand[k0*d_threads + rngIndex] = seed;
    
    k0 = SP(k0 + 1, N_MT);
    k1 = SP(k1 + 1, N_MT);
    k2 = SP(k2 + 1, N_MT);
    
    seed ^= (seed >> 11);
    seed ^= (seed << 7) & 0x9d2c5680U;
    seed ^= (seed << 15) & 0xefc60000U;
    seed ^= (seed >> 18);

    UPDATE_BLUE_PRNG(s3, texture_R3);

    count += 4;
  }

  if(count < spins && count > 0){
    for(int i = count; i < spins; i++){
      RNGT seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK);
      seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
      rand[k0*d_threads + rngIndex] = seed;
      
      k0 = SP(k0 + 1, N_MT);
      k1 = SP(k1 + 1, N_MT);
      k2 = SP(k2 + 1, N_MT);      
    }
  }

  return;
}

__global__ void MCStepBlues_Standard_MTGP32(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, curandStateMtgp32 *state, int spins){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = blockIdx.x + gridDim.x*blockIdx.y;
  
  int count = 0;
  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    int x = k%d_hL; int y = (k/d_hL)%d_L; int z = k/d_hA;
    int par = (y^z)&1;
    
    int spx = i - x + SP((x + 1 - par), d_hL);
    int smx = i - x + SM((x - 1 + (par^1)), d_hL);
    int spy = i + (SP(y + 1, d_L) - y)*d_hL;
    int smy = i + (SM(y - 1, d_L) - y)*d_hL;
    int spz = i + (SP(z + 1, d_L) - z)*d_hA;
    int smz = i + (SM(z - 1, d_L) - z)*d_hA;
    
    RNGT seed = curand(&state[rngIndex]);

    UPDATE_BLUE_PRNG(s0, texture_R0);

    seed = curand(&state[rngIndex]);

    UPDATE_BLUE_PRNG(s1, texture_R1);

    seed = curand(&state[rngIndex]);

    UPDATE_BLUE_PRNG(s2, texture_R2);

    seed = curand(&state[rngIndex]);

    UPDATE_BLUE_PRNG(s3, texture_R3);

    count += 4;
  }

  if(count < spins && count > 0){
    for(int i = count; i < spins; i++){
      curand(&state[rngIndex]);
    }
  }

  return;
}

__global__ void MCStepBlues_Standard_XORWOW(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, curandState *state){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = kk + blockIdx.y*d_loopStride;
  curandState localPrng = state[rngIndex];
    
  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    int x = k%d_hL; int y = (k/d_hL)%d_L; int z = k/d_hA;
    int par = (y^z)&1;
    
    int spx = i - x + SP((x + 1 - par), d_hL);
    int smx = i - x + SM((x - 1 + (par^1)), d_hL);
    int spy = i + (SP(y + 1, d_L) - y)*d_hL;
    int smy = i + (SM(y - 1, d_L) - y)*d_hL;
    int spz = i + (SP(z + 1, d_L) - z)*d_hA;
    int smz = i + (SM(z - 1, d_L) - z)*d_hA;
    
    RNGT seed = curand(&localPrng);

    UPDATE_BLUE_PRNG(s0, texture_R0);

    seed = curand(&localPrng);

    UPDATE_BLUE_PRNG(s1, texture_R1);

    seed = curand(&localPrng);

    UPDATE_BLUE_PRNG(s2, texture_R2);

    seed = curand(&localPrng);

    UPDATE_BLUE_PRNG(s3, texture_R3);
  }

  state[rngIndex] = localPrng;

  return;
}

__global__ void MCStepReds_StandardGrid(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *rand){

  const int i = threadIdx.x + threadIdx.y*blockDim.x + blockIdx.x*d_block_xStrideS + blockIdx.y*d_block_yStrideS + blockIdx.z*d_hV;
  const int y = threadIdx.y + blockIdx.x*blockDim.y;

  RNGT seed = rand[i];
  int par = (y^blockIdx.y)&1;

  int spx = i - threadIdx.x + SP(threadIdx.x + 1 - (par^1), d_hL);
  int smx = i - threadIdx.x + SMM(threadIdx.x + par, d_hL);
  int spy = i + (SP(y + 1, d_L) - y)*d_hL;
  int smy = i + (SM(y - 1, d_L) - y)*d_hL;
  int spz = i + (SP(blockIdx.y + 1, d_L) - blockIdx.y)*d_hA;
  int smz = i + (SMM(blockIdx.y, d_L) - blockIdx.y)*d_hA;
  
  UPDATE_RED(s0, texture_B0);
  UPDATE_RED(s1, texture_B1);
  UPDATE_RED(s2, texture_B2);
  UPDATE_RED(s3, texture_B3);
  
  rand[i] = seed;

  return;
}

__global__ void MCStepReds_StandardGrid_PR(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *ira, int ip0, int ip1, int ip2, int ip3, int spins){

  const int i = threadIdx.x + threadIdx.y*blockDim.x + blockIdx.x*d_block_xStrideS + blockIdx.y*d_block_yStrideS + blockIdx.z*d_hV;
  const int y = threadIdx.y + blockIdx.x*blockDim.y;

  int par = (y^blockIdx.y)&1;

  int spx = i - threadIdx.x + SP(threadIdx.x + 1 - (par^1), d_hL);
  int smx = i - threadIdx.x + SMM(threadIdx.x + par, d_hL);
  int spy = i + (SP(y + 1, d_L) - y)*d_hL;
  int smy = i + (SM(y - 1, d_L) - y)*d_hL;
  int spz = i + (SP(blockIdx.y + 1, d_L) - blockIdx.y)*d_hA;
  int smz = i + (SMM(blockIdx.y, d_L) - blockIdx.y)*d_hA;
  
  RNGT seed = ira[ip1*d_threads + i] + ira[ip2*d_threads + i];
  ira[ip0*d_threads + i] = seed;
  seed ^= ira[ip3*d_threads + i];
  
  ip0 = ((ip0 + 1)&PRWHEELMASK);
  ip1 = ((ip1 + 1)&PRWHEELMASK);
  ip2 = ((ip2 + 1)&PRWHEELMASK);
  ip3 = ((ip3 + 1)&PRWHEELMASK);

  UPDATE_RED_PRNG(s0, texture_B0);

  seed = ira[ip1*d_threads + i] + ira[ip2*d_threads + i];
  ira[ip0*d_threads + i] = seed;
  seed ^= ira[ip3*d_threads + i];   
  
  ip0 = ((ip0 + 1)&PRWHEELMASK);
  ip1 = ((ip1 + 1)&PRWHEELMASK);
  ip2 = ((ip2 + 1)&PRWHEELMASK);
  ip3 = ((ip3 + 1)&PRWHEELMASK);

  UPDATE_RED_PRNG(s1, texture_B1);

  seed = ira[ip1*d_threads + i] + ira[ip2*d_threads + i];
  ira[ip0*d_threads + i] = seed;
  seed ^= ira[ip3*d_threads + i];   
  
  ip0 = ((ip0 + 1)&PRWHEELMASK);
  ip1 = ((ip1 + 1)&PRWHEELMASK);
  ip2 = ((ip2 + 1)&PRWHEELMASK);
  ip3 = ((ip3 + 1)&PRWHEELMASK);

  UPDATE_RED_PRNG(s2, texture_B2);

  seed = ira[ip1*d_threads + i] + ira[ip2*d_threads + i];
  ira[ip0*d_threads + i] = seed;
  seed ^= ira[ip3*d_threads + i];   
  
  ip0 = ((ip0 + 1)&PRWHEELMASK);
  ip1 = ((ip1 + 1)&PRWHEELMASK);
  ip2 = ((ip2 + 1)&PRWHEELMASK);
  ip3 = ((ip3 + 1)&PRWHEELMASK);

  UPDATE_RED_PRNG(s3, texture_B3);  

  return;
}

__global__ void MCStepReds_StandardGrid_MT19937(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *rand, int k0, int k1, int k2, int spins){

  const int i = threadIdx.x + threadIdx.y*blockDim.x + blockIdx.x*d_block_xStrideS + blockIdx.y*d_block_yStrideS + blockIdx.z*d_hV;
  const int y = threadIdx.y + blockIdx.x*blockDim.y;

  int par = (y^blockIdx.y)&1;

  int spx = i - threadIdx.x + SP(threadIdx.x + 1 - (par^1), d_hL);
  int smx = i - threadIdx.x + SMM(threadIdx.x + par, d_hL);
  int spy = i + (SP(y + 1, d_L) - y)*d_hL;
  int smy = i + (SM(y - 1, d_L) - y)*d_hL;
  int spz = i + (SP(blockIdx.y + 1, d_L) - blockIdx.y)*d_hA;
  int smz = i + (SMM(blockIdx.y, d_L) - blockIdx.y)*d_hA;
  
  RNGT seed = (rand[k0*d_threads + i] & UPPER_MASK)|(rand[k1*d_threads + i] & LOWER_MASK); 
  seed = rand[k2*d_threads + i]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
  rand[k0*d_threads + i] = seed;
  
  k0 = SP(k0 + 1, N_MT);
  k1 = SP(k1 + 1, N_MT);
  k2 = SP(k2 + 1, N_MT);
  
  seed ^= (seed >> 11);
  seed ^= (seed << 7) & 0x9d2c5680U;
  seed ^= (seed << 15) & 0xefc60000U;
  seed ^= (seed >> 18);

  UPDATE_RED_PRNG(s0, texture_B0);

  seed = (rand[k0*d_threads + i] & UPPER_MASK)|(rand[k1*d_threads + i] & LOWER_MASK); 
  seed = rand[k2*d_threads + i]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
  rand[k0*d_threads + i] = seed;
  
  k0 = SP(k0 + 1, N_MT);
  k1 = SP(k1 + 1, N_MT);
  k2 = SP(k2 + 1, N_MT);
  
  seed ^= (seed >> 11);
  seed ^= (seed << 7) & 0x9d2c5680U;
  seed ^= (seed << 15) & 0xefc60000U;
  seed ^= (seed >> 18);

  UPDATE_RED_PRNG(s1, texture_B1);

  seed = (rand[k0*d_threads + i] & UPPER_MASK)|(rand[k1*d_threads + i] & LOWER_MASK); 
  seed = rand[k2*d_threads + i]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
  rand[k0*d_threads + i] = seed;
  
  k0 = SP(k0 + 1, N_MT);
  k1 = SP(k1 + 1, N_MT);
  k2 = SP(k2 + 1, N_MT);
  
  seed ^= (seed >> 11);
  seed ^= (seed << 7) & 0x9d2c5680U;
  seed ^= (seed << 15) & 0xefc60000U;
  seed ^= (seed >> 18);

  UPDATE_RED_PRNG(s2, texture_B2);

  seed = (rand[k0*d_threads + i] & UPPER_MASK)|(rand[k1*d_threads + i] & LOWER_MASK); 
  seed = rand[k2*d_threads + i]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
  rand[k0*d_threads + i] = seed;
  
  k0 = SP(k0 + 1, N_MT);
  k1 = SP(k1 + 1, N_MT);
  k2 = SP(k2 + 1, N_MT);
  
  seed ^= (seed >> 11);
  seed ^= (seed << 7) & 0x9d2c5680U;
  seed ^= (seed << 15) & 0xefc60000U;
  seed ^= (seed >> 18);

  UPDATE_RED_PRNG(s3, texture_B3);  

  return;
}

__global__ void MCStepReds_StandardGrid_MTGP32(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, curandStateMtgp32 *state, int spins){

  const int i = threadIdx.x + threadIdx.y*blockDim.x + blockIdx.x*d_block_xStrideS + blockIdx.y*d_block_yStrideS + blockIdx.z*d_hV;
  const int y = threadIdx.y + blockIdx.x*blockDim.y;
  const int rngIndex = blockIdx.x + gridDim.x*blockIdx.y;

  int par = (y^blockIdx.y)&1;

  int spx = i - threadIdx.x + SP(threadIdx.x + 1 - (par^1), d_hL);
  int smx = i - threadIdx.x + SMM(threadIdx.x + par, d_hL);
  int spy = i + (SP(y + 1, d_L) - y)*d_hL;
  int smy = i + (SM(y - 1, d_L) - y)*d_hL;
  int spz = i + (SP(blockIdx.y + 1, d_L) - blockIdx.y)*d_hA;
  int smz = i + (SMM(blockIdx.y, d_L) - blockIdx.y)*d_hA;
  
  RNGT seed = curand(&state[rngIndex]);

  UPDATE_RED_PRNG(s0, texture_B0);

  seed = curand(&state[rngIndex]);

  UPDATE_RED_PRNG(s1, texture_B1);

  seed = curand(&state[rngIndex]);

  UPDATE_RED_PRNG(s2, texture_B2);

  seed = curand(&state[rngIndex]);

  UPDATE_RED_PRNG(s3, texture_B3);  

  return;
}

__global__ void MCStepReds_StandardGrid_XORWOW(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, curandState *state){

  const int i = threadIdx.x + threadIdx.y*blockDim.x + blockIdx.x*d_block_xStrideS + blockIdx.y*d_block_yStrideS + blockIdx.z*d_hV;
  const int y = threadIdx.y + blockIdx.x*blockDim.y;

  curandState localPrng = state[i];

  int par = (y^blockIdx.y)&1;

  int spx = i - threadIdx.x + SP(threadIdx.x + 1 - (par^1), d_hL);
  int smx = i - threadIdx.x + SMM(threadIdx.x + par, d_hL);
  int spy = i + (SP(y + 1, d_L) - y)*d_hL;
  int smy = i + (SM(y - 1, d_L) - y)*d_hL;
  int spz = i + (SP(blockIdx.y + 1, d_L) - blockIdx.y)*d_hA;
  int smz = i + (SMM(blockIdx.y, d_L) - blockIdx.y)*d_hA;
  
  RNGT seed = curand(&localPrng);

  UPDATE_RED_PRNG(s0, texture_B0);

  seed = curand(&localPrng);

  UPDATE_RED_PRNG(s1, texture_B1);

  seed = curand(&localPrng);

  UPDATE_RED_PRNG(s2, texture_B2);

  seed = curand(&localPrng);

  UPDATE_RED_PRNG(s3, texture_B3);  

  state[i] = localPrng;
  return;
}

__global__ void MCStepBlues_StandardGrid(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *rand){

  const int i = threadIdx.x + threadIdx.y*blockDim.x + blockIdx.x*d_block_xStrideS + blockIdx.y*d_block_yStrideS + blockIdx.z*d_hV;
  const int y = threadIdx.y + blockIdx.x*blockDim.y;

  RNGT seed = rand[i];

  int par = (y^blockIdx.y)&1;

  int spx = i - threadIdx.x + SP(threadIdx.x + 1 - par, d_hL);
  int smx = i - threadIdx.x + SMM(threadIdx.x + (par^1), d_hL);
  int spy = i + (SP(y + 1, d_L) - y)*d_hL;
  int smy = i + (SM(y - 1, d_L) - y)*d_hL;
  int spz = i + (SP(blockIdx.y + 1, d_L) - blockIdx.y)*d_hA;
  int smz = i + (SMM(blockIdx.y, d_L) - blockIdx.y)*d_hA;
  
  UPDATE_BLUE(s0, texture_R0);
  UPDATE_BLUE(s1, texture_R1);
  UPDATE_BLUE(s2, texture_R2);
  UPDATE_BLUE(s3, texture_R3);
  
  rand[i] = seed;

  return;
}

__global__ void MCStepBlues_StandardGrid_PR(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *ira, int ip0, int ip1, int ip2, int ip3, int spins){

  const int i = threadIdx.x + threadIdx.y*blockDim.x + blockIdx.x*d_block_xStrideS + blockIdx.y*d_block_yStrideS + blockIdx.z*d_hV;
  const int y = threadIdx.y + blockIdx.x*blockDim.y;

  int par = (y^blockIdx.y)&1;

  int spx = i - threadIdx.x + SP(threadIdx.x + 1 - par, d_hL);
  int smx = i - threadIdx.x + SMM(threadIdx.x + (par^1), d_hL);
  int spy = i + (SP(y + 1, d_L) - y)*d_hL;
  int smy = i + (SM(y - 1, d_L) - y)*d_hL;
  int spz = i + (SP(blockIdx.y + 1, d_L) - blockIdx.y)*d_hA;
  int smz = i + (SMM(blockIdx.y, d_L) - blockIdx.y)*d_hA;

  RNGT seed = ira[ip1*d_threads + i] + ira[ip2*d_threads + i];
  ira[ip0*d_threads + i] = seed;
  seed ^= ira[ip3*d_threads + i];
  
  ip0 = ((ip0 + 1)&PRWHEELMASK);
  ip1 = ((ip1 + 1)&PRWHEELMASK);
  ip2 = ((ip2 + 1)&PRWHEELMASK);
  ip3 = ((ip3 + 1)&PRWHEELMASK);
  
  UPDATE_BLUE_PRNG(s0, texture_R0);

  seed = ira[ip1*d_threads + i] + ira[ip2*d_threads + i];
  ira[ip0*d_threads + i] = seed;
  seed ^= ira[ip3*d_threads + i];   
  
  ip0 = ((ip0 + 1)&PRWHEELMASK);
  ip1 = ((ip1 + 1)&PRWHEELMASK);
  ip2 = ((ip2 + 1)&PRWHEELMASK);
  ip3 = ((ip3 + 1)&PRWHEELMASK);

  UPDATE_BLUE_PRNG(s1, texture_R1);

  seed = ira[ip1*d_threads + i] + ira[ip2*d_threads + i];
  ira[ip0*d_threads + i] = seed;
  seed ^= ira[ip3*d_threads + i];   
  
  ip0 = ((ip0 + 1)&PRWHEELMASK);
  ip1 = ((ip1 + 1)&PRWHEELMASK);
  ip2 = ((ip2 + 1)&PRWHEELMASK);
  ip3 = ((ip3 + 1)&PRWHEELMASK);

  UPDATE_BLUE_PRNG(s2, texture_R2);

  seed = ira[ip1*d_threads + i] + ira[ip2*d_threads + i];
  ira[ip0*d_threads + i] = seed;
  seed ^= ira[ip3*d_threads + i];   
  
  ip0 = ((ip0 + 1)&PRWHEELMASK);
  ip1 = ((ip1 + 1)&PRWHEELMASK);
  ip2 = ((ip2 + 1)&PRWHEELMASK);
  ip3 = ((ip3 + 1)&PRWHEELMASK);

  UPDATE_BLUE(s3, texture_R3);
  
  return;
}

__global__ void MCStepBlues_StandardGrid_MT19937(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *rand, int k0, int k1, int k2, int spins){

  const int i = threadIdx.x + threadIdx.y*blockDim.x + blockIdx.x*d_block_xStrideS + blockIdx.y*d_block_yStrideS + blockIdx.z*d_hV;
  const int y = threadIdx.y + blockIdx.x*blockDim.y;

  int par = (y^blockIdx.y)&1;

  int spx = i - threadIdx.x + SP(threadIdx.x + 1 - par, d_hL);
  int smx = i - threadIdx.x + SMM(threadIdx.x + (par^1), d_hL);
  int spy = i + (SP(y + 1, d_L) - y)*d_hL;
  int smy = i + (SM(y - 1, d_L) - y)*d_hL;
  int spz = i + (SP(blockIdx.y + 1, d_L) - blockIdx.y)*d_hA;
  int smz = i + (SMM(blockIdx.y, d_L) - blockIdx.y)*d_hA;

  RNGT seed = (rand[k0*d_threads + i] & UPPER_MASK)|(rand[k1*d_threads + i] & LOWER_MASK); 
  seed = rand[k2*d_threads + i]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
  rand[k0*d_threads + i] = seed;
  
  k0 = SP(k0 + 1, N_MT);
  k1 = SP(k1 + 1, N_MT);
  k2 = SP(k2 + 1, N_MT);
  
  seed ^= (seed >> 11);
  seed ^= (seed << 7) & 0x9d2c5680U;
  seed ^= (seed << 15) & 0xefc60000U;
  seed ^= (seed >> 18);
  
  UPDATE_BLUE_PRNG(s0, texture_R0);

  seed = (rand[k0*d_threads + i] & UPPER_MASK)|(rand[k1*d_threads + i] & LOWER_MASK); 
  seed = rand[k2*d_threads + i]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
  rand[k0*d_threads + i] = seed;
  
  k0 = SP(k0 + 1, N_MT);
  k1 = SP(k1 + 1, N_MT);
  k2 = SP(k2 + 1, N_MT);
  
  seed ^= (seed >> 11);
  seed ^= (seed << 7) & 0x9d2c5680U;
  seed ^= (seed << 15) & 0xefc60000U;
  seed ^= (seed >> 18);

  UPDATE_BLUE_PRNG(s1, texture_R1);

  seed = (rand[k0*d_threads + i] & UPPER_MASK)|(rand[k1*d_threads + i] & LOWER_MASK); 
  seed = rand[k2*d_threads + i]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
  rand[k0*d_threads + i] = seed;
  
  k0 = SP(k0 + 1, N_MT);
  k1 = SP(k1 + 1, N_MT);
  k2 = SP(k2 + 1, N_MT);
  
  seed ^= (seed >> 11);
  seed ^= (seed << 7) & 0x9d2c5680U;
  seed ^= (seed << 15) & 0xefc60000U;
  seed ^= (seed >> 18);

  UPDATE_BLUE_PRNG(s2, texture_R2);

  seed = (rand[k0*d_threads + i] & UPPER_MASK)|(rand[k1*d_threads + i] & LOWER_MASK); 
  seed = rand[k2*d_threads + i]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
  rand[k0*d_threads + i] = seed;
  
  k0 = SP(k0 + 1, N_MT);
  k1 = SP(k1 + 1, N_MT);
  k2 = SP(k2 + 1, N_MT);
  
  seed ^= (seed >> 11);
  seed ^= (seed << 7) & 0x9d2c5680U;
  seed ^= (seed << 15) & 0xefc60000U;
  seed ^= (seed >> 18);

  UPDATE_BLUE(s3, texture_R3);
  
  return;
}

__global__ void MCStepBlues_StandardGrid_MTGP32(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, curandStateMtgp32 *state, int spins){

  const int i = threadIdx.x + threadIdx.y*blockDim.x + blockIdx.x*d_block_xStrideS + blockIdx.y*d_block_yStrideS + blockIdx.z*d_hV;
  const int y = threadIdx.y + blockIdx.x*blockDim.y;
  const int rngIndex = blockIdx.x + gridDim.x*blockIdx.y;

  int par = (y^blockIdx.y)&1;

  int spx = i - threadIdx.x + SP(threadIdx.x + 1 - par, d_hL);
  int smx = i - threadIdx.x + SMM(threadIdx.x + (par^1), d_hL);
  int spy = i + (SP(y + 1, d_L) - y)*d_hL;
  int smy = i + (SM(y - 1, d_L) - y)*d_hL;
  int spz = i + (SP(blockIdx.y + 1, d_L) - blockIdx.y)*d_hA;
  int smz = i + (SMM(blockIdx.y, d_L) - blockIdx.y)*d_hA;

  RNGT seed = curand(&state[rngIndex]);
  
  UPDATE_BLUE_PRNG(s0, texture_R0);

  seed = curand(&state[rngIndex]);

  UPDATE_BLUE_PRNG(s1, texture_R1);

  seed = curand(&state[rngIndex]);

  UPDATE_BLUE_PRNG(s2, texture_R2);

  seed = curand(&state[rngIndex]);

  UPDATE_BLUE(s3, texture_R3);
  
  return;
}

__global__ void MCStepBlues_StandardGrid_XORWOW(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, curandState *state){

  const int i = threadIdx.x + threadIdx.y*blockDim.x + blockIdx.x*d_block_xStrideS + blockIdx.y*d_block_yStrideS + blockIdx.z*d_hV;
  const int y = threadIdx.y + blockIdx.x*blockDim.y;

  curandState localPrng = state[i];

  int par = (y^blockIdx.y)&1;

  int spx = i - threadIdx.x + SP(threadIdx.x + 1 - par, d_hL);
  int smx = i - threadIdx.x + SMM(threadIdx.x + (par^1), d_hL);
  int spy = i + (SP(y + 1, d_L) - y)*d_hL;
  int smy = i + (SM(y - 1, d_L) - y)*d_hL;
  int spz = i + (SP(blockIdx.y + 1, d_L) - blockIdx.y)*d_hA;
  int smz = i + (SMM(blockIdx.y, d_L) - blockIdx.y)*d_hA;

  RNGT seed = curand(&localPrng);
  
  UPDATE_BLUE_PRNG(s0, texture_R0);

  seed = curand(&localPrng);

  UPDATE_BLUE_PRNG(s1, texture_R1);

  seed = curand(&localPrng);

  UPDATE_BLUE_PRNG(s2, texture_R2);

  seed = curand(&localPrng);

  UPDATE_BLUE(s3, texture_R3);
  
  state[i] = localPrng;

  return;
}


//////////////////////////////////////////////////////////////////////
// KERNEL SLICED
//////////////////////////////////////////////////////////////////////

__global__ void MCStepReds_Sliced(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *rand){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = kk + blockIdx.y*d_loopStride;

  RNGT seed = rand[rngIndex];

  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    int x = k%d_L;
    int y = (k/d_L)%d_L;
    int z = k/d_A;

    int smz = i + (SM(z - 1, d_hL) - z)*d_A;
    int spy = smz + (SP(y + 1, d_L) - y)*d_L; 
    int smy = i + (SM(y - 1, d_L) - y)*d_L; 
    int smx = spy - x + SM(x - 1, d_L);
    int spx = smy - x + SP(x + 1, d_L);
        
    UPDATE_RED_SLICED(s0, texture_B0);
    UPDATE_RED_SLICED(s1, texture_B1);
    UPDATE_RED_SLICED(s2, texture_B2);
    UPDATE_RED_SLICED(s3, texture_B3);
  }

  rand[rngIndex] = seed;

  return;
}

__global__ void MCStepReds_Sliced_MT19937(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *rand, int k0, int k1, int k2, int spins){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = kk + blockIdx.y*d_loopStride;

  int count = 0;
  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    int x = k%d_L;
    int y = (k/d_L)%d_L;
    int z = k/d_A;

    int smz = i + (SM(z - 1, d_hL) - z)*d_A;
    int spy = smz + (SP(y + 1, d_L) - y)*d_L; 
    int smy = i + (SM(y - 1, d_L) - y)*d_L; 
    int smx = spy - x + SM(x - 1, d_L);
    int spx = smy - x + SP(x + 1, d_L);

    RNGT seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK); 
    seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
    rand[k0*d_threads + rngIndex] = seed;
    
    k0 = SP(k0 + 1, N_MT);
    k1 = SP(k1 + 1, N_MT);
    k2 = SP(k2 + 1, N_MT);
    
    seed ^= (seed >> 11);
    seed ^= (seed << 7) & 0x9d2c5680U;
    seed ^= (seed << 15) & 0xefc60000U;
    seed ^= (seed >> 18);
    
    UPDATE_RED_SLICED_PRNG(s0, texture_B0);

    seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK);
    seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
    rand[k0*d_threads + rngIndex] = seed;
    
    k0 = SP(k0 + 1, N_MT);
    k1 = SP(k1 + 1, N_MT);
    k2 = SP(k2 + 1, N_MT);
    
    seed ^= (seed >> 11);
    seed ^= (seed << 7) & 0x9d2c5680U;
    seed ^= (seed << 15) & 0xefc60000U;
    seed ^= (seed >> 18);    

    UPDATE_RED_SLICED_PRNG(s1, texture_B1);

    seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK);
    seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
    rand[k0*d_threads + rngIndex] = seed;
    
    k0 = SP(k0 + 1, N_MT);
    k1 = SP(k1 + 1, N_MT);
    k2 = SP(k2 + 1, N_MT);
    
    seed ^= (seed >> 11);
    seed ^= (seed << 7) & 0x9d2c5680U;
    seed ^= (seed << 15) & 0xefc60000U;
    seed ^= (seed >> 18);    

    UPDATE_RED_SLICED_PRNG(s2, texture_B2);

    seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK);
    seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
    rand[k0*d_threads + rngIndex] = seed;
    
    k0 = SP(k0 + 1, N_MT);
    k1 = SP(k1 + 1, N_MT);
    k2 = SP(k2 + 1, N_MT);
    
    seed ^= (seed >> 11);
    seed ^= (seed << 7) & 0x9d2c5680U;
    seed ^= (seed << 15) & 0xefc60000U;
    seed ^= (seed >> 18);    

    UPDATE_RED_SLICED_PRNG(s3, texture_B3);

    count += 4;
  }

  if(count < spins && count != 0){
    for(int i = count; i < spins; i++){
      RNGT seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK);
      seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
      rand[k0*d_threads + rngIndex] = seed;
      
      k0 = SP(k0 + 1, N_MT);
      k1 = SP(k1 + 1, N_MT);
      k2 = SP(k2 + 1, N_MT);      
    }
  }
  return;
}

__global__ void MCStepReds_Sliced_PR(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *ira, int ip0, int ip1, int ip2, int ip3, int spins){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = kk + blockIdx.y*d_loopStride;

  int count = 0;
  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    int x = k%d_L;
    int y = (k/d_L)%d_L;
    int z = k/d_A;

    int smz = i + (SM(z - 1, d_hL) - z)*d_A;
    int spy = smz + (SP(y + 1, d_L) - y)*d_L; 
    int smy = i + (SM(y - 1, d_L) - y)*d_L; 
    int smx = spy - x + SM(x - 1, d_L);
    int spx = smy - x + SP(x + 1, d_L);

    RNGT seed = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];
    ira[ip0*d_threads + rngIndex] = seed;
    seed ^= ira[ip3*d_threads + rngIndex];
  
    ip0 = ((ip0 + 1)&PRWHEELMASK);
    ip1 = ((ip1 + 1)&PRWHEELMASK);
    ip2 = ((ip2 + 1)&PRWHEELMASK);
    ip3 = ((ip3 + 1)&PRWHEELMASK);
    
    UPDATE_RED_SLICED_PRNG(s0, texture_B0);

    seed = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];
    ira[ip0*d_threads + rngIndex] = seed;
    seed ^= ira[ip3*d_threads + rngIndex];

    ip0 = ((ip0 + 1)&PRWHEELMASK);
    ip1 = ((ip1 + 1)&PRWHEELMASK);
    ip2 = ((ip2 + 1)&PRWHEELMASK);
    ip3 = ((ip3 + 1)&PRWHEELMASK);

    UPDATE_RED_SLICED_PRNG(s1, texture_B1);

    seed = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];
    ira[ip0*d_threads + rngIndex] = seed;
    seed ^= ira[ip3*d_threads + rngIndex];
    
    ip0 = ((ip0 + 1)&PRWHEELMASK);
    ip1 = ((ip1 + 1)&PRWHEELMASK);
    ip2 = ((ip2 + 1)&PRWHEELMASK);
    ip3 = ((ip3 + 1)&PRWHEELMASK);

    UPDATE_RED_SLICED_PRNG(s2, texture_B2);

    seed = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];
    ira[ip0*d_threads + rngIndex] = seed;
    seed ^= ira[ip3*d_threads + rngIndex];

    ip0 = ((ip0 + 1)&PRWHEELMASK);
    ip1 = ((ip1 + 1)&PRWHEELMASK);
    ip2 = ((ip2 + 1)&PRWHEELMASK);
    ip3 = ((ip3 + 1)&PRWHEELMASK);

    UPDATE_RED_SLICED_PRNG(s3, texture_B3);

    count += 4;
  }

  if(count < spins && count > 0){
    for(int i = count; i < spins; i++){
      ira[ip0*d_threads + rngIndex] = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];
       
      ip0 = ((ip0 + 1)&PRWHEELMASK);
      ip1 = ((ip1 + 1)&PRWHEELMASK);
      ip2 = ((ip2 + 1)&PRWHEELMASK);
      ip3 = ((ip3 + 1)&PRWHEELMASK);
    }
  }
  return;
}

__global__ void MCStepReds_Sliced_MTGP32(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, curandStateMtgp32 *state, int spins){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = blockIdx.x + gridDim.x*blockIdx.y;

  int count = 0;
  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    int x = k%d_L;
    int y = (k/d_L)%d_L;
    int z = k/d_A;

    int smz = i + (SM(z - 1, d_hL) - z)*d_A;
    int spy = smz + (SP(y + 1, d_L) - y)*d_L; 
    int smy = i + (SM(y - 1, d_L) - y)*d_L; 
    int smx = spy - x + SM(x - 1, d_L);
    int spx = smy - x + SP(x + 1, d_L);

    RNGT seed = curand(&state[rngIndex]);
    
    UPDATE_RED_SLICED_PRNG(s0, texture_B0);

    seed = curand(&state[rngIndex]);

    UPDATE_RED_SLICED_PRNG(s1, texture_B1);

    seed = curand(&state[rngIndex]);

    UPDATE_RED_SLICED_PRNG(s2, texture_B2);

    seed = curand(&state[rngIndex]);

    UPDATE_RED_SLICED_PRNG(s3, texture_B3);

    count += 4;
  }

  if(count < spins && count > 0){
    for(int i = count; i < spins; i++){
      curand(&state[rngIndex]);
    }
  }

  return;
}

__global__ void MCStepReds_Sliced_XORWOW(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, curandState *state){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = kk + blockIdx.y*d_loopStride;
  curandState localPrng = state[rngIndex];

  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    int x = k%d_L;
    int y = (k/d_L)%d_L;
    int z = k/d_A;

    int smz = i + (SM(z - 1, d_hL) - z)*d_A;
    int spy = smz + (SP(y + 1, d_L) - y)*d_L; 
    int smy = i + (SM(y - 1, d_L) - y)*d_L; 
    int smx = spy - x + SM(x - 1, d_L);
    int spx = smy - x + SP(x + 1, d_L);

    RNGT seed = curand(&localPrng);
    
    UPDATE_RED_SLICED_PRNG(s0, texture_B0);

    seed = curand(&localPrng);

    UPDATE_RED_SLICED_PRNG(s1, texture_B1);

    seed = curand(&localPrng);

    UPDATE_RED_SLICED_PRNG(s2, texture_B2);

    seed = curand(&localPrng);

    UPDATE_RED_SLICED_PRNG(s3, texture_B3);
  }

  state[rngIndex] = localPrng;
  return;
}
  
__global__ void MCStepBlues_Sliced(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *rand){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = kk + blockIdx.y*d_loopStride;
  RNGT seed = rand[rngIndex];


  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    int x = k%d_L;
    int y = (k/d_L)%d_L;
    int z = k/d_A;
    
    int spz = i + (SP(z + 1, d_hL) - z)*d_A;    
    int spy = i + (SP(y + 1, d_L) - y)*d_L;
    int smy = spz + (SM(y - 1, d_L) - y)*d_L;
    int smx = spy - x + SM(x - 1, d_L);
    int spx = smy - x + SP(x + 1, d_L);
    
    UPDATE_BLUE_SLICED(s0, texture_R0);
    UPDATE_BLUE_SLICED(s1, texture_R1);
    UPDATE_BLUE_SLICED(s2, texture_R2);
    UPDATE_BLUE_SLICED(s3, texture_R3);
  }

  rand[rngIndex] = seed;

  return;
}

__global__ void MCStepBlues_Sliced_MT19937(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *rand, int k0, int k1, int k2, int spins){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = kk + blockIdx.y*d_loopStride;

  int count = 0;
  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    int x = k%d_L;
    int y = (k/d_L)%d_L;
    int z = k/d_A;
    
    int spz = i + (SP(z + 1, d_hL) - z)*d_A;    
    int spy = i + (SP(y + 1, d_L) - y)*d_L;
    int smy = spz + (SM(y - 1, d_L) - y)*d_L;
    int smx = spy - x + SM(x - 1, d_L);
    int spx = smy - x + SP(x + 1, d_L);

    RNGT seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK);
    seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
    rand[k0*d_threads + rngIndex] = seed;
    
    k0 = SP(k0 + 1, N_MT);
    k1 = SP(k1 + 1, N_MT);
    k2 = SP(k2 + 1, N_MT);
    
    seed ^= (seed >> 11);
    seed ^= (seed << 7) & 0x9d2c5680U;
    seed ^= (seed << 15) & 0xefc60000U;
    seed ^= (seed >> 18);    
    
    UPDATE_BLUE_SLICED_PRNG(s0, texture_R0);

    seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK);
    seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
    rand[k0*d_threads + rngIndex] = seed;
  
    k0 = SP(k0 + 1, N_MT);
    k1 = SP(k1 + 1, N_MT);
    k2 = SP(k2 + 1, N_MT);
    
    seed ^= (seed >> 11);
    seed ^= (seed << 7) & 0x9d2c5680U;
    seed ^= (seed << 15) & 0xefc60000U;
    seed ^= (seed >> 18);    

    UPDATE_BLUE_SLICED_PRNG(s1, texture_R1);

    seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK);
    seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
    rand[k0*d_threads + rngIndex] = seed;
  
    k0 = SP(k0 + 1, N_MT);
    k1 = SP(k1 + 1, N_MT);
    k2 = SP(k2 + 1, N_MT);
    
    seed ^= (seed >> 11);
    seed ^= (seed << 7) & 0x9d2c5680U;
    seed ^= (seed << 15) & 0xefc60000U;
    seed ^= (seed >> 18);    

    UPDATE_BLUE_SLICED_PRNG(s2, texture_R2);

    seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK);
    seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
    rand[k0*d_threads + rngIndex] = seed;
  
    k0 = SP(k0 + 1, N_MT);
    k1 = SP(k1 + 1, N_MT);
    k2 = SP(k2 + 1, N_MT);
    
    seed ^= (seed >> 11);
    seed ^= (seed << 7) & 0x9d2c5680U;
    seed ^= (seed << 15) & 0xefc60000U;
    seed ^= (seed >> 18);    

    UPDATE_BLUE_SLICED_PRNG(s3, texture_R3);

    count += 4;
  }

  if(count < spins && count != 0){
    for(int i = count; i < spins; i++){
      RNGT seed = (rand[k0*d_threads + rngIndex] & UPPER_MASK)|(rand[k1*d_threads + rngIndex] & LOWER_MASK);
      seed = rand[k2*d_threads + rngIndex]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
      rand[k0*d_threads + rngIndex] = seed;
      
      k0 = SP(k0 + 1, N_MT);
      k1 = SP(k1 + 1, N_MT);
      k2 = SP(k2 + 1, N_MT);
    }
  }

  return;
}

__global__ void MCStepBlues_Sliced_PR(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *ira, int ip0, int ip1, int ip2, int ip3, int spins){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = kk + blockIdx.y*d_loopStride;
  
  int count = 0;
  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    int x = k%d_L;
    int y = (k/d_L)%d_L;
    int z = k/d_A;
    
    int spz = i + (SP(z + 1, d_hL) - z)*d_A;    
    int spy = i + (SP(y + 1, d_L) - y)*d_L;
    int smy = spz + (SM(y - 1, d_L) - y)*d_L;
    int smx = spy - x + SM(x - 1, d_L);
    int spx = smy - x + SP(x + 1, d_L);

    RNGT seed = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];
    ira[ip0*d_threads + rngIndex] = seed;
    seed ^= ira[ip3*d_threads + rngIndex];

    ip0 = ((ip0 + 1)&PRWHEELMASK);
    ip1 = ((ip1 + 1)&PRWHEELMASK);
    ip2 = ((ip2 + 1)&PRWHEELMASK);
    ip3 = ((ip3 + 1)&PRWHEELMASK);
    
    UPDATE_BLUE_SLICED_PRNG(s0, texture_R0);

    seed = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];
    ira[ip0*d_threads + rngIndex] = seed;
    seed ^= ira[ip3*d_threads + rngIndex];

    ip0 = ((ip0 + 1)&PRWHEELMASK);
    ip1 = ((ip1 + 1)&PRWHEELMASK);
    ip2 = ((ip2 + 1)&PRWHEELMASK);
    ip3 = ((ip3 + 1)&PRWHEELMASK);

    UPDATE_BLUE_SLICED_PRNG(s1, texture_R1);

    seed = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];
    ira[ip0*d_threads + rngIndex] = seed;
    seed ^= ira[ip3*d_threads + rngIndex];

    ip0 = ((ip0 + 1)&PRWHEELMASK);
    ip1 = ((ip1 + 1)&PRWHEELMASK);
    ip2 = ((ip2 + 1)&PRWHEELMASK);
    ip3 = ((ip3 + 1)&PRWHEELMASK);

    UPDATE_BLUE_SLICED_PRNG(s2, texture_R2);

    seed = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];
    ira[ip0*d_threads + rngIndex] = seed;
    seed ^= ira[ip3*d_threads + rngIndex];

    ip0 = ((ip0 + 1)&PRWHEELMASK);
    ip1 = ((ip1 + 1)&PRWHEELMASK);
    ip2 = ((ip2 + 1)&PRWHEELMASK);
    ip3 = ((ip3 + 1)&PRWHEELMASK);

    UPDATE_BLUE_SLICED_PRNG(s3, texture_R3);

    count += 4;
  }


  if(count < spins && count > 0){
    for(int i = count; i < spins; i++){
      ira[ip0*d_threads + rngIndex] = ira[ip1*d_threads + rngIndex] + ira[ip2*d_threads + rngIndex];

      ip0 = ((ip0 + 1)&PRWHEELMASK);
      ip1 = ((ip1 + 1)&PRWHEELMASK);
      ip2 = ((ip2 + 1)&PRWHEELMASK);
      ip3 = ((ip3 + 1)&PRWHEELMASK);
    }
  }

  return;
}

__global__ void MCStepBlues_Sliced_MTGP32(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, curandStateMtgp32 *state, int spins){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = blockIdx.x + gridDim.x*blockIdx.y; 

  int count = 0;
  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    int x = k%d_L;
    int y = (k/d_L)%d_L;
    int z = k/d_A;
    
    int spz = i + (SP(z + 1, d_hL) - z)*d_A;    
    int spy = i + (SP(y + 1, d_L) - y)*d_L;
    int smy = spz + (SM(y - 1, d_L) - y)*d_L;
    int smx = spy - x + SM(x - 1, d_L);
    int spx = smy - x + SP(x + 1, d_L);

    RNGT seed = curand(&state[rngIndex]);
    
    UPDATE_BLUE_SLICED_PRNG(s0, texture_R0);

    seed = curand(&state[rngIndex]);

    UPDATE_BLUE_SLICED_PRNG(s1, texture_R1);

    seed = curand(&state[rngIndex]);

    UPDATE_BLUE_SLICED_PRNG(s2, texture_R2);

    seed = curand(&state[rngIndex]);

    UPDATE_BLUE_SLICED_PRNG(s3, texture_R3);

    count += 4;
  }

  if(count < spins && count > 0){
    for(int i = count; i < spins; i++){
      curand(&state[rngIndex]);
    }
  }

  return;
}

__global__ void MCStepBlues_Sliced_XORWOW(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, curandState *state){

  const int off = blockIdx.y*d_hV;
  const int kk = blockIdx.x*blockDim.x + threadIdx.x;
  const int rngIndex = kk + blockIdx.y*d_loopStride; 
  curandState localPrng = state[rngIndex];

  for(int k = kk; k < d_hV; k += d_loopStride){ 

    int i = k + off;

    int x = k%d_L;
    int y = (k/d_L)%d_L;
    int z = k/d_A;
    
    int spz = i + (SP(z + 1, d_hL) - z)*d_A;    
    int spy = i + (SP(y + 1, d_L) - y)*d_L;
    int smy = spz + (SM(y - 1, d_L) - y)*d_L;
    int smx = spy - x + SM(x - 1, d_L);
    int spx = smy - x + SP(x + 1, d_L);

    RNGT seed = curand(&localPrng);
    
    UPDATE_BLUE_SLICED_PRNG(s0, texture_R0);

    seed = curand(&localPrng);

    UPDATE_BLUE_SLICED_PRNG(s1, texture_R1);

    seed = curand(&localPrng);

    UPDATE_BLUE_SLICED_PRNG(s2, texture_R2);

    seed = curand(&localPrng);

    UPDATE_BLUE_SLICED_PRNG(s3, texture_R3);

  }

  state[rngIndex] = localPrng;

  return;
}

__global__ void MCStepReds_SlicedGrid(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *rand){

  const int i = threadIdx.x + threadIdx.y*blockDim.x + blockIdx.x*d_block_xStride + blockIdx.y*d_block_yStride + blockIdx.z*d_hV;
  const int y = threadIdx.y + blockIdx.x*blockDim.y;

  RNGT seed = rand[i];

  int smz = i + (SMM(blockIdx.y, d_hL) - blockIdx.y)*d_A;
  int spy = smz + (SP(y + 1, d_L) - y)*d_L; 
  int smy = i + (SMM(y, d_L) - y)*d_L;   
  int smx = spy - threadIdx.x + SMM(threadIdx.x, d_L);
  int spx = smy - threadIdx.x + SP(threadIdx.x + 1, d_L);
  
  UPDATE_RED_SLICED(s0, texture_B0);
  UPDATE_RED_SLICED(s1, texture_B1);
  UPDATE_RED_SLICED(s2, texture_B2);
  UPDATE_RED_SLICED(s3, texture_B3);
  
  rand[i] = seed;

  return;
}

__global__ void MCStepReds_SlicedGrid_PR(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *ira, int ip0, int ip1, int ip2, int ip3, int spins){

  const int i = threadIdx.x + threadIdx.y*blockDim.x + blockIdx.x*d_block_xStride + blockIdx.y*d_block_yStride + blockIdx.z*d_hV;
  const int y = threadIdx.y + blockIdx.x*blockDim.y;

  int smz = i + (SMM(blockIdx.y, d_hL) - blockIdx.y)*d_A;
  int spy = smz + (SP(y + 1, d_L) - y)*d_L; 
  int smy = i + (SMM(y, d_L) - y)*d_L;   
  int smx = spy - threadIdx.x + SMM(threadIdx.x, d_L);
  int spx = smy - threadIdx.x + SP(threadIdx.x + 1, d_L);

  RNGT seed = ira[ip1*d_threads + i] + ira[ip2*d_threads + i];
  ira[ip0*d_threads + i] = seed;
  seed ^= ira[ip3*d_threads + i];
  
  ip0 = ((ip0 + 1)&PRWHEELMASK);
  ip1 = ((ip1 + 1)&PRWHEELMASK);
  ip2 = ((ip2 + 1)&PRWHEELMASK);
  ip3 = ((ip3 + 1)&PRWHEELMASK);
  
  UPDATE_RED_SLICED_PRNG(s0, texture_B0);

  seed = ira[ip1*d_threads + i] + ira[ip2*d_threads + i];
  ira[ip0*d_threads + i] = seed;
  seed ^= ira[ip3*d_threads + i];   
  
  ip0 = ((ip0 + 1)&PRWHEELMASK);
  ip1 = ((ip1 + 1)&PRWHEELMASK);
  ip2 = ((ip2 + 1)&PRWHEELMASK);
  ip3 = ((ip3 + 1)&PRWHEELMASK);

  UPDATE_RED_SLICED_PRNG(s1, texture_B1);

  seed = ira[ip1*d_threads + i] + ira[ip2*d_threads + i];
  ira[ip0*d_threads + i] = seed;
  seed ^= ira[ip3*d_threads + i];   
  
  ip0 = ((ip0 + 1)&PRWHEELMASK);
  ip1 = ((ip1 + 1)&PRWHEELMASK);
  ip2 = ((ip2 + 1)&PRWHEELMASK);
  ip3 = ((ip3 + 1)&PRWHEELMASK);

  UPDATE_RED_SLICED_PRNG(s2, texture_B2);

  seed = ira[ip1*d_threads + i] + ira[ip2*d_threads + i];
  ira[ip0*d_threads + i] = seed;
  seed ^= ira[ip3*d_threads + i];   
  
  ip0 = ((ip0 + 1)&PRWHEELMASK);
  ip1 = ((ip1 + 1)&PRWHEELMASK);
  ip2 = ((ip2 + 1)&PRWHEELMASK);
  ip3 = ((ip3 + 1)&PRWHEELMASK);

  UPDATE_RED_SLICED_PRNG(s3, texture_B3);

  return;
}

__global__ void MCStepReds_SlicedGrid_MT19937(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *rand, int k0, int k1, int k2, int spins){

  const int i = threadIdx.x + threadIdx.y*blockDim.x + blockIdx.x*d_block_xStride + blockIdx.y*d_block_yStride + blockIdx.z*d_hV;
  const int y = threadIdx.y + blockIdx.x*blockDim.y;
  
  int smz = i + (SMM(blockIdx.y, d_hL) - blockIdx.y)*d_A;
  int spy = smz + (SP(y + 1, d_L) - y)*d_L; 
  int smy = i + (SMM(y, d_L) - y)*d_L;   
  int smx = spy - threadIdx.x + SMM(threadIdx.x, d_L);
  int spx = smy - threadIdx.x + SP(threadIdx.x + 1, d_L);

  RNGT seed = (rand[k0*d_threads + i] & UPPER_MASK)|(rand[k1*d_threads + i] & LOWER_MASK); 
  seed = rand[k2*d_threads + i]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
  rand[k0*d_threads + i] = seed;
  
  k0 = SP(k0 + 1, N_MT);
  k1 = SP(k1 + 1, N_MT);
  k2 = SP(k2 + 1, N_MT);
  
  seed ^= (seed >> 11);
  seed ^= (seed << 7) & 0x9d2c5680U;
  seed ^= (seed << 15) & 0xefc60000U;
  seed ^= (seed >> 18);
  
  UPDATE_RED_SLICED(s0, texture_B0);

  seed = (rand[k0*d_threads + i] & UPPER_MASK)|(rand[k1*d_threads + i] & LOWER_MASK); 
  seed = rand[k2*d_threads + i]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
  rand[k0*d_threads + i] = seed;
  
  k0 = SP(k0 + 1, N_MT);
  k1 = SP(k1 + 1, N_MT);
  k2 = SP(k2 + 1, N_MT);
  
  seed ^= (seed >> 11);
  seed ^= (seed << 7) & 0x9d2c5680U;
  seed ^= (seed << 15) & 0xefc60000U;
  seed ^= (seed >> 18);

  UPDATE_RED_SLICED(s1, texture_B1);

  seed = (rand[k0*d_threads + i] & UPPER_MASK)|(rand[k1*d_threads + i] & LOWER_MASK); 
  seed = rand[k2*d_threads + i]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
  rand[k0*d_threads + i] = seed;
  
  k0 = SP(k0 + 1, N_MT);
  k1 = SP(k1 + 1, N_MT);
  k2 = SP(k2 + 1, N_MT);
  
  seed ^= (seed >> 11);
  seed ^= (seed << 7) & 0x9d2c5680U;
  seed ^= (seed << 15) & 0xefc60000U;
  seed ^= (seed >> 18);

  UPDATE_RED_SLICED(s2, texture_B2);

  seed = (rand[k0*d_threads + i] & UPPER_MASK)|(rand[k1*d_threads + i] & LOWER_MASK); 
  seed = rand[k2*d_threads + i]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
  rand[k0*d_threads + i] = seed;
  
  k0 = SP(k0 + 1, N_MT);
  k1 = SP(k1 + 1, N_MT);
  k2 = SP(k2 + 1, N_MT);
  
  seed ^= (seed >> 11);
  seed ^= (seed << 7) & 0x9d2c5680U;
  seed ^= (seed << 15) & 0xefc60000U;
  seed ^= (seed >> 18);

  UPDATE_RED_SLICED(s3, texture_B3);
  
  return;
}


__global__ void MCStepBlues_SlicedGrid(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *rand){

  const int i = threadIdx.x + threadIdx.y*blockDim.x + blockIdx.x*d_block_xStride + blockIdx.y*d_block_yStride + blockIdx.z*d_hV;
  const int y = threadIdx.y + blockIdx.x*blockDim.y;

  RNGT seed = rand[i];

  int spz = i + (SP(blockIdx.y + 1, d_hL) - blockIdx.y)*d_A;  
  int spy = i + (SP(y + 1, d_L) - y)*d_L; 
  int smy = spz + (SMM(y, d_L) - y)*d_L;   
  int smx = spy - threadIdx.x + SMM(threadIdx.x, d_L);
  int spx = smy - threadIdx.x + SP(threadIdx.x + 1, d_L);
  
  UPDATE_BLUE_SLICED(s0, texture_R0);
  UPDATE_BLUE_SLICED(s1, texture_R1);
  UPDATE_BLUE_SLICED(s2, texture_R2);
  UPDATE_BLUE_SLICED(s3, texture_R3);
  
  rand[i] = seed;

  return;
}

__global__ void MCStepBlues_SlicedGrid_PR(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *ira, int ip0, int ip1, int ip2, int ip3, int spins){

  const int i = threadIdx.x + threadIdx.y*blockDim.x + blockIdx.x*d_block_xStride + blockIdx.y*d_block_yStride + blockIdx.z*d_hV;
  const int y = threadIdx.y + blockIdx.x*blockDim.y;

  int spz = i + (SP(blockIdx.y + 1, d_hL) - blockIdx.y)*d_A;  
  int spy = i + (SP(y + 1, d_L) - y)*d_L; 
  int smy = spz + (SMM(y, d_L) - y)*d_L;   
  int smx = spy - threadIdx.x + SMM(threadIdx.x, d_L);
  int spx = smy - threadIdx.x + SP(threadIdx.x + 1, d_L);

  RNGT seed = ira[ip1*d_threads + i] + ira[ip2*d_threads + i];
  ira[ip0*d_threads + i] = seed;
  seed ^= ira[ip3*d_threads + i];
  
  ip0 = ((ip0 + 1)&PRWHEELMASK);
  ip1 = ((ip1 + 1)&PRWHEELMASK);
  ip2 = ((ip2 + 1)&PRWHEELMASK);
  ip3 = ((ip3 + 1)&PRWHEELMASK);
  
  UPDATE_BLUE_SLICED_PRNG(s0, texture_R0);

  seed = ira[ip1*d_threads + i] + ira[ip2*d_threads + i];
  ira[ip0*d_threads + i] = seed;
  seed ^= ira[ip3*d_threads + i];   
  
  ip0 = ((ip0 + 1)&PRWHEELMASK);
  ip1 = ((ip1 + 1)&PRWHEELMASK);
  ip2 = ((ip2 + 1)&PRWHEELMASK);
  ip3 = ((ip3 + 1)&PRWHEELMASK);

  UPDATE_BLUE_SLICED_PRNG(s1, texture_R1);

  seed = ira[ip1*d_threads + i] + ira[ip2*d_threads + i];
  ira[ip0*d_threads + i] = seed;
  seed ^= ira[ip3*d_threads + i];   
  
  ip0 = ((ip0 + 1)&PRWHEELMASK);
  ip1 = ((ip1 + 1)&PRWHEELMASK);
  ip2 = ((ip2 + 1)&PRWHEELMASK);
  ip3 = ((ip3 + 1)&PRWHEELMASK);

  UPDATE_BLUE_SLICED_PRNG(s2, texture_R2);

  seed = ira[ip1*d_threads + i] + ira[ip2*d_threads + i];
  ira[ip0*d_threads + i] = seed;
  seed ^= ira[ip3*d_threads + i];   
  
  ip0 = ((ip0 + 1)&PRWHEELMASK);
  ip1 = ((ip1 + 1)&PRWHEELMASK);
  ip2 = ((ip2 + 1)&PRWHEELMASK);
  ip3 = ((ip3 + 1)&PRWHEELMASK);

  UPDATE_BLUE_SLICED_PRNG(s3, texture_R3);

  return;
}

__global__ void MCStepBlues_SlicedGrid_MT19937(__restrict MSC *s0, __restrict MSC *s1, __restrict MSC *s2, __restrict MSC *s3, __restrict RNGT *rand, int k0, int k1, int k2, int spins){

  const int i = threadIdx.x + threadIdx.y*blockDim.x + blockIdx.x*d_block_xStride + blockIdx.y*d_block_yStride + blockIdx.z*d_hV;
  const int y = threadIdx.y + blockIdx.x*blockDim.y;

  int spz = i + (SP(blockIdx.y + 1, d_hL) - blockIdx.y)*d_A;  
  int spy = i + (SP(y + 1, d_L) - y)*d_L; 
  int smy = spz + (SMM(y, d_L) - y)*d_L;   
  int smx = spy - threadIdx.x + SMM(threadIdx.x, d_L);
  int spx = smy - threadIdx.x + SP(threadIdx.x + 1, d_L);

  RNGT seed = (rand[k0*d_threads + i] & UPPER_MASK)|(rand[k1*d_threads + i] & LOWER_MASK); 
  seed = rand[k2*d_threads + i]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
  rand[k0*d_threads + i] = seed;
  
  k0 = SP(k0 + 1, N_MT);
  k1 = SP(k1 + 1, N_MT);
  k2 = SP(k2 + 1, N_MT);
  
  seed ^= (seed >> 11);
  seed ^= (seed << 7) & 0x9d2c5680U;
  seed ^= (seed << 15) & 0xefc60000U;
  seed ^= (seed >> 18);
  
  UPDATE_BLUE_SLICED(s0, texture_R0);

  seed = (rand[k0*d_threads + i] & UPPER_MASK)|(rand[k1*d_threads + i] & LOWER_MASK); 
  seed = rand[k2*d_threads + i]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
  rand[k0*d_threads + i] = seed;
  
  k0 = SP(k0 + 1, N_MT);
  k1 = SP(k1 + 1, N_MT);
  k2 = SP(k2 + 1, N_MT);
  
  seed ^= (seed >> 11);
  seed ^= (seed << 7) & 0x9d2c5680U;
  seed ^= (seed << 15) & 0xefc60000U;
  seed ^= (seed >> 18);

  UPDATE_BLUE_SLICED(s1, texture_R1);

  seed = (rand[k0*d_threads + i] & UPPER_MASK)|(rand[k1*d_threads + i] & LOWER_MASK); 
  seed = rand[k2*d_threads + i]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
  rand[k0*d_threads + i] = seed;
  
  k0 = SP(k0 + 1, N_MT);
  k1 = SP(k1 + 1, N_MT);
  k2 = SP(k2 + 1, N_MT);
  
  seed ^= (seed >> 11);
  seed ^= (seed << 7) & 0x9d2c5680U;
  seed ^= (seed << 15) & 0xefc60000U;
  seed ^= (seed >> 18);

  UPDATE_BLUE_SLICED(s2, texture_R2);

  seed = (rand[k0*d_threads + i] & UPPER_MASK)|(rand[k1*d_threads + i] & LOWER_MASK); 
  seed = rand[k2*d_threads + i]^(seed >> 1)^((-(seed & 1U)) & MATRIX_A);
  rand[k0*d_threads + i] = seed;
  
  k0 = SP(k0 + 1, N_MT);
  k1 = SP(k1 + 1, N_MT);
  k2 = SP(k2 + 1, N_MT);
  
  seed ^= (seed >> 11);
  seed ^= (seed << 7) & 0x9d2c5680U;
  seed ^= (seed << 15) & 0xefc60000U;
  seed ^= (seed >> 18);

  UPDATE_BLUE_SLICED(s3, texture_R3);
  
  return;
}

// END KERNELS

__global__ void testParisiRapuanoKernel(__restrict RNGT *ira, int steps,
					int ip, int ip1, int ip2, int ip3){
  const int off = blockIdx.y*d_hV;
  const int k = blockIdx.x*blockDim.x + threadIdx.x;
  const int i = k + off;

  for(int j=0; j<steps; j++){
    PR32(ira, ip + i, ip1 + i, ip2 + i, ip3 + i);
    ip = (ip + d_threads)%d_iraLen;
    ip1 = (ip1 + d_threads)%d_iraLen;
    ip2 = (ip2 + d_threads)%d_iraLen;
    ip3 = (ip3 + d_threads)%d_iraLen;
  }

  return;
}

__global__ void testParisiRapuanoKernelMod(__restrict RNGT *ira, int steps,
					int ip, int ip1, int ip2, int ip3){
  const int off = blockIdx.y*d_hV;
  const int k = blockIdx.x*blockDim.x + threadIdx.x;
  const int i = k + off;

  for(int j=0; j<steps; j++){
    PR32(ira, ip*d_threads + i, ip1*d_threads + i, ip2*d_threads + i, ip3*d_threads + i);
    ip = ((ip + 1)&PRWHEELMASK);
    ip1 = ((ip1 + 1)&PRWHEELMASK);
    ip2 = ((ip2 + 1)&PRWHEELMASK);
    ip3 = ((ip3 + 1)&PRWHEELMASK);
  }

  return;
}

__global__ void testMinStdKernel(__restrict RNGT *rng, int steps){
  const int off = blockIdx.y*d_hV;
  const int k = blockIdx.x*blockDim.x + threadIdx.x;
  const int i = k + off;

  RNGT seed = rng[i];
  for(int j=0; j<steps; j++){
    GPURAND(seed);
  }
  rng[i] = seed;

  return;
}

extern int testParisiRapuano(sysEA3D_t *sys){
  dim3 block(sys->blockSize,1,1);
  dim3 grid(sys->V/2/block.x, sys->sysN, 1);
  int threads = block.x*block.y*block.z*grid.x*grid.y*grid.z;
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_threads, &threads, sizeof(int)) );
  int iraLen = threads*PRWHEEL;
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_iraLen, &iraLen, sizeof(int)) );

  int steps = 2<<20;

  PR_RNGArray_t *ira = NULL;
  ira = PR_RNGArrayInit(iraLen, 0);
  RNGT *iraHost = (RNGT *)malloc(iraLen*sizeof(RNGT));

  //int ip = 61*threads, ip1 = (61-24)*threads, ip2 = (61-55)*threads, ip3 = 0;
  int ip00 = 61, ip01 = (61-24), ip02 = (61-55), ip03 = 0;

  /*printf("GPU start\n");
  testParisiRapuanoKernel<<<grid, block>>>(ira->d_rngSeeds, steps, ip, ip1, ip2, ip3);
  MYCUDA_ERROR( cudaDeviceSynchronize() );*/

  printf("GPU start\n");
  testParisiRapuanoKernelMod<<<grid, block>>>(ira->d_rngSeeds, steps, ip00, ip01, ip02, ip03);
  MYCUDA_ERROR( cudaDeviceSynchronize() );

  printf("CPU start\n");
  PR32EvolveMod(ira, steps, ip00, ip01, ip02, ip03);

  for(int i=0; i<iraLen; i++) iraHost[i] = ira->h_rngSeeds[i];
  D2HPR_RNGArray(ira);
  for(int i=0; i<iraLen; i++){
    if(iraHost[i] != ira->h_rngSeeds[i]){
      printf("%d %d %d\n", iraHost[i], ira->h_rngSeeds[i], i);
      getchar();
    }
  }

  return 0;
}



extern int SimulationFullRecoverTestCongruential(sysEA3D_t *sys){
  int hV = sys->V/2, hA = sys->L*sys->L/2;
  dim3 block(sys->blockSize,1,1);
  dim3 grid(sys->V/sys->k/block.x, sys->sysN, 1);
  int threads = block.x*block.y*block.z*grid.x*grid.y*grid.z;
  int rngOffset = grid.x*block.x;
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_threads, &threads, sizeof(int)) );
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_rngOffset, &rngOffset, sizeof(int)) );

  printf("block.x: %d, grid.x: %d, threads: %d, spins: %d\n", 
	 block.x, grid.x, threads, hV/(block.x*grid.x));

  CRNGArray_t *rng = CRNGArrayInit(threads);

  if(rng == NULL){
    fprintf(stderr, "\nNo Alloc rng Simulation\n");
    return 1;
  }

  if(sys->dumpF == 1 && sys->recoverF != 1){
    // RNG Sequence transform
    if(sys->kernelChoice == 1 || sys->kernelChoice == 2){
      for(int i=0; i<sys->sysN; i++){
	int offset = i*hV;
	arrayTransformRNGT(rng->h_rngSeeds + offset, sys->realI_R, hV);
      }
    }

    if(DumpCRNGArray(rng, 0)){
      fprintf(stderr, "\nError DumpCRNGArray Simulation\n");
      return 1;
    }
  }

  if(sys->recoverF == 1){
    if(ReadCRNGArray(rng, 0)){
      fprintf(stderr, "\nError DumpCRNGArray Simulation\n");
      return 1;
    }

    // RNG Sequence transform
    if(sys->kernelChoice == 1 || sys->kernelChoice == 2){
      for(int i=0; i<sys->sysN; i++){
	int offset = i*hV;
	arrayTransformRNGT(rng->h_rngSeeds + offset, sys->myI_R, hV);
      }
    }

    H2DCRNGArray(rng);

    testMinStdKernel<<<grid, block>>>(rng->d_rngSeeds, 8*sys->startValue);
    MYCUDA_ERROR( cudaDeviceSynchronize() );

    D2HCRNGArray(rng);

    if(sys->kernelChoice == 1 || sys->kernelChoice == 2){
      for(int i=0; i<sys->sysN; i++){
	int offset = i*hV;
	arrayTransformRNGT(rng->h_rngSeeds + offset, sys->realI_R, hV);
      }
    }

    if(DumpCRNGArray(rng, (unsigned)-1)){
      fprintf(stderr, "\nError DumpCRNGArray Simulation\n");
      return 1;
    }
  }


  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Bitwise, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Bitwise, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Standard, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Standard, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_StandardGrid, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_StandardGrid, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Sliced, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Sliced, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_SlicedGrid, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_SlicedGrid, cudaFuncCachePreferL1) );

  int dumpIndex = 0;
  int maxStep = sys->dumpTimes[sys->dumpN - 1];

  int A = sys->L*sys->L;
  dim3 blockG(sys->L, sys->k, 1);
  dim3 gridG(A/(blockG.x*blockG.y), sys->L/2, sys->sysN);

  //int hA = A/2;
  dim3 blockS(sys->L/2, sys->k, 1);
  dim3 gridS(hA/(blockS.x*blockS.y), sys->L, sys->sysN);

  int threadsG = blockG.x*blockG.y*blockG.z*gridG.x*gridG.y*gridG.z;
  int threadsS = blockS.x*blockS.y*blockS.z*gridS.x*gridS.y*gridS.z;

  printf("%d %d %d\n", threadsG, threadsS, gridS.x);

  int h_block_xStride = blockG.x*blockG.y;
  int h_block_yStride = gridG.x*h_block_xStride;
  int h_block_xStrideS = blockS.x*blockS.y;
  int h_block_yStrideS = gridS.x*h_block_xStrideS;

  int h_loopStride = block.x*grid.x;

  MYCUDA_ERROR( cudaMemcpyToSymbol(d_block_xStride, &h_block_xStride, sizeof(int)) );
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_block_yStride, &h_block_yStride, sizeof(int)) );
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_block_xStrideS, &h_block_xStrideS, sizeof(int)) );
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_block_yStrideS, &h_block_yStrideS, sizeof(int)) );
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_loopStride, &h_loopStride, sizeof(int)) );

  for(int i=sys->startValue + 1; i<=maxStep; i++){
    
    if(sys->kernelChoice == 0){
      MCStepReds_Bitwise<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
					    sys->d_reds2, sys->d_reds3,
					    rng->d_rngSeeds);
      MYCUDA_ERROR( cudaDeviceSynchronize() );

      MCStepBlues_Bitwise<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
					     sys->d_blues2, sys->d_blues3,
					     rng->d_rngSeeds);
      MYCUDA_ERROR( cudaDeviceSynchronize() );
    }

    if(sys->kernelChoice == 1){
      MCStepReds_Sliced<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
					   sys->d_reds2, sys->d_reds3,
					   rng->d_rngSeeds);
      MYCUDA_ERROR( cudaDeviceSynchronize() );

      // RNG sequence transformation
      D2HCRNGArray(rng);
      for(int i=0; i<sys->sysN; i++){
	int offset = i*hV;
	arrayTransformRNGT(rng->h_rngSeeds + offset, sys->myI_R2B, hV);
      }
      H2DCRNGArray(rng);

      MCStepBlues_Sliced<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
					    sys->d_blues2, sys->d_blues3,
					    rng->d_rngSeeds);
      MYCUDA_ERROR( cudaDeviceSynchronize() );

      // RNG sequence transformation
      D2HCRNGArray(rng);
      for(int i=0; i<sys->sysN; i++){
	int offset = i*hV;
	arrayTransformRNGT(rng->h_rngSeeds + offset, sys->myI_B2R, hV);
      }
      H2DCRNGArray(rng);
    }

    if(sys->kernelChoice == 2){
      MCStepReds_SlicedGrid<<< gridG, blockG >>>(sys->d_reds0, sys->d_reds1,
						 sys->d_reds2, sys->d_reds3,
						 rng->d_rngSeeds);
      MYCUDA_ERROR( cudaDeviceSynchronize() );

      // RNG sequence transformation
      D2HCRNGArray(rng);
      for(int i=0; i<sys->sysN; i++){
	int offset = i*hV;
	arrayTransformRNGT(rng->h_rngSeeds + offset, sys->myI_R2B, hV);
      }
      H2DCRNGArray(rng);

      MCStepBlues_SlicedGrid<<< gridG, blockG >>>(sys->d_blues0, sys->d_blues1,
						  sys->d_blues2, sys->d_blues3,
						  rng->d_rngSeeds);
      MYCUDA_ERROR( cudaDeviceSynchronize() );

      // RNG sequence transformation
      D2HCRNGArray(rng);
      for(int i=0; i<sys->sysN; i++){
	int offset = i*hV;
	arrayTransformRNGT(rng->h_rngSeeds + offset, sys->myI_B2R, hV);
      }
      H2DCRNGArray(rng);
    }

    if(sys->kernelChoice == 3){
      MCStepReds_Standard<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
					     sys->d_reds2, sys->d_reds3,
					     rng->d_rngSeeds);
      MYCUDA_ERROR( cudaDeviceSynchronize() );

      MCStepBlues_Standard<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
					      sys->d_blues2, sys->d_blues3,
					      rng->d_rngSeeds);
      MYCUDA_ERROR( cudaDeviceSynchronize() );
    }

    if(sys->kernelChoice == 4){
      MCStepReds_StandardGrid<<< gridS, blockS >>>(sys->d_reds0, sys->d_reds1,
						   sys->d_reds2, sys->d_reds3,
						   rng->d_rngSeeds);
      MYCUDA_ERROR( cudaDeviceSynchronize() );

      MCStepBlues_StandardGrid<<< gridS, blockS >>>(sys->d_blues0, sys->d_blues1,
						    sys->d_blues2, sys->d_blues3,
						    rng->d_rngSeeds);
      MYCUDA_ERROR( cudaDeviceSynchronize() );
    }
        
    if(sys->dumpTimes[dumpIndex] == i){
      if(D2HSpins(sys)){
	fprintf(stderr,"\nError in D2HSpins time %d in Simulation\n", i);
	return 1;
      }
      
      if( sys->kernelChoice == 1 || sys->kernelChoice == 2 ){
	SpinsTransform(sys);
      }
      
      if(sys->dumpF){
	if(SpinsDump(sys, i)){
	  fprintf(stderr,"\nError in SpinsDump time %d Simulation\n", i);
	  return 1;
	}
      }

      dumpIndex++;
    }
  }

  D2HCRNGArray(rng);
  if(sys->kernelChoice == 1 || sys->kernelChoice == 2 ){
    for(int i=0; i<sys->sysN; i++){
      int offset = i*hV;
      arrayTransformRNGT(rng->h_rngSeeds + offset, sys->realI_R, hV);
    }
  }

  if(sys->recoverF != 1){    
    if(DumpCRNGArray(rng, maxStep)){
      fprintf(stderr, "\nError DumpCRNGArray Simulation\n");
      return 1;
    }
  }

  if(sys->recoverF == 1){    
    if(DumpCRNGArray(rng, maxStep + 1)){
      fprintf(stderr, "\nError DumpCRNGArray Simulation\n");
      return 1;
    }
  }

  return 0;
}

extern int benchmark(int argc, char **argv){
  if(argc != 9){
    fprintf(stderr, "usage: %s <gpu> <L> <sysN> <block> <samples> <steps> <memory> <spinsPerThread>\n\n", argv[0]);
    fprintf(stderr, "Systems' GPUs are:\n");
    int numberOfGPUs = 0;
    MYCUDA_ERROR( cudaGetDeviceCount(&numberOfGPUs) );
    for(int gpu = 0; gpu < numberOfGPUs; gpu++){
      cudaDeviceProp prop;
      MYCUDA_ERROR( cudaGetDeviceProperties(&prop, gpu) );
      printf("GPU %d: %s\n", gpu, prop.name);
    }
    return 1;
  }

  char fileName[1024];
  //struct timespec start, stop;

  sysEA3D_t *sys = (sysEA3D_t *)malloc(sizeof(sysEA3D_t));
  int initL, initSysN, initSamples, initSteps, initBlock, spinsPerThread;
  float memory = 0.;
  sscanf(argv[1], "%d", &sys->cardN);
  sscanf(argv[2], "%d", &initL);
  sscanf(argv[3], "%d", &initSysN);
  sscanf(argv[4], "%d", &initBlock);
  sscanf(argv[5], "%d", &initSamples);
  sscanf(argv[6], "%d", &initSteps);
  sscanf(argv[7], "%f", &memory);
  sscanf(argv[8], "%d", &spinsPerThread);
  
  if(initL == 0) initL = 8;
  if(initSamples == 0) initSamples = 100;
  if(initSteps == 0) initSteps = 50;

  sys->beta = 1.101;

  int numberOfGPUs = 0;
  cudaGetDeviceCount(&numberOfGPUs);
  if(sys->cardN >= numberOfGPUs){
    fprintf(stderr, "gpu %d >= gpuN %d\n", sys->cardN, numberOfGPUs);
    return 1;
  }

  MYCUDA_ERROR( cudaSetDevice(sys->cardN) );
  cudaDeviceProp prop;
  MYCUDA_ERROR( cudaGetDeviceProperties(&prop, sys->cardN) );
  printf("GPU %d: %s Compute Capability: %d.%d\n", sys->cardN, prop.name, prop.major, prop.minor);

  cudaEvent_t startEvent, stopEvent;
  MYCUDA_ERROR( cudaEventCreate(&startEvent) );
  MYCUDA_ERROR( cudaEventCreate(&stopEvent) );

  int maxThreadsPerBlock = 0;
  if(prop.major == 2) maxThreadsPerBlock = 512;
  if(prop.major == 3) maxThreadsPerBlock = 1024;

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Bitwise, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Bitwise, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Standard, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Standard, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Sliced, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Sliced, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_StandardGrid, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_StandardGrid, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_SlicedGrid, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_SlicedGrid, cudaFuncCachePreferL1) );


  for(int L = initL; L <= 256; L += 2){

    int hV = L*L*L/2;
    size_t NJ_Size = 6*hV*sizeof(MSC), NS_Size = 4*2*hV*sizeof(MSC), RNG_Size = hV*sizeof(RNGT);
    size_t devFree = 0, devTotal = 0;
    MYCUDA_ERROR( cudaMemGetInfo(&devFree, &devTotal) );
    // here to decide the percentage of device memory to use. 
    int MaxSysN = (int)(memory*(devFree/(NJ_Size + NS_Size + RNG_Size)));
    printf("\n\nL = %d, Free memory on device: %lu, MaxSysN: %d\n", L, devFree, MaxSysN);

    if(initSysN == 0) initSysN = 1;

    for(int sysN = initSysN; sysN <= MaxSysN; sysN *= 2){
      sys->L = L;
      sys->V = sys->L*sys->L*sys->L;
      sys->sysN = sysN;
      sys->NJ = sys->sysN*sys->V/2;
      sys->NS = sys->sysN*sys->V/2;
      
      if(masksInit(sys)){
	fprintf(stderr,"\nError in masksInit\n");
	return 1;
      }
      if(MCWeightsInit(sys)){
	fprintf(stderr,"\nError in MCWeightsInit\n");
	return 1;
      }
      if(dMemInit(sys)){
	fprintf(stderr,"\nError in dMemInit\n");
	return 1;
      }
      if(hMemInit2(sys)){
	fprintf(stderr,"\nError in hMemInit\n");
	return 1;
      }
      if(H2DSysCopy(sys)){
	fprintf(stderr,"\nError in H2DSysCopy\n");
	return 1;
      }

      MYCUDA_ERROR( cudaMemGetInfo(&devFree, &devTotal) );
      printf("\n\n---\tL = %d, sysN = %d, free memory after allocation: %lu\t---\n\n",
	     L, sysN, devFree);

      ////////////////////////////////////////
      // NO-GRID KERNELS
      ///////////////////////////////////////

      int blockInit = 0;

      if(initBlock == 0){	
	blockInit = sys->V/2;
	//printf("blockInitI: %d\n", blockInit);
	while(blockInit > maxThreadsPerBlock) blockInit /= 2;
	//printf("blockInitF: %d\n", blockInit);
      }else blockInit = initBlock;

      int count = 0, test = 1;
      for(sys->blockSize = blockInit; test; ){

	if(count > 0){
	  do{sys->blockSize--;}while((sys->V/2)%sys->blockSize != 0);
	}
	count++;

	//printf("blockSize: %d\n", sys->blockSize);

	if(sys->blockSize >= 32){
	  int kMax = sys->V/sys->blockSize/2;
	  if(kMax > spinsPerThread && spinsPerThread != 0) kMax = spinsPerThread;
	  //printf("blockSizeIF: %d\n", sys->blockSize);

	  for(int k = 0; k < kMax ;){
	    
	    float redMean = 0., redMean2 = 0., swap = 0.;
	    float blueMean = 0., blueMean2 = 0.;
	    float errorRed = 0., errorBlue = 0.;

	    do{k++;}while(kMax%k != 0);	    
	    
	    dim3 block(sys->blockSize,1,1);
	    dim3 grid(sys->V/block.x/(2*k), sys->sysN, 1);
	    int h_loopStride = block.x*grid.x;
	    MYCUDA_ERROR( cudaMemcpyToSymbol(d_loopStride, &h_loopStride, sizeof(int)) );
	    
	    printf("Measuring Times No-Grid L:%d block.x:%d grid.x:%d spins per thread:%d\n",
		   L, block.x, grid.x, k);
	    
	    int threads = block.x*block.y*block.z*grid.x*grid.y*grid.z;
	    MYCUDA_ERROR( cudaMemcpyToSymbol(d_threads, &threads, sizeof(int)) );
	    
	    CRNGArray_t *rng = CRNGArrayInit(threads);
	    
	    // KERNEL: Bitwise; Only L = 2^k are examined
	    if( ((1U<<findLog2(sys->L))^sys->L) == 0 ){
	      for(int sample = 0; sample < initSamples; sample++){
		
		MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
		
		for(int steps = 0; steps < initSteps; steps++){	  
		  MCStepReds_Bitwise<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
							sys->d_reds2, sys->d_reds3,
							rng->d_rngSeeds);
		  MYCUDA_ERROR( cudaDeviceSynchronize() );
		}
		
		MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
		MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
		MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
		redMean += (swap/(initSteps)); 
		redMean2 += (swap/(initSteps))*(swap/(initSteps));
		swap = 0.;
		
		MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
		
		for(int steps = 0; steps < initSteps; steps++){	  
		  MCStepBlues_Bitwise<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
							 sys->d_blues2, sys->d_blues3,
							 rng->d_rngSeeds);
		  MYCUDA_ERROR( cudaDeviceSynchronize() );
		}
		
		MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
		MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
		MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
		blueMean += (swap/(initSteps)); 
		blueMean2 += (swap/(initSteps))*(swap/(initSteps));
		swap = 0.;
	      }
	      
	      redMean /= initSamples;  redMean2 /= initSamples;      
	      blueMean /= initSamples; blueMean2 /= initSamples;
	      
	      errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
	      errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
	      
	      snprintf(fileName, sizeof(fileName), "GPU%d_bitwiseKernel", sys->cardN);
	      FILE *outputFile = fopen(fileName, "a+");
	      fprintf(outputFile, "%d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
		      L, sysN, k, block.x, grid.x,
		      redMean*1.0e-3, errorRed*1.0e-3, blueMean*1.0e-3, errorBlue*1.0e-3);
	      fflush(outputFile);
	      fclose(outputFile);
	    }
	    
	    // KERNEL: Standard
	    redMean = blueMean = redMean2 = blueMean2 = 0.;
	    for(int sample = 0; sample < initSamples; sample++){
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepReds_Standard<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
						       sys->d_reds2, sys->d_reds3,
						       rng->d_rngSeeds);
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      redMean += (swap/(initSteps)); 
	      redMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepBlues_Standard<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
							sys->d_blues2, sys->d_blues3,
							rng->d_rngSeeds);
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      blueMean += (swap/(initSteps)); 
	      blueMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	    }
	    
	    redMean /= initSamples;  redMean2 /= initSamples;      
	    blueMean /= initSamples; blueMean2 /= initSamples;
	    
	    errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
	    errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
	    
	    snprintf(fileName, sizeof(fileName), "GPU%d_standardKernel", sys->cardN);
	    FILE *outputFile = fopen(fileName, "a+");
	    fprintf(outputFile, "%d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
		    L, sysN, k, block.x, grid.x,
		    redMean*1.0e-3, errorRed*1.0e-3, blueMean*1.0e-3, errorBlue*1.0e-3);
	    fflush(outputFile);
	    fclose(outputFile);
	    
	    // KERNEL: Sliced
	    redMean = blueMean = redMean2 = blueMean2 = 0.;
	    for(int sample = 0; sample < initSamples; sample++){
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepReds_Sliced<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
						     sys->d_reds2, sys->d_reds3,
						     rng->d_rngSeeds);
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      redMean += (swap/(initSteps)); 
	      redMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepBlues_Sliced<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
						      sys->d_blues2, sys->d_blues3,
						      rng->d_rngSeeds);
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      blueMean += (swap/(initSteps)); 
	      blueMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	    }
	    
	    redMean /= initSamples;  redMean2 /= initSamples;      
	    blueMean /= initSamples; blueMean2 /= initSamples;
	    
	    errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
	    errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
	    
	    snprintf(fileName, sizeof(fileName), "GPU%d_slicedKernel", sys->cardN);
	    outputFile = fopen(fileName, "a+");
	    fprintf(outputFile, "%d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
		    L, sysN, k, block.x, grid.x,
		    redMean*1.0e-3, errorRed*1.0e-3, blueMean*1.0e-3, errorBlue*1.0e-3);
	    fflush(outputFile);
	    fclose(outputFile);
	  	  
	    CRNGArrayReset(rng);
	    free(rng);
	  }
	}else test = 0;
      }

      initBlock = 0;
    
      ////////////////////////////////////////
      // GRID KERNELS
      ///////////////////////////////////////
	  
      int A = sys->L*sys->L;
      int hA = A/2;

      int count3 = 0, test3 = 1;

      for(sys->k = 1; test3; ){

	float redMean = 0., redMean2 = 0., swap = 0.;
	float blueMean = 0., blueMean2 = 0.;
	float errorRed = 0., errorBlue = 0.;

	if(count3 > 0) do{sys->k++;}while(sys->L%sys->k != 0 && sys->k <= sys->L);
	count3++;

	dim3 blockS(sys->L/2, sys->k, 1);
	dim3 gridS(hA/(blockS.x*blockS.y), sys->L, sys->sysN);
	dim3 blockG(sys->L, sys->k, 1);
	dim3 gridG(A/(blockG.x*blockG.y), sys->L/2, sys->sysN);

	int threads = blockS.x*blockS.y*blockS.z*gridS.x*gridS.y*gridS.z;
	CRNGArray_t *rng;

	if(threads > 0) rng = CRNGArrayInit(threads);

	if(sys->k*sys->L <= maxThreadsPerBlock && sys->k <= sys->L && sys->k*sys->L >= 32){	  
	  printf("Measuring Times Grid    L:%d lines:%d\n",
		 L, sys->k);
	}


	if(sys->k*sys->L/2 <= maxThreadsPerBlock && sys->k <= sys->L && sys->k*sys->L/2 >= 32){
	  // KERNEL: Standard
	  	  	  
	  int h_block_xStrideS = blockS.x*blockS.y;
	  int h_block_yStrideS = gridS.x*h_block_xStrideS;
	  
	  MYCUDA_ERROR( cudaMemcpyToSymbol(d_block_xStrideS, &h_block_xStrideS, sizeof(int)) );
	  MYCUDA_ERROR( cudaMemcpyToSymbol(d_block_yStrideS, &h_block_yStrideS, sizeof(int)) );	  
	  
	  for(int sample = 0; sample < initSamples; sample++){
	    
	    MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	    
	    for(int steps = 0; steps < initSteps; steps++){	  
	      MCStepReds_StandardGrid<<< gridS, blockS >>>(sys->d_reds0, sys->d_reds1,
							   sys->d_reds2, sys->d_reds3,
							   rng->d_rngSeeds);
	      MYCUDA_ERROR( cudaDeviceSynchronize() );
	    }
	    
	    MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	    MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	    MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	    redMean += (swap/(initSteps)); 
	    redMean2 += (swap/(initSteps))*(swap/(initSteps));
	    swap = 0.;
	    
	    MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	    
	    for(int steps = 0; steps < initSteps; steps++){	  
	      MCStepBlues_StandardGrid<<< gridS, blockS >>>(sys->d_blues0, sys->d_blues1,
							    sys->d_blues2, sys->d_blues3,
							    rng->d_rngSeeds);
	      MYCUDA_ERROR( cudaDeviceSynchronize() );
	    }
	    
	    MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	    MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	    MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	    blueMean += (swap/(initSteps)); 
	    blueMean2 += (swap/(initSteps))*(swap/(initSteps));
	    swap = 0.;
	  }
	  
	  redMean /= initSamples;  redMean2 /= initSamples;      
	  blueMean /= initSamples; blueMean2 /= initSamples;
	  
	  errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
	  errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
	  
	  snprintf(fileName, sizeof(fileName), "GPU%d_standardGridKernel", sys->cardN);
	  FILE *outputFile = fopen(fileName, "a+");
	  fprintf(outputFile, "%d %d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
		  L, sysN, sys->k, blockS.x, blockS.y, gridS.x,
		  redMean*1.0e-3, errorRed*1.0e-3, blueMean*1.0e-3, errorBlue*1.0e-3);
	  fflush(outputFile);
	  fclose(outputFile);
	}

	if(sys->k*sys->L <= maxThreadsPerBlock && sys->k <= sys->L){	  
	  if(sys->k*sys->L >= 32){
	    // KERNEL: Sliced
	    redMean = blueMean = redMean2 = blueMean2 = 0.;
	    
	    int h_block_xStride = blockG.x*blockG.y;
	    int h_block_yStride = gridG.x*h_block_xStride;
	    
	    MYCUDA_ERROR( cudaMemcpyToSymbol(d_block_xStride, &h_block_xStride, sizeof(int)) );
	    MYCUDA_ERROR( cudaMemcpyToSymbol(d_block_yStride, &h_block_yStride, sizeof(int)) );
	    
	    for(int sample = 0; sample < initSamples; sample++){
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	    
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepReds_SlicedGrid<<< gridG, blockG >>>(sys->d_reds0, sys->d_reds1,
							   sys->d_reds2, sys->d_reds3,
							   rng->d_rngSeeds);
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      redMean += (swap/(initSteps)); 
	      redMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepBlues_SlicedGrid<<< gridG, blockG >>>(sys->d_blues0, sys->d_blues1,
							    sys->d_blues2, sys->d_blues3,
							    rng->d_rngSeeds);
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      blueMean += (swap/(initSteps)); 
	      blueMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	    }
	    
	    redMean /= initSamples;  redMean2 /= initSamples;      
	    blueMean /= initSamples; blueMean2 /= initSamples;
	    
	    errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
	    errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
	    
	    snprintf(fileName, sizeof(fileName), "GPU%d_slicedGridKernel", sys->cardN);
	    FILE *outputFile = fopen(fileName, "a+");
	    fprintf(outputFile, "%d %d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
		    L, sysN, sys->k, blockG.x, blockG.y, gridG.x,
		    redMean*1.0e-3, errorRed*1.0e-3, blueMean*1.0e-3, errorBlue*1.0e-3);
	    fflush(outputFile);
	    fclose(outputFile);
	  }
	} else test3 = 0;
	
	if(threads > 0){
	  CRNGArrayReset(rng);
	  free(rng);
	}
      }
      // CLEANING UP
      //CRNGArrayFree(&rng);
      if(dMemFree(sys)){
	fprintf(stderr,"\nError in dMemFree\n");
	return 1;
      }
      if(hMemFree(sys)){
	fprintf(stderr,"\nError in hMemFree\n");
	return 1;
      }
    }

    initSysN = 0;
  }
  // allocazione
  // deallocazione
  // linearizzazione nel numero dei sistemi con limite la memoria
  return 0;
}

extern int benchmarkGrid(int argc, char **argv){
  if(argc != 8){
    fprintf(stderr, "usage: %s <gpu> <L> <sysN> <block> <samples> <steps> <memory>\n\n", argv[0]);
    fprintf(stderr, "Systems' GPUs are:\n");
    int numberOfGPUs = 0;
    MYCUDA_ERROR( cudaGetDeviceCount(&numberOfGPUs) );
    for(int gpu = 0; gpu < numberOfGPUs; gpu++){
      cudaDeviceProp prop;
      MYCUDA_ERROR( cudaGetDeviceProperties(&prop, gpu) );
      printf("GPU %d: %s\n", gpu, prop.name);
    }
    return 1;
  }

  char fileName[1024];
  //struct timespec start, stop;

  sysEA3D_t *sys = (sysEA3D_t *)malloc(sizeof(sysEA3D_t));
  int initL, initSysN, initSamples, initSteps, initBlock;
  float memory = 0.;
  sscanf(argv[1], "%d", &sys->cardN);
  sscanf(argv[2], "%d", &initL);
  sscanf(argv[3], "%d", &initSysN);
  sscanf(argv[4], "%d", &initBlock);
  sscanf(argv[5], "%d", &initSamples);
  sscanf(argv[6], "%d", &initSteps);
  sscanf(argv[7], "%f", &memory);
  
  if(initL == 0) initL = 8;
  if(initSamples == 0) initSamples = 100;
  if(initSteps == 0) initSteps = 50;

  sys->beta = 1.101;

  int numberOfGPUs = 0;
  cudaGetDeviceCount(&numberOfGPUs);
  if(sys->cardN >= numberOfGPUs){
    fprintf(stderr, "gpu %d >= gpuN %d\n", sys->cardN, numberOfGPUs);
    return 1;
  }

  MYCUDA_ERROR( cudaSetDevice(sys->cardN) );
  cudaDeviceProp prop;
  MYCUDA_ERROR( cudaGetDeviceProperties(&prop, sys->cardN) );
  printf("GPU %d: %s Compute Capability: %d.%d\n", sys->cardN, prop.name, prop.major, prop.minor);

  cudaEvent_t startEvent, stopEvent;
  MYCUDA_ERROR( cudaEventCreate(&startEvent) );
  MYCUDA_ERROR( cudaEventCreate(&stopEvent) );

  int maxThreadsPerBlock = 0;
  if(prop.major == 2) maxThreadsPerBlock = 512;
  if(prop.major == 3) maxThreadsPerBlock = 1024;

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Bitwise, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Bitwise, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Standard, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Standard, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Sliced, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Sliced, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_StandardGrid, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_StandardGrid, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_SlicedGrid, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_SlicedGrid, cudaFuncCachePreferL1) );


  for(int L = initL; L <= 256; L += 2){

    int hV = L*L*L/2;
    size_t NJ_Size = 6*hV*sizeof(MSC), NS_Size = 4*2*hV*sizeof(MSC), RNG_Size = hV*sizeof(RNGT);
    size_t devFree = 0, devTotal = 0;
    MYCUDA_ERROR( cudaMemGetInfo(&devFree, &devTotal) );
    // here to decide the percentage of device memory to use. 
    int MaxSysN = (int)(memory*(devFree/(NJ_Size + NS_Size + RNG_Size)));
    printf("\n\nL = %d, Free memory on device: %lu, MaxSysN: %d\n", L, devFree, MaxSysN);

    if(initSysN == 0) initSysN = 1;

    for(int sysN = initSysN; sysN <= MaxSysN; sysN *= 2){
      sys->L = L;
      sys->V = sys->L*sys->L*sys->L;
      sys->sysN = sysN;
      sys->NJ = sys->sysN*sys->V/2;
      sys->NS = sys->sysN*sys->V/2;
      
      if(masksInit(sys)){
	fprintf(stderr,"\nError in masksInit\n");
	return 1;
      }
      if(MCWeightsInit(sys)){
	fprintf(stderr,"\nError in MCWeightsInit\n");
	return 1;
      }
      if(dMemInit(sys)){
	fprintf(stderr,"\nError in dMemInit\n");
	return 1;
      }
      if(hMemInit2(sys)){
	fprintf(stderr,"\nError in hMemInit\n");
	return 1;
      }
      if(H2DSysCopy(sys)){
	fprintf(stderr,"\nError in H2DSysCopy\n");
	return 1;
      }

      MYCUDA_ERROR( cudaMemGetInfo(&devFree, &devTotal) );
      printf("\n\n---\tL = %d, sysN = %d, free memory after allocation: %lu\t---\n\n",
	     L, sysN, devFree);


      ////////////////////////////////////////
      // GRID KERNELS
      ///////////////////////////////////////
	  
      int A = sys->L*sys->L;
      int hA = A/2;

      int count3 = 0, test3 = 1;

      for(sys->k = 1; test3; ){

	float redMean = 0., redMean2 = 0., swap = 0.;
	float blueMean = 0., blueMean2 = 0.;
	float errorRed = 0., errorBlue = 0.;

	if(count3 > 0) do{sys->k++;}while(sys->L%sys->k != 0 && sys->k <= sys->L);
	count3++;

	dim3 blockS(sys->L/2, sys->k, 1);
	dim3 gridS(hA/(blockS.x*blockS.y), sys->L, sys->sysN);
	dim3 blockG(sys->L, sys->k, 1);
	dim3 gridG(A/(blockG.x*blockG.y), sys->L/2, sys->sysN);

	int threads = blockS.x*blockS.y*blockS.z*gridS.x*gridS.y*gridS.z;
	CRNGArray_t *rng;

	if(threads > 0) rng = CRNGArrayInit(threads);

	if(sys->k*sys->L <= maxThreadsPerBlock && sys->k <= sys->L && sys->k*sys->L >= 32){	  
	  printf("Measuring Times Grid    L:%d lines:%d\n",
		 L, sys->k);
	}


	if(sys->k*sys->L/2 <= maxThreadsPerBlock && sys->k <= sys->L && sys->k*sys->L/2 >= 32){
	  // KERNEL: Standard
	  	  	  
	  int h_block_xStrideS = blockS.x*blockS.y;
	  int h_block_yStrideS = gridS.x*h_block_xStrideS;
	  
	  MYCUDA_ERROR( cudaMemcpyToSymbol(d_block_xStrideS, &h_block_xStrideS, sizeof(int)) );
	  MYCUDA_ERROR( cudaMemcpyToSymbol(d_block_yStrideS, &h_block_yStrideS, sizeof(int)) );	  
	  
	  for(int sample = 0; sample < initSamples; sample++){
	    
	    MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	    
	    for(int steps = 0; steps < initSteps; steps++){	  
	      MCStepReds_StandardGrid<<< gridS, blockS >>>(sys->d_reds0, sys->d_reds1,
							   sys->d_reds2, sys->d_reds3,
							   rng->d_rngSeeds);
	      MYCUDA_ERROR( cudaDeviceSynchronize() );
	    }
	    
	    MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	    MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	    MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	    redMean += (swap/(initSteps)); 
	    redMean2 += (swap/(initSteps))*(swap/(initSteps));
	    swap = 0.;
	    
	    MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	    
	    for(int steps = 0; steps < initSteps; steps++){	  
	      MCStepBlues_StandardGrid<<< gridS, blockS >>>(sys->d_blues0, sys->d_blues1,
							    sys->d_blues2, sys->d_blues3,
							    rng->d_rngSeeds);
	      MYCUDA_ERROR( cudaDeviceSynchronize() );
	    }
	    
	    MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	    MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	    MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	    blueMean += (swap/(initSteps)); 
	    blueMean2 += (swap/(initSteps))*(swap/(initSteps));
	    swap = 0.;
	  }
	  
	  redMean /= initSamples;  redMean2 /= initSamples;      
	  blueMean /= initSamples; blueMean2 /= initSamples;
	  
	  errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
	  errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
	  
	  snprintf(fileName, sizeof(fileName), "GPU%d_standardGridKernel", sys->cardN);
	  FILE *outputFile = fopen(fileName, "a+");
	  fprintf(outputFile, "%d %d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
		  L, sysN, sys->k, blockS.x, blockS.y, gridS.x,
		  redMean*1.0e-3, errorRed*1.0e-3, blueMean*1.0e-3, errorBlue*1.0e-3);
	  fflush(outputFile);
	  fclose(outputFile);
	}

	if(sys->k*sys->L <= maxThreadsPerBlock && sys->k <= sys->L){	  
	  if(sys->k*sys->L >= 32){
	    // KERNEL: Sliced
	    redMean = blueMean = redMean2 = blueMean2 = 0.;
	    
	    int h_block_xStride = blockG.x*blockG.y;
	    int h_block_yStride = gridG.x*h_block_xStride;
	    
	    MYCUDA_ERROR( cudaMemcpyToSymbol(d_block_xStride, &h_block_xStride, sizeof(int)) );
	    MYCUDA_ERROR( cudaMemcpyToSymbol(d_block_yStride, &h_block_yStride, sizeof(int)) );
	    
	    for(int sample = 0; sample < initSamples; sample++){
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	    
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepReds_SlicedGrid<<< gridG, blockG >>>(sys->d_reds0, sys->d_reds1,
							   sys->d_reds2, sys->d_reds3,
							   rng->d_rngSeeds);
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      redMean += (swap/(initSteps)); 
	      redMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepBlues_SlicedGrid<<< gridG, blockG >>>(sys->d_blues0, sys->d_blues1,
							    sys->d_blues2, sys->d_blues3,
							    rng->d_rngSeeds);
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      blueMean += (swap/(initSteps)); 
	      blueMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	    }
	    
	    redMean /= initSamples;  redMean2 /= initSamples;      
	    blueMean /= initSamples; blueMean2 /= initSamples;
	    
	    errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
	    errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
	    
	    snprintf(fileName, sizeof(fileName), "GPU%d_slicedGridKernel", sys->cardN);
	    FILE *outputFile = fopen(fileName, "a+");
	    fprintf(outputFile, "%d %d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
		    L, sysN, sys->k, blockG.x, blockG.y, gridG.x,
		    redMean*1.0e-3, errorRed*1.0e-3, blueMean*1.0e-3, errorBlue*1.0e-3);
	    fflush(outputFile);
	    fclose(outputFile);
	  }
	} else test3 = 0;
	
	if(threads > 0){
	  CRNGArrayReset(rng);
	  free(rng);
	}
      }
      // CLEANING UP
      //CRNGArrayFree(&rng);
      if(dMemFree(sys)){
	fprintf(stderr,"\nError in dMemFree\n");
	return 1;
      }
      if(hMemFree(sys)){
	fprintf(stderr,"\nError in hMemFree\n");
	return 1;
      }
    }

    initSysN = 0;
  }
  // allocazione
  // deallocazione
  // linearizzazione nel numero dei sistemi con limite la memoria
  return 0;
}

extern int benchmarkStandardGrid_NVPROF(int argc, char **argv){
  if(argc != 8){
    fprintf(stderr, "usage: %s <gpu> <L> <sysN> <lines> <dummyBlock> <samples> <steps>\n\n", argv[0]);
    fprintf(stderr, "Systems' GPUs are:\n");
    int numberOfGPUs = 0;
    MYCUDA_ERROR( cudaGetDeviceCount(&numberOfGPUs) );
    for(int gpu = 0; gpu < numberOfGPUs; gpu++){
      cudaDeviceProp prop;
      MYCUDA_ERROR( cudaGetDeviceProperties(&prop, gpu) );
      printf("GPU %d: %s\n", gpu, prop.name);
    }
    return 1;
  }

  sysEA3D_t *sys = (sysEA3D_t *)malloc(sizeof(sysEA3D_t));
  int initL, initSysN, initSamples, initSteps, initBlock, lines;

  sscanf(argv[1], "%d", &sys->cardN);
  sscanf(argv[2], "%d", &initL);
  sscanf(argv[3], "%d", &initSysN);
  sscanf(argv[4], "%d", &lines);
  sscanf(argv[5], "%d", &initBlock);
  sscanf(argv[6], "%d", &initSamples);
  sscanf(argv[7], "%d", &initSteps);
  
  sys->beta = 1.101;

  int numberOfGPUs = 0;
  cudaGetDeviceCount(&numberOfGPUs);
  if(sys->cardN >= numberOfGPUs){
    fprintf(stderr, "gpu %d >= gpuN %d\n", sys->cardN, numberOfGPUs);
    return 1;
  }

  MYCUDA_ERROR( cudaSetDevice(sys->cardN) );
  cudaDeviceProp prop;
  MYCUDA_ERROR( cudaGetDeviceProperties(&prop, sys->cardN) );
  printf("GPU %d: %s Compute Capability: %d.%d\n", sys->cardN, prop.name, prop.major, prop.minor);




  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_StandardGrid, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_StandardGrid, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_SlicedGrid, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_SlicedGrid, cudaFuncCachePreferL1) );



  int L = initL;


  int sysN = initSysN;
  sys->L = L;
  sys->V = sys->L*sys->L*sys->L;
  sys->sysN = sysN;
  sys->NJ = sys->sysN*sys->V/2;
  sys->NS = sys->sysN*sys->V/2;
  
  if(masksInit(sys)){
    fprintf(stderr,"\nError in masksInit\n");
    return 1;
  }
  if(MCWeightsInit(sys)){
    fprintf(stderr,"\nError in MCWeightsInit\n");
    return 1;
  }
  if(dMemInit(sys)){
    fprintf(stderr,"\nError in dMemInit\n");
    return 1;
  }
  if(hMemInit2(sys)){
    fprintf(stderr,"\nError in hMemInit\n");
    return 1;
  }
  if(H2DSysCopy(sys)){
    fprintf(stderr,"\nError in H2DSysCopy\n");
    return 1;
  }
     
  ////////////////////////////////////////
  // GRID KERNELS
  ///////////////////////////////////////
  
  int A = sys->L*sys->L;
  int hA = A/2;
  
  sys->k = lines;
        
  dim3 blockS(sys->L/2, sys->k, 1);
  dim3 gridS(hA/(blockS.x*blockS.y), sys->L, sys->sysN);
  
  int threads = blockS.x*blockS.y*blockS.z*gridS.x*gridS.y*gridS.z;
  CRNGArray_t *rng;
  
  if(threads > 0) rng = CRNGArrayInit(threads);
	  	  	  
  int h_block_xStrideS = blockS.x*blockS.y;
  int h_block_yStrideS = gridS.x*h_block_xStrideS;
  
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_block_xStrideS, &h_block_xStrideS, sizeof(int)) );
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_block_yStrideS, &h_block_yStrideS, sizeof(int)) );	  

  // Kernel: Standard
	  
  for(int sample = 0; sample < initSamples; sample++){
	    	    
    for(int steps = 0; steps < initSteps; steps++){	  
      MCStepReds_StandardGrid<<< gridS, blockS >>>(sys->d_reds0, sys->d_reds1,
						   sys->d_reds2, sys->d_reds3,
						   rng->d_rngSeeds);
      MYCUDA_ERROR( cudaDeviceSynchronize() );
    }
	    	    
    
    for(int steps = 0; steps < initSteps; steps++){	  
      MCStepBlues_StandardGrid<<< gridS, blockS >>>(sys->d_blues0, sys->d_blues1,
						    sys->d_blues2, sys->d_blues3,
						    rng->d_rngSeeds);
      MYCUDA_ERROR( cudaDeviceSynchronize() );
    }
  }
	  	    
  CRNGArrayReset(rng);
  free(rng);

  // CLEANING UP
  //CRNGArrayFree(&rng);
  if(dMemFree(sys)){
    fprintf(stderr,"\nError in dMemFree\n");
    return 1;
  }
  if(hMemFree(sys)){
    fprintf(stderr,"\nError in hMemFree\n");
    return 1;
  }

  return 0;
}

extern int benchmarkSlicedGrid_NVPROF(int argc, char **argv){
  if(argc != 8){
    fprintf(stderr, "usage: %s <gpu> <L> <sysN> <lines> <dummyBlock> <samples> <steps>\n\n", argv[0]);
    fprintf(stderr, "Systems' GPUs are:\n");
    int numberOfGPUs = 0;
    MYCUDA_ERROR( cudaGetDeviceCount(&numberOfGPUs) );
    for(int gpu = 0; gpu < numberOfGPUs; gpu++){
      cudaDeviceProp prop;
      MYCUDA_ERROR( cudaGetDeviceProperties(&prop, gpu) );
      printf("GPU %d: %s\n", gpu, prop.name);
    }
    return 1;
  }


  //struct timespec start, stop;

  sysEA3D_t *sys = (sysEA3D_t *)malloc(sizeof(sysEA3D_t));
  int initL, initSysN, initSamples, initSteps, initBlock, lines;

  sscanf(argv[1], "%d", &sys->cardN);
  sscanf(argv[2], "%d", &initL);
  sscanf(argv[3], "%d", &initSysN);
  sscanf(argv[4], "%d", &lines);
  sscanf(argv[5], "%d", &initBlock);
  sscanf(argv[6], "%d", &initSamples);
  sscanf(argv[7], "%d", &initSteps);
  
  sys->beta = 1.101;

  int numberOfGPUs = 0;
  cudaGetDeviceCount(&numberOfGPUs);
  if(sys->cardN >= numberOfGPUs){
    fprintf(stderr, "gpu %d >= gpuN %d\n", sys->cardN, numberOfGPUs);
    return 1;
  }

  MYCUDA_ERROR( cudaSetDevice(sys->cardN) );
  cudaDeviceProp prop;
  MYCUDA_ERROR( cudaGetDeviceProperties(&prop, sys->cardN) );
  printf("GPU %d: %s Compute Capability: %d.%d\n", sys->cardN, prop.name, prop.major, prop.minor);




  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_StandardGrid, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_StandardGrid, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_SlicedGrid, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_SlicedGrid, cudaFuncCachePreferL1) );


  int L = initL;


  int sysN = initSysN;
  sys->L = L;
  sys->V = sys->L*sys->L*sys->L;
  sys->sysN = sysN;
  sys->NJ = sys->sysN*sys->V/2;
  sys->NS = sys->sysN*sys->V/2;
  
  if(masksInit(sys)){
    fprintf(stderr,"\nError in masksInit\n");
    return 1;
  }
  if(MCWeightsInit(sys)){
    fprintf(stderr,"\nError in MCWeightsInit\n");
    return 1;
  }
  if(dMemInit(sys)){
    fprintf(stderr,"\nError in dMemInit\n");
    return 1;
  }
  if(hMemInit2(sys)){
    fprintf(stderr,"\nError in hMemInit\n");
    return 1;
  }
  if(H2DSysCopy(sys)){
    fprintf(stderr,"\nError in H2DSysCopy\n");
    return 1;
  }
     
  
  ////////////////////////////////////////
  // GRID KERNELS
  ///////////////////////////////////////
  
  int A = sys->L*sys->L;

  
  sys->k = lines;
        
  dim3 blockG(sys->L, sys->k, 1);
  dim3 gridG(A/(blockG.x*blockG.y), sys->L/2, sys->sysN);
  
  int threads = blockG.x*blockG.y*blockG.z*gridG.x*gridG.y*gridG.z;
  CRNGArray_t *rng;
  
  if(threads > 0) rng = CRNGArrayInit(threads);
	  	  	  	  
  // KERNEL: Sliced
  int h_block_xStride = blockG.x*blockG.y;
  int h_block_yStride = gridG.x*h_block_xStride;
	    
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_block_xStride, &h_block_xStride, sizeof(int)) );
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_block_yStride, &h_block_yStride, sizeof(int)) );
	    
  for(int sample = 0; sample < initSamples; sample++){    
    
    for(int steps = 0; steps < initSteps; steps++){	  
      MCStepReds_SlicedGrid<<< gridG, blockG >>>(sys->d_reds0, sys->d_reds1,
						 sys->d_reds2, sys->d_reds3,
						 rng->d_rngSeeds);
      MYCUDA_ERROR( cudaDeviceSynchronize() );
    }
	      
    for(int steps = 0; steps < initSteps; steps++){	  
      MCStepBlues_SlicedGrid<<< gridG, blockG >>>(sys->d_blues0, sys->d_blues1,
						  sys->d_blues2, sys->d_blues3,
						  rng->d_rngSeeds);
      MYCUDA_ERROR( cudaDeviceSynchronize() );
    }
  }
	    
  CRNGArrayReset(rng);
  free(rng);

  // CLEANING UP
  //CRNGArrayFree(&rng);
  if(dMemFree(sys)){
    fprintf(stderr,"\nError in dMemFree\n");
    return 1;
  }
  if(hMemFree(sys)){
    fprintf(stderr,"\nError in hMemFree\n");
    return 1;
  }

  return 0;
}

extern int benchmarkSmooth(int argc, char **argv){
  if(argc != 9){
    fprintf(stderr, "usage: %s <gpu> <L> <sysN> <block> <samples> <steps> <memory> <spinsPerThread>\n\n", argv[0]);
    fprintf(stderr, "Systems' GPUs are:\n");
    int numberOfGPUs = 0;
    MYCUDA_ERROR( cudaGetDeviceCount(&numberOfGPUs) );
    for(int gpu = 0; gpu < numberOfGPUs; gpu++){
      cudaDeviceProp prop;
      MYCUDA_ERROR( cudaGetDeviceProperties(&prop, gpu) );
      printf("GPU %d: %s\n", gpu, prop.name);
    }
    return 1;
  }

  char fileName[1024];
  //struct timespec start, stop;

  sysEA3D_t *sys = (sysEA3D_t *)malloc(sizeof(sysEA3D_t));
  int initL, initSysN, initSamples, initSteps, initBlock, spinsPerThread;
  float memory;
  sscanf(argv[1], "%d", &sys->cardN);
  sscanf(argv[2], "%d", &initL);
  sscanf(argv[3], "%d", &initSysN);
  sscanf(argv[4], "%d", &initBlock);
  sscanf(argv[5], "%d", &initSamples);
  sscanf(argv[6], "%d", &initSteps);
  sscanf(argv[7], "%f", &memory);
  sscanf(argv[8], "%d", &spinsPerThread);

  if(initL == 0) initL = 8;
  if(initSamples == 0) initSamples = 100;
  if(initSteps == 0) initSteps = 50;

  sys->beta = 1.101;

  int numberOfGPUs = 0;
  cudaGetDeviceCount(&numberOfGPUs);
  if(sys->cardN >= numberOfGPUs){
    fprintf(stderr, "gpu %d >= gpuN %d\n", sys->cardN, numberOfGPUs);
    return 1;
  }

  MYCUDA_ERROR( cudaSetDevice(sys->cardN) );
  cudaDeviceProp prop;
  MYCUDA_ERROR( cudaGetDeviceProperties(&prop, sys->cardN) );
  printf("GPU %d: %s Compute Capability: %d.%d\n", sys->cardN, prop.name, prop.major, prop.minor);

  cudaEvent_t startEvent, stopEvent;
  MYCUDA_ERROR( cudaEventCreate(&startEvent) );
  MYCUDA_ERROR( cudaEventCreate(&stopEvent) );

  int maxThreadsPerBlock = 0;
  if(prop.major == 2) maxThreadsPerBlock = 512;
  if(prop.major == 3) maxThreadsPerBlock = 1024;

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Bitwise, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Bitwise, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Standard, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Standard, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Sliced, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Sliced, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_StandardGrid, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_StandardGrid, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_SlicedGrid, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_SlicedGrid, cudaFuncCachePreferL1) );


  for(int L = initL; L <= 256; L += 2){

    int hV = L*L*L/2;
    size_t NJ_Size = 6*hV*sizeof(MSC), NS_Size = 4*2*hV*sizeof(MSC), RNG_Size = hV*sizeof(RNGT);
    size_t devFree = 0, devTotal = 0;
    MYCUDA_ERROR( cudaMemGetInfo(&devFree, &devTotal) );
    // here to decide the percentage of device memory to use. 
    int MaxSysN = (int)(memory*(devFree/(NJ_Size + NS_Size + RNG_Size)));
    printf("\n\nL = %d, Free memory on device: %lu, MaxSysN: %d\n", L, devFree, MaxSysN);

    if(initSysN == 0) initSysN = 1;

    for(int sysN = initSysN; sysN <= MaxSysN; sysN *= 2){
      sys->L = L;
      sys->V = sys->L*sys->L*sys->L;
      sys->sysN = sysN;
      sys->NJ = sys->sysN*sys->V/2;
      sys->NS = sys->sysN*sys->V/2;
      
      if(masksInit(sys)){
	fprintf(stderr,"\nError in masksInit\n");
	return 1;
      }
      if(MCWeightsInit(sys)){
	fprintf(stderr,"\nError in MCWeightsInit\n");
	return 1;
      }
      if(dMemInit(sys)){
	fprintf(stderr,"\nError in dMemInit\n");
	return 1;
      }
      if(hMemInit2(sys)){
	fprintf(stderr,"\nError in hMemInit\n");
	return 1;
      }
      if(H2DSysCopy(sys)){
	fprintf(stderr,"\nError in H2DSysCopy\n");
	return 1;
      }

      MYCUDA_ERROR( cudaMemGetInfo(&devFree, &devTotal) );
      printf("\n\n---\tL = %d, sysN = %d, free memory after allocation: %lu\t---\n\n",
	     L, sysN, devFree);

      ////////////////////////////////////////
      // NO-GRID KERNELS
      ///////////////////////////////////////

      int blockInit = 0;

      if(initBlock == 0){
	blockInit = maxThreadsPerBlock;
	while(blockInit > sys->V/2) blockInit /= 2;
      }


      for(sys->blockSize = blockInit; sys->blockSize >= 32; sys->blockSize /= 2){

	//printf("blockSize: %d\n", sys->blockSize);

	if(sys->blockSize >= 32){
	  int kMax = sys->V/sys->blockSize/2;
	  if(kMax > spinsPerThread && spinsPerThread != 0) kMax = spinsPerThread;
	  //printf("blockSizeIF: %d\n", sys->blockSize);

	  for(int k = 0; k < kMax ;){
	    
	    float redMean = 0., redMean2 = 0., swap = 0.;
	    float blueMean = 0., blueMean2 = 0.;
	    float errorRed = 0., errorBlue = 0.;

	    do{k++;}while(kMax%k != 0);	    
	    
	    dim3 block(sys->blockSize,1,1);

	    int fitGrid = (sys->V/2/k + sys->blockSize - 1)/sys->blockSize;
	    dim3 grid(fitGrid, sys->sysN, 1);
	    int h_loopStride = block.x*grid.x;
	    MYCUDA_ERROR( cudaMemcpyToSymbol(d_loopStride, &h_loopStride, sizeof(int)) );
	    
	    printf("Measuring Times No-Grid L:%d block.x:%d grid.x:%d spins per thread:%d\n",
		   L, block.x, grid.x, k);
	    
	    int threads = block.x*block.y*block.z*grid.x*grid.y*grid.z;
	    MYCUDA_ERROR( cudaMemcpyToSymbol(d_threads, &threads, sizeof(int)) );
	    
	    CRNGArray_t *rng = CRNGArrayInit(threads);
	    
	    // KERNEL: Bitwise; Only L = 2^k are examined
	    if( ((1U<<findLog2(sys->L))^sys->L) == 0 ){
	      for(int sample = 0; sample < initSamples; sample++){
		
		MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
		
		for(int steps = 0; steps < initSteps; steps++){	  
		  MCStepReds_Bitwise<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
							sys->d_reds2, sys->d_reds3,
							rng->d_rngSeeds);
		  MYCUDA_ERROR( cudaDeviceSynchronize() );
		}
		
		MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
		MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
		MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
		redMean += (swap/(initSteps)); 
		redMean2 += (swap/(initSteps))*(swap/(initSteps));
		swap = 0.;
		
		MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
		
		for(int steps = 0; steps < initSteps; steps++){	  
		  MCStepBlues_Bitwise<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
							 sys->d_blues2, sys->d_blues3,
							 rng->d_rngSeeds);
		  MYCUDA_ERROR( cudaDeviceSynchronize() );
		}
		
		MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
		MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
		MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
		blueMean += (swap/(initSteps)); 
		blueMean2 += (swap/(initSteps))*(swap/(initSteps));
		swap = 0.;
	      }
	      
	      redMean /= initSamples;  redMean2 /= initSamples;      
	      blueMean /= initSamples; blueMean2 /= initSamples;
	      
	      errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
	      errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
	      
	      snprintf(fileName, sizeof(fileName), "GPU%d_bitwiseKernel_smooth", sys->cardN);
	      FILE *outputFile = fopen(fileName, "a+");
	      fprintf(outputFile, "%d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
		      L, sysN, k, block.x, grid.x,
		      redMean*1.0e-3, errorRed*1.0e-3, blueMean*1.0e-3, errorBlue*1.0e-3);
	      fflush(outputFile);
	      fclose(outputFile);
	    }
	    
	    // KERNEL: Standard
	    redMean = blueMean = redMean2 = blueMean2 = 0.;
	    for(int sample = 0; sample < initSamples; sample++){
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepReds_Standard<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
						       sys->d_reds2, sys->d_reds3,
						       rng->d_rngSeeds);
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      redMean += (swap/(initSteps)); 
	      redMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepBlues_Standard<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
							sys->d_blues2, sys->d_blues3,
							rng->d_rngSeeds);
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      blueMean += (swap/(initSteps)); 
	      blueMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	    }
	    
	    redMean /= initSamples;  redMean2 /= initSamples;      
	    blueMean /= initSamples; blueMean2 /= initSamples;
	    
	    errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
	    errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
	    
	    snprintf(fileName, sizeof(fileName), "GPU%d_standardKernel_smooth", sys->cardN);
	    FILE *outputFile = fopen(fileName, "a+");
	    fprintf(outputFile, "%d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
		    L, sysN, k, block.x, grid.x,
		    redMean*1.0e-3, errorRed*1.0e-3, blueMean*1.0e-3, errorBlue*1.0e-3);
	    fflush(outputFile);
	    fclose(outputFile);
	    
	    // KERNEL: Sliced
	    redMean = blueMean = redMean2 = blueMean2 = 0.;
	    for(int sample = 0; sample < initSamples; sample++){
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepReds_Sliced<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
						     sys->d_reds2, sys->d_reds3,
						     rng->d_rngSeeds);
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      redMean += (swap/(initSteps)); 
	      redMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepBlues_Sliced<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
						      sys->d_blues2, sys->d_blues3,
						      rng->d_rngSeeds);
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      blueMean += (swap/(initSteps)); 
	      blueMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	    }
	    
	    redMean /= initSamples;  redMean2 /= initSamples;      
	    blueMean /= initSamples; blueMean2 /= initSamples;
	    
	    errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
	    errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
	    
	    snprintf(fileName, sizeof(fileName), "GPU%d_slicedKernel_smooth", sys->cardN);
	    outputFile = fopen(fileName, "a+");
	    fprintf(outputFile, "%d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
		    L, sysN, k, block.x, grid.x,
		    redMean*1.0e-3, errorRed*1.0e-3, blueMean*1.0e-3, errorBlue*1.0e-3);
	    fflush(outputFile);
	    fclose(outputFile);
	  	  
	    CRNGArrayReset(rng);
	    free(rng);
	  }
	}
      }

      // CLEANING UP
      //CRNGArrayFree(&rng);
      if(dMemFree(sys)){
	fprintf(stderr,"\nError in dMemFree\n");
	return 1;
      }
      if(hMemFree(sys)){
	fprintf(stderr,"\nError in hMemFree\n");
	return 1;
      }
    }

    initSysN = 0;
  }
  // allocazione
  // deallocazione
  // linearizzazione nel numero dei sistemi con limite la memoria
  return 0;
}

extern int benchmarkBitwise_NVPROF(int argc, char **argv){
  if(argc != 8){
    fprintf(stderr, "usage: %s <gpu> <L> <sysN> <spinsPerThread> <block> <samples> <steps>\n\n", argv[0]);
    fprintf(stderr, "Systems' GPUs are:\n");
    int numberOfGPUs = 0;
    MYCUDA_ERROR( cudaGetDeviceCount(&numberOfGPUs) );
    for(int gpu = 0; gpu < numberOfGPUs; gpu++){
      cudaDeviceProp prop;
      MYCUDA_ERROR( cudaGetDeviceProperties(&prop, gpu) );
      printf("GPU %d: %s\n", gpu, prop.name);
    }
    return 1;
  }


  //struct timespec start, stop;

  sysEA3D_t *sys = (sysEA3D_t *)malloc(sizeof(sysEA3D_t));
  int initL, initSysN, initSamples, initSteps, initBlock, spinsPerThread;

  sscanf(argv[1], "%d", &sys->cardN);
  sscanf(argv[2], "%d", &initL);
  sscanf(argv[3], "%d", &initSysN);
  sscanf(argv[4], "%d", &spinsPerThread);
  sscanf(argv[5], "%d", &initBlock);
  sscanf(argv[6], "%d", &initSamples);
  sscanf(argv[7], "%d", &initSteps);

  sys->beta = 1.101;

  int numberOfGPUs = 0;
  cudaGetDeviceCount(&numberOfGPUs);
  if(sys->cardN >= numberOfGPUs){
    fprintf(stderr, "gpu %d >= gpuN %d\n", sys->cardN, numberOfGPUs);
    return 1;
  }

  MYCUDA_ERROR( cudaSetDevice(sys->cardN) );
  cudaDeviceProp prop;
  MYCUDA_ERROR( cudaGetDeviceProperties(&prop, sys->cardN) );
  printf("GPU %d: %s Compute Capability: %d.%d\n", sys->cardN, prop.name, prop.major, prop.minor);


  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Bitwise, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Bitwise, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Standard, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Standard, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Sliced, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Sliced, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_StandardGrid, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_StandardGrid, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_SlicedGrid, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_SlicedGrid, cudaFuncCachePreferL1) );


  int L = initL;

  int sysN = initSysN;

  sys->L = L;
  sys->V = sys->L*sys->L*sys->L;
  sys->sysN = sysN;
  sys->NJ = sys->sysN*sys->V/2;
  sys->NS = sys->sysN*sys->V/2;
      
  if(masksInit(sys)){
    fprintf(stderr,"\nError in masksInit\n");
    return 1;
  }
  if(MCWeightsInit(sys)){
    fprintf(stderr,"\nError in MCWeightsInit\n");
    return 1;
  }
  if(dMemInit(sys)){
    fprintf(stderr,"\nError in dMemInit\n");
    return 1;
  }
  if(hMemInit2(sys)){
    fprintf(stderr,"\nError in hMemInit\n");
    return 1;
  }
  if(H2DSysCopy(sys)){
    fprintf(stderr,"\nError in H2DSysCopy\n");
    return 1;
  }

  ////////////////////////////////////////
  // NO-GRID KERNELS
  ///////////////////////////////////////
  
  int blockInit = initBlock;
  sys->blockSize = blockInit;

  int k = spinsPerThread;
	    	    
  dim3 block(sys->blockSize,1,1);

  int fitGrid = (sys->V/2/k + sys->blockSize - 1)/sys->blockSize;
  dim3 grid(fitGrid, sys->sysN, 1);
  int h_loopStride = block.x*grid.x;
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_loopStride, &h_loopStride, sizeof(int)) );
	    	    
  int threads = block.x*block.y*block.z*grid.x*grid.y*grid.z;
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_threads, &threads, sizeof(int)) );
	    
  CRNGArray_t *rng = CRNGArrayInit(threads);
  
  // KERNEL: Bitwise; Only L = 2^k are examined

  for(int sample = 0; sample < initSamples; sample++){
        
    for(int steps = 0; steps < initSteps; steps++){	  
      MCStepReds_Bitwise<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
					    sys->d_reds2, sys->d_reds3,
					    rng->d_rngSeeds);
      MYCUDA_ERROR( cudaDeviceSynchronize() );
    }
			       		
    for(int steps = 0; steps < initSteps; steps++){	  
      MCStepBlues_Bitwise<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
					     sys->d_blues2, sys->d_blues3,
					     rng->d_rngSeeds);
      MYCUDA_ERROR( cudaDeviceSynchronize() );
    }
		
  }
	      	      	   	  	  
  CRNGArrayReset(rng);
  free(rng);

  if(dMemFree(sys)){
    fprintf(stderr,"\nError in dMemFree\n");
    return 1;
  }
  if(hMemFree(sys)){
    fprintf(stderr,"\nError in hMemFree\n");
    return 1;
  }
  
  return 0;
}

extern int benchmarkStandard_NVPROF(int argc, char **argv){
  if(argc != 8){
    fprintf(stderr, "usage: %s <gpu> <L> <sysN> <spinsPerThread> <block> <samples> <steps>\n\n", argv[0]);
    fprintf(stderr, "Systems' GPUs are:\n");
    int numberOfGPUs = 0;
    MYCUDA_ERROR( cudaGetDeviceCount(&numberOfGPUs) );
    for(int gpu = 0; gpu < numberOfGPUs; gpu++){
      cudaDeviceProp prop;
      MYCUDA_ERROR( cudaGetDeviceProperties(&prop, gpu) );
      printf("GPU %d: %s\n", gpu, prop.name);
    }
    return 1;
  }


  //struct timespec start, stop;

  sysEA3D_t *sys = (sysEA3D_t *)malloc(sizeof(sysEA3D_t));
  int initL, initSysN, initSamples, initSteps, initBlock, spinsPerThread;

  sscanf(argv[1], "%d", &sys->cardN);
  sscanf(argv[2], "%d", &initL);
  sscanf(argv[3], "%d", &initSysN);
  sscanf(argv[4], "%d", &spinsPerThread);
  sscanf(argv[5], "%d", &initBlock);
  sscanf(argv[6], "%d", &initSamples);
  sscanf(argv[7], "%d", &initSteps);

  sys->beta = 1.101;

  int numberOfGPUs = 0;
  cudaGetDeviceCount(&numberOfGPUs);
  if(sys->cardN >= numberOfGPUs){
    fprintf(stderr, "gpu %d >= gpuN %d\n", sys->cardN, numberOfGPUs);
    return 1;
  }

  MYCUDA_ERROR( cudaSetDevice(sys->cardN) );
  cudaDeviceProp prop;
  MYCUDA_ERROR( cudaGetDeviceProperties(&prop, sys->cardN) );
  printf("GPU %d: %s Compute Capability: %d.%d\n", sys->cardN, prop.name, prop.major, prop.minor);


  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Bitwise, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Bitwise, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Standard, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Standard, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Sliced, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Sliced, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_StandardGrid, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_StandardGrid, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_SlicedGrid, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_SlicedGrid, cudaFuncCachePreferL1) );


  int L = initL;

  int sysN = initSysN;

  sys->L = L;
  sys->V = sys->L*sys->L*sys->L;
  sys->sysN = sysN;
  sys->NJ = sys->sysN*sys->V/2;
  sys->NS = sys->sysN*sys->V/2;
      
  if(masksInit(sys)){
    fprintf(stderr,"\nError in masksInit\n");
    return 1;
  }
  if(MCWeightsInit(sys)){
    fprintf(stderr,"\nError in MCWeightsInit\n");
    return 1;
  }
  if(dMemInit(sys)){
    fprintf(stderr,"\nError in dMemInit\n");
    return 1;
  }
  if(hMemInit2(sys)){
    fprintf(stderr,"\nError in hMemInit\n");
    return 1;
  }
  if(H2DSysCopy(sys)){
    fprintf(stderr,"\nError in H2DSysCopy\n");
    return 1;
  }

  ////////////////////////////////////////
  // NO-GRID KERNELS
  ///////////////////////////////////////
  
  int blockInit = initBlock;
  sys->blockSize = blockInit;

  int k = spinsPerThread;
	    	    
  dim3 block(sys->blockSize,1,1);

  int fitGrid = (sys->V/2/k + sys->blockSize - 1)/sys->blockSize;
  dim3 grid(fitGrid, sys->sysN, 1);
  int h_loopStride = block.x*grid.x;
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_loopStride, &h_loopStride, sizeof(int)) );
	    	    
  int threads = block.x*block.y*block.z*grid.x*grid.y*grid.z;
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_threads, &threads, sizeof(int)) );
	    
  CRNGArray_t *rng = CRNGArrayInit(threads);	      
	      
  // KERNEL: Standard  
  for(int sample = 0; sample < initSamples; sample++){
	      	      
    for(int steps = 0; steps < initSteps; steps++){	  
      MCStepReds_Standard<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
					     sys->d_reds2, sys->d_reds3,
					     rng->d_rngSeeds);
      MYCUDA_ERROR( cudaDeviceSynchronize() );
    }
	      	      
    for(int steps = 0; steps < initSteps; steps++){	  
      MCStepBlues_Standard<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
					      sys->d_blues2, sys->d_blues3,
					      rng->d_rngSeeds);
      MYCUDA_ERROR( cudaDeviceSynchronize() );
    }
	      
  }	    
	  	  
  CRNGArrayReset(rng);
  free(rng);

  if(dMemFree(sys)){
    fprintf(stderr,"\nError in dMemFree\n");
    return 1;
  }
  if(hMemFree(sys)){
    fprintf(stderr,"\nError in hMemFree\n");
    return 1;
  }
  
  return 0;
}

extern int benchmarkSliced_NVPROF(int argc, char **argv){
  if(argc != 8){
    fprintf(stderr, "usage: %s <gpu> <L> <sysN> <spinsPerThread> <block> <samples> <steps>\n\n", argv[0]);
    fprintf(stderr, "Systems' GPUs are:\n");
    int numberOfGPUs = 0;
    MYCUDA_ERROR( cudaGetDeviceCount(&numberOfGPUs) );
    for(int gpu = 0; gpu < numberOfGPUs; gpu++){
      cudaDeviceProp prop;
      MYCUDA_ERROR( cudaGetDeviceProperties(&prop, gpu) );
      printf("GPU %d: %s\n", gpu, prop.name);
    }
    return 1;
  }


  //struct timespec start, stop;

  sysEA3D_t *sys = (sysEA3D_t *)malloc(sizeof(sysEA3D_t));
  int initL, initSysN, initSamples, initSteps, initBlock, spinsPerThread;

  sscanf(argv[1], "%d", &sys->cardN);
  sscanf(argv[2], "%d", &initL);
  sscanf(argv[3], "%d", &initSysN);
  sscanf(argv[4], "%d", &spinsPerThread);
  sscanf(argv[5], "%d", &initBlock);
  sscanf(argv[6], "%d", &initSamples);
  sscanf(argv[7], "%d", &initSteps);

  sys->beta = 1.101;

  int numberOfGPUs = 0;
  cudaGetDeviceCount(&numberOfGPUs);
  if(sys->cardN >= numberOfGPUs){
    fprintf(stderr, "gpu %d >= gpuN %d\n", sys->cardN, numberOfGPUs);
    return 1;
  }

  MYCUDA_ERROR( cudaSetDevice(sys->cardN) );
  cudaDeviceProp prop;
  MYCUDA_ERROR( cudaGetDeviceProperties(&prop, sys->cardN) );
  printf("GPU %d: %s Compute Capability: %d.%d\n", sys->cardN, prop.name, prop.major, prop.minor);


  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Bitwise, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Bitwise, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Standard, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Standard, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Sliced, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Sliced, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_StandardGrid, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_StandardGrid, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_SlicedGrid, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_SlicedGrid, cudaFuncCachePreferL1) );


  int L = initL;

  int sysN = initSysN;

  sys->L = L;
  sys->V = sys->L*sys->L*sys->L;
  sys->sysN = sysN;
  sys->NJ = sys->sysN*sys->V/2;
  sys->NS = sys->sysN*sys->V/2;
      
  if(masksInit(sys)){
    fprintf(stderr,"\nError in masksInit\n");
    return 1;
  }
  if(MCWeightsInit(sys)){
    fprintf(stderr,"\nError in MCWeightsInit\n");
    return 1;
  }
  if(dMemInit(sys)){
    fprintf(stderr,"\nError in dMemInit\n");
    return 1;
  }
  if(hMemInit2(sys)){
    fprintf(stderr,"\nError in hMemInit\n");
    return 1;
  }
  if(H2DSysCopy(sys)){
    fprintf(stderr,"\nError in H2DSysCopy\n");
    return 1;
  }

  ////////////////////////////////////////
  // NO-GRID KERNELS
  ///////////////////////////////////////
  
  int blockInit = initBlock;
  sys->blockSize = blockInit;

  int k = spinsPerThread;
	    	    
  dim3 block(sys->blockSize,1,1);

  int fitGrid = (sys->V/2/k + sys->blockSize - 1)/sys->blockSize;
  dim3 grid(fitGrid, sys->sysN, 1);
  int h_loopStride = block.x*grid.x;
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_loopStride, &h_loopStride, sizeof(int)) );
	    	    
  int threads = block.x*block.y*block.z*grid.x*grid.y*grid.z;
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_threads, &threads, sizeof(int)) );
	    
  CRNGArray_t *rng = CRNGArrayInit(threads);
  	    
  // KERNEL: Sliced

  for(int sample = 0; sample < initSamples; sample++){
    
    for(int steps = 0; steps < initSteps; steps++){	  
      MCStepReds_Sliced<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
					   sys->d_reds2, sys->d_reds3,
					   rng->d_rngSeeds);
      MYCUDA_ERROR( cudaDeviceSynchronize() );
    }
	      	      
    for(int steps = 0; steps < initSteps; steps++){	  
      MCStepBlues_Sliced<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
					    sys->d_blues2, sys->d_blues3,
					    rng->d_rngSeeds);
      MYCUDA_ERROR( cudaDeviceSynchronize() );
    }
	      
  }
	    
	  	  
  CRNGArrayReset(rng);
  free(rng);

  if(dMemFree(sys)){
    fprintf(stderr,"\nError in dMemFree\n");
    return 1;
  }
  if(hMemFree(sys)){
    fprintf(stderr,"\nError in hMemFree\n");
    return 1;
  }
  
  return 0;
}

extern int benchmarkSliced_MT19937(int argc, char **argv){
  if(argc != 8){
    fprintf(stderr, "usage: %s <gpu> <L> <sysN> <spinsPerThread> <block> <samples> <steps>\n\n", argv[0]);
    fprintf(stderr, "Systems' GPUs are:\n");
    int numberOfGPUs = 0;
    MYCUDA_ERROR( cudaGetDeviceCount(&numberOfGPUs) );
    for(int gpu = 0; gpu < numberOfGPUs; gpu++){
      cudaDeviceProp prop;
      MYCUDA_ERROR( cudaGetDeviceProperties(&prop, gpu) );
      printf("GPU %d: %s\n", gpu, prop.name);
    }
    return 1;
  }

  char fileName[1024];
  struct timespec start, stop;
  double redMean = 0., redMean2 = 0., swap = 0.;
  double blueMean = 0., blueMean2 = 0.;
  double errorRed = 0., errorBlue = 0.;


  sysEA3D_t *sys = (sysEA3D_t *)malloc(sizeof(sysEA3D_t));
  int initL, initSysN, initSamples, initSteps, initBlock, spinsPerThread;

  sscanf(argv[1], "%d", &sys->cardN);
  sscanf(argv[2], "%d", &initL);
  sscanf(argv[3], "%d", &initSysN);
  sscanf(argv[4], "%d", &spinsPerThread);
  sscanf(argv[5], "%d", &initBlock);
  sscanf(argv[6], "%d", &initSamples);
  sscanf(argv[7], "%d", &initSteps);

  sys->beta = 1.101;

  int numberOfGPUs = 0;
  cudaGetDeviceCount(&numberOfGPUs);
  if(sys->cardN >= numberOfGPUs){
    fprintf(stderr, "gpu %d >= gpuN %d\n", sys->cardN, numberOfGPUs);
    return 1;
  }

  MYCUDA_ERROR( cudaSetDevice(sys->cardN) );
  cudaDeviceProp prop;
  MYCUDA_ERROR( cudaGetDeviceProperties(&prop, sys->cardN) );
  printf("GPU %d: %s Compute Capability: %d.%d\n", sys->cardN, prop.name, prop.major, prop.minor);

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Sliced_MT19937, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Sliced_MT19937, cudaFuncCachePreferL1) );

  int L = initL;

  int sysN = initSysN;

  sys->L = L;
  sys->V = sys->L*sys->L*sys->L;
  sys->sysN = sysN;
  sys->NJ = sys->sysN*sys->V/2;
  sys->NS = sys->sysN*sys->V/2;
      
  if(masksInit(sys)){
    fprintf(stderr,"\nError in masksInit\n");
    return 1;
  }
  if(MCWeightsInit(sys)){
    fprintf(stderr,"\nError in MCWeightsInit\n");
    return 1;
  }
  if(dMemInit(sys)){
    fprintf(stderr,"\nError in dMemInit\n");
    return 1;
  }
  if(hMemInit2(sys)){
    fprintf(stderr,"\nError in hMemInit\n");
    return 1;
  }
  if(H2DSysCopy(sys)){
    fprintf(stderr,"\nError in H2DSysCopy\n");
    return 1;
  }

  ////////////////////////////////////////
  // NO-GRID KERNELS
  ///////////////////////////////////////
  
  int blockInit = initBlock;
  sys->blockSize = blockInit;

  int k = spinsPerThread;
	    	    
  dim3 block(sys->blockSize,1,1);

  int fitGrid = (sys->V/2/k + sys->blockSize - 1)/sys->blockSize;
  dim3 grid(fitGrid, sys->sysN, 1);
  int h_loopStride = block.x*grid.x;
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_loopStride, &h_loopStride, sizeof(int)) );
	    	    
  int threads = block.x*block.y*block.z*grid.x*grid.y*grid.z;
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_threads, &threads, sizeof(int)) );
	    
  MT19937_RNGArray_t *rng = MT19937_RNGArrayInit(threads*N_MT);
  int k0 = 0, k1 = 1, k2 = M_MT;  	    
  int spins = 4*k;
  // KERNEL: Sliced

  for(int sample = 0; sample < initSamples; sample++){
  
    clock_gettime(CLOCK_MONOTONIC, &start);
  
    for(int steps = 0; steps < initSteps; steps++){	  
      MCStepReds_Sliced_MT19937<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
						   sys->d_reds2, sys->d_reds3,
						   rng->d_rngSeeds, 
						   k0, k1, k2, spins);
      k0 = (k0 + 4*k)%N_MT;
      k1 = (k1 + 4*k)%N_MT;
      k2 = (k2 + 4*k)%N_MT;

      MYCUDA_ERROR( cudaDeviceSynchronize() );
    }

    clock_gettime(CLOCK_MONOTONIC, &stop);

    swap = ((stop.tv_sec - start.tv_sec) + (stop.tv_nsec - start.tv_nsec)/1.0e9);
    redMean += (swap/(initSteps)); 
    redMean2 += (swap/(initSteps))*(swap/(initSteps));
    swap = 0.;

    clock_gettime(CLOCK_MONOTONIC, &start);
	      	      
    /*    for(int steps = 0; steps < initSteps; steps++){	  
      MCStepBlues_Sliced_MT19937<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
						    sys->d_blues2, sys->d_blues3,
						    rng->d_rngSeeds,
						    (k0 + (k<<2))%N_MT, (k1 + (k<<2))%N_MT,
						    (k2 + (k<<2))%N_MT, 4*k);

      MYCUDA_ERROR( cudaDeviceSynchronize() );
      }*/

    clock_gettime(CLOCK_MONOTONIC, &stop);

    swap = ((stop.tv_sec - start.tv_sec) + (stop.tv_nsec - start.tv_nsec)/1.0e9); 	    
    blueMean += (swap/(initSteps)); 
    blueMean2 += (swap/(initSteps))*(swap/(initSteps));
    swap = 0.;
	      
  }
	    
  redMean /= initSamples;  redMean2 /= initSamples;      
  blueMean /= initSamples; blueMean2 /= initSamples;
  
  errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
  errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
  
  snprintf(fileName, sizeof(fileName), "GPU%d_slicedKernel_MT19937_smooth", sys->cardN);
  FILE *outputFile = fopen(fileName, "a+");
  fprintf(outputFile, "%d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
	  L, sysN, k, block.x, grid.x,
	  redMean, errorRed, blueMean, errorBlue);
  fflush(outputFile);
  fclose(outputFile);

  MT19937_RNGArrayReset(rng);
  free(rng);

  if(dMemFree(sys)){
    fprintf(stderr,"\nError in dMemFree\n");
    return 1;
  }
  if(hMemFree(sys)){
    fprintf(stderr,"\nError in hMemFree\n");
    return 1;
  }
  
  return 0;
}

extern int benchmarkSliced_PR(int argc, char **argv){
  if(argc != 8){
    fprintf(stderr, "usage: %s <gpu> <L> <sysN> <spinsPerThread> <block> <samples> <steps>\n\n", argv[0]);
    fprintf(stderr, "Systems' GPUs are:\n");
    int numberOfGPUs = 0;
    MYCUDA_ERROR( cudaGetDeviceCount(&numberOfGPUs) );
    for(int gpu = 0; gpu < numberOfGPUs; gpu++){
      cudaDeviceProp prop;
      MYCUDA_ERROR( cudaGetDeviceProperties(&prop, gpu) );
      printf("GPU %d: %s\n", gpu, prop.name);
    }
    return 1;
  }

  char fileName[1024];
  struct timespec start, stop;
  double redMean = 0., redMean2 = 0., swap = 0.;
  double blueMean = 0., blueMean2 = 0.;
  double errorRed = 0., errorBlue = 0.;


  sysEA3D_t *sys = (sysEA3D_t *)malloc(sizeof(sysEA3D_t));
  int initL, initSysN, initSamples, initSteps, initBlock, spinsPerThread;

  sscanf(argv[1], "%d", &sys->cardN);
  sscanf(argv[2], "%d", &initL);
  sscanf(argv[3], "%d", &initSysN);
  sscanf(argv[4], "%d", &spinsPerThread);
  sscanf(argv[5], "%d", &initBlock);
  sscanf(argv[6], "%d", &initSamples);
  sscanf(argv[7], "%d", &initSteps);

  sys->beta = 1.101;

  int numberOfGPUs = 0;
  cudaGetDeviceCount(&numberOfGPUs);
  if(sys->cardN >= numberOfGPUs){
    fprintf(stderr, "gpu %d >= gpuN %d\n", sys->cardN, numberOfGPUs);
    return 1;
  }

  MYCUDA_ERROR( cudaSetDevice(sys->cardN) );
  cudaDeviceProp prop;
  MYCUDA_ERROR( cudaGetDeviceProperties(&prop, sys->cardN) );
  printf("GPU %d: %s Compute Capability: %d.%d\n", sys->cardN, prop.name, prop.major, prop.minor);

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Sliced_MT19937, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Sliced_MT19937, cudaFuncCachePreferL1) );

  int L = initL;

  int sysN = initSysN;

  sys->L = L;
  sys->V = sys->L*sys->L*sys->L;
  sys->sysN = sysN;
  sys->NJ = sys->sysN*sys->V/2;
  sys->NS = sys->sysN*sys->V/2;
      
  if(masksInit(sys)){
    fprintf(stderr,"\nError in masksInit\n");
    return 1;
  }
  if(MCWeightsInit(sys)){
    fprintf(stderr,"\nError in MCWeightsInit\n");
    return 1;
  }
  if(dMemInit(sys)){
    fprintf(stderr,"\nError in dMemInit\n");
    return 1;
  }
  if(hMemInit2(sys)){
    fprintf(stderr,"\nError in hMemInit\n");
    return 1;
  }
  if(H2DSysCopy(sys)){
    fprintf(stderr,"\nError in H2DSysCopy\n");
    return 1;
  }

  ////////////////////////////////////////
  // NO-GRID KERNELS
  ///////////////////////////////////////
  
  int blockInit = initBlock;
  sys->blockSize = blockInit;

  int k = spinsPerThread;
	    	    
  dim3 block(sys->blockSize,1,1);

  int fitGrid = (sys->V/2/k + sys->blockSize - 1)/sys->blockSize;
  dim3 grid(fitGrid, sys->sysN, 1);
  int h_loopStride = block.x*grid.x;
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_loopStride, &h_loopStride, sizeof(int)) );
	    	    
  int threads = block.x*block.y*block.z*grid.x*grid.y*grid.z;
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_threads, &threads, sizeof(int)) );
	    
  PR_RNGArray_t *rng = PR_RNGArrayInit(threads*PRWHEEL, 0);
  int ip0 = 61, ip1 = (61 - 24), ip2 = (61 - 57), ip3 = 0;  	    
  int spins = 4*k;
  // KERNEL: Sliced

  for(int sample = 0; sample < initSamples; sample++){
  
    clock_gettime(CLOCK_MONOTONIC, &start);
  
    for(int steps = 0; steps < initSteps; steps++){	  
      MCStepReds_Sliced_PR<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
					      sys->d_reds2, sys->d_reds3,
					      rng->d_rngSeeds, 
					      ip0, ip1, ip2, ip3, spins);
      ip0 = (ip0 + 4*k)&PRWHEELMASK;
      ip1 = (ip1 + 4*k)&PRWHEELMASK;
      ip2 = (ip2 + 4*k)&PRWHEELMASK;
      ip2 = (ip2 + 4*k)&PRWHEELMASK;

      MYCUDA_ERROR( cudaDeviceSynchronize() );
    }

    clock_gettime(CLOCK_MONOTONIC, &stop);

    swap = ((stop.tv_sec - start.tv_sec) + (stop.tv_nsec - start.tv_nsec)/1.0e9);
    redMean += (swap/(initSteps)); 
    redMean2 += (swap/(initSteps))*(swap/(initSteps));
    swap = 0.;

    clock_gettime(CLOCK_MONOTONIC, &start);
	      	      
    for(int steps = 0; steps < initSteps; steps++){	  
      MCStepBlues_Sliced_PR<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
					       sys->d_blues2, sys->d_blues3,
					       rng->d_rngSeeds,
					       ip0, ip1, ip2, ip3, spins);

      ip0 = (ip0 + 4*k)&PRWHEELMASK;
      ip1 = (ip1 + 4*k)&PRWHEELMASK;
      ip2 = (ip2 + 4*k)&PRWHEELMASK;
      ip2 = (ip2 + 4*k)&PRWHEELMASK;

      MYCUDA_ERROR( cudaDeviceSynchronize() );
    }

    clock_gettime(CLOCK_MONOTONIC, &stop);

    swap = ((stop.tv_sec - start.tv_sec) + (stop.tv_nsec - start.tv_nsec)/1.0e9); 	    
    blueMean += (swap/(initSteps)); 
    blueMean2 += (swap/(initSteps))*(swap/(initSteps));
    swap = 0.;
	      
  }
	    
  redMean /= initSamples;  redMean2 /= initSamples;      
  blueMean /= initSamples; blueMean2 /= initSamples;
  
  errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
  errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
  
  snprintf(fileName, sizeof(fileName), "GPU%d_slicedKernel_PR_smooth", sys->cardN);
  FILE *outputFile = fopen(fileName, "a+");
  fprintf(outputFile, "%d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
	  L, sysN, k, block.x, grid.x,
	  redMean, errorRed, blueMean, errorBlue);
  fflush(outputFile);
  fclose(outputFile);

  PR_RNGArrayReset(rng);
  free(rng);

  if(dMemFree(sys)){
    fprintf(stderr,"\nError in dMemFree\n");
    return 1;
  }
  if(hMemFree(sys)){
    fprintf(stderr,"\nError in hMemFree\n");
    return 1;
  }
  
  return 0;
}

extern int benchmarkSmooth_PR(int argc, char **argv){
  if(argc != 9){
    fprintf(stderr, "usage: %s <gpu> <L> <sysN> <block> <samples> <steps> <memory> <spinsPerThread>\n\n", argv[0]);
    fprintf(stderr, "Systems' GPUs are:\n");
    int numberOfGPUs = 0;
    MYCUDA_ERROR( cudaGetDeviceCount(&numberOfGPUs) );
    for(int gpu = 0; gpu < numberOfGPUs; gpu++){
      cudaDeviceProp prop;
      MYCUDA_ERROR( cudaGetDeviceProperties(&prop, gpu) );
      printf("GPU %d: %s\n", gpu, prop.name);
    }
    return 1;
  }

  char fileName[1024];
  //struct timespec start, stop;

  sysEA3D_t *sys = (sysEA3D_t *)malloc(sizeof(sysEA3D_t));
  int initL, initSysN, initSamples, initSteps, initBlock, spinsPerThread;
  float memory;
  sscanf(argv[1], "%d", &sys->cardN);
  sscanf(argv[2], "%d", &initL);
  sscanf(argv[3], "%d", &initSysN);
  sscanf(argv[4], "%d", &initBlock);
  sscanf(argv[5], "%d", &initSamples);
  sscanf(argv[6], "%d", &initSteps);
  sscanf(argv[7], "%f", &memory);
  sscanf(argv[8], "%d", &spinsPerThread);

  if(initL == 0) initL = 8;
  if(initSamples == 0) initSamples = 100;
  if(initSteps == 0) initSteps = 50;

  sys->beta = 1.101;

  int numberOfGPUs = 0;
  cudaGetDeviceCount(&numberOfGPUs);
  if(sys->cardN >= numberOfGPUs){
    fprintf(stderr, "gpu %d >= gpuN %d\n", sys->cardN, numberOfGPUs);
    return 1;
  }

  MYCUDA_ERROR( cudaSetDevice(sys->cardN) );
  cudaDeviceProp prop;
  MYCUDA_ERROR( cudaGetDeviceProperties(&prop, sys->cardN) );
  printf("GPU %d: %s Compute Capability: %d.%d\n", sys->cardN, prop.name, prop.major, prop.minor);

  cudaEvent_t startEvent, stopEvent;
  MYCUDA_ERROR( cudaEventCreate(&startEvent) );
  MYCUDA_ERROR( cudaEventCreate(&stopEvent) );

  int maxThreadsPerBlock = 0;
  if(prop.major == 2) maxThreadsPerBlock = 512;
  if(prop.major == 3) maxThreadsPerBlock = 1024;

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Bitwise_PR, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Bitwise_PR, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Standard_PR, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Standard_PR, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Sliced_PR, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Sliced_PR, cudaFuncCachePreferL1) );


  for(int L = initL; L <= 256; L += 2){

    int hV = L*L*L/2;
    size_t NJ_Size = 6*hV*sizeof(MSC), NS_Size = 4*2*hV*sizeof(MSC), RNG_Size = PRWHEEL*hV*sizeof(RNGT);
    size_t devFree = 0, devTotal = 0;
    MYCUDA_ERROR( cudaMemGetInfo(&devFree, &devTotal) );
    // here to decide the percentage of device memory to use. 
    int MaxSysN = (int)(memory*(devFree/(NJ_Size + NS_Size + RNG_Size)));
    printf("\n\nL = %d, Free memory on device: %lu, MaxSysN: %d\n", L, devFree, MaxSysN);

    if(initSysN == 0) initSysN = 1;

    for(int sysN = initSysN; sysN <= MaxSysN; sysN *= 2){
      sys->L = L;
      sys->V = sys->L*sys->L*sys->L;
      sys->sysN = sysN;
      sys->NJ = sys->sysN*sys->V/2;
      sys->NS = sys->sysN*sys->V/2;
      
      if(masksInit(sys)){
	fprintf(stderr,"\nError in masksInit\n");
	return 1;
      }
      if(MCWeightsInit(sys)){
	fprintf(stderr,"\nError in MCWeightsInit\n");
	return 1;
      }
      if(dMemInit(sys)){
	fprintf(stderr,"\nError in dMemInit\n");
	return 1;
      }
      if(hMemInit2(sys)){
	fprintf(stderr,"\nError in hMemInit\n");
	return 1;
      }
      if(H2DSysCopy(sys)){
	fprintf(stderr,"\nError in H2DSysCopy\n");
	return 1;
      }

      MYCUDA_ERROR( cudaMemGetInfo(&devFree, &devTotal) );
      printf("\n\n---\tL = %d, sysN = %d, free memory after allocation: %lu\t---\n\n",
	     L, sysN, devFree);

      ////////////////////////////////////////
      // NO-GRID KERNELS
      ///////////////////////////////////////

      int blockInit = 0;

      if(initBlock == 0){
	blockInit = maxThreadsPerBlock;
	while(blockInit > sys->V/2) blockInit /= 2;
      }


      for(sys->blockSize = blockInit; sys->blockSize >= 32; sys->blockSize /= 2){

	//printf("blockSize: %d\n", sys->blockSize);

	if(sys->blockSize >= 32){
	  int kMax = sys->V/sys->blockSize/2;
	  if(kMax > spinsPerThread && spinsPerThread != 0) kMax = spinsPerThread;
	  //printf("blockSizeIF: %d\n", sys->blockSize);

	  for(int k = 0; k < kMax ;){
	    
	    float redMean = 0., redMean2 = 0., swap = 0.;
	    float blueMean = 0., blueMean2 = 0.;
	    float errorRed = 0., errorBlue = 0.;

	    do{k++;}while(kMax%k != 0);	    
	    
	    dim3 block(sys->blockSize,1,1);

	    int fitGrid = (sys->V/2/k + sys->blockSize - 1)/sys->blockSize;
	    dim3 grid(fitGrid, sys->sysN, 1);
	    int h_loopStride = block.x*grid.x;
	    MYCUDA_ERROR( cudaMemcpyToSymbol(d_loopStride, &h_loopStride, sizeof(int)) );
	    
	    printf("Measuring Times No-Grid L:%d block.x:%d grid.x:%d spins per thread:%d\n",
		   L, block.x, grid.x, k);
	    
	    int threads = block.x*block.y*block.z*grid.x*grid.y*grid.z;
	    MYCUDA_ERROR( cudaMemcpyToSymbol(d_threads, &threads, sizeof(int)) );
	    
	    PR_RNGArray_t *rng = PR_RNGArrayInit(threads*PRWHEEL,0);
	    int ip0 = 61, ip1 = (61 - 24), ip2 = (61 - 57), ip3 = 0;  	    
	    int spins = 4*k;

	    // KERNEL: Bitwise; Only L = 2^k are examined
	    if( ((1U<<findLog2(sys->L))^sys->L) == 0 ){
	      for(int sample = 0; sample < initSamples; sample++){
		
		MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
		
		for(int steps = 0; steps < initSteps; steps++){	  
		  MCStepReds_Bitwise_PR<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
							   sys->d_reds2, sys->d_reds3,
							   rng->d_rngSeeds,
							   ip0, ip1, ip2, ip3, spins);
		  ip0 = (ip0 + 4*k)&PRWHEELMASK;
		  ip1 = (ip1 + 4*k)&PRWHEELMASK;
		  ip2 = (ip2 + 4*k)&PRWHEELMASK;
		  ip2 = (ip2 + 4*k)&PRWHEELMASK;
							   
		  MYCUDA_ERROR( cudaDeviceSynchronize() );
		}
		
		MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
		MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
		MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
		redMean += (swap/(initSteps)); 
		redMean2 += (swap/(initSteps))*(swap/(initSteps));
		swap = 0.;
		
		MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
		
		for(int steps = 0; steps < initSteps; steps++){	  
		  MCStepBlues_Bitwise_PR<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
							    sys->d_blues2, sys->d_blues3,
							    rng->d_rngSeeds,
							    ip0, ip1, ip2, ip3, spins);
		  ip0 = (ip0 + 4*k)&PRWHEELMASK;
		  ip1 = (ip1 + 4*k)&PRWHEELMASK;
		  ip2 = (ip2 + 4*k)&PRWHEELMASK;
		  ip2 = (ip2 + 4*k)&PRWHEELMASK;
							    
		  MYCUDA_ERROR( cudaDeviceSynchronize() );
		}
		
		MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
		MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
		MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
		blueMean += (swap/(initSteps)); 
		blueMean2 += (swap/(initSteps))*(swap/(initSteps));
		swap = 0.;
	      }
	      
	      redMean /= initSamples;  redMean2 /= initSamples;      
	      blueMean /= initSamples; blueMean2 /= initSamples;
	      
	      errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
	      errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
	      
	      snprintf(fileName, sizeof(fileName), "GPU%d_bitwiseKernel_PR_smooth", sys->cardN);
	      FILE *outputFile = fopen(fileName, "a+");
	      fprintf(outputFile, "%d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
		      L, sysN, k, block.x, grid.x,
		      redMean*1.0e-3, errorRed*1.0e-3, blueMean*1.0e-3, errorBlue*1.0e-3);
	      fflush(outputFile);
	      fclose(outputFile);
	    }
	    
	    // KERNEL: Standard
	    redMean = blueMean = redMean2 = blueMean2 = 0.;
	    for(int sample = 0; sample < initSamples; sample++){
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepReds_Standard_PR<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
							  sys->d_reds2, sys->d_reds3,
							  rng->d_rngSeeds,
							  ip0, ip1, ip2, ip3, spins);
		ip0 = (ip0 + 4*k)&PRWHEELMASK;
		ip1 = (ip1 + 4*k)&PRWHEELMASK;
		ip2 = (ip2 + 4*k)&PRWHEELMASK;
		ip2 = (ip2 + 4*k)&PRWHEELMASK;

		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      redMean += (swap/(initSteps)); 
	      redMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepBlues_Standard_PR<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
							   sys->d_blues2, sys->d_blues3,
							   rng->d_rngSeeds,
							   ip0, ip1, ip2, ip3, spins);
		ip0 = (ip0 + 4*k)&PRWHEELMASK;
		ip1 = (ip1 + 4*k)&PRWHEELMASK;
		ip2 = (ip2 + 4*k)&PRWHEELMASK;
		ip2 = (ip2 + 4*k)&PRWHEELMASK;
		
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      blueMean += (swap/(initSteps)); 
	      blueMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	    }
	    
	    redMean /= initSamples;  redMean2 /= initSamples;      
	    blueMean /= initSamples; blueMean2 /= initSamples;
	    
	    errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
	    errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
	    
	    snprintf(fileName, sizeof(fileName), "GPU%d_standardKernel_PR_smooth", sys->cardN);
	    FILE *outputFile = fopen(fileName, "a+");
	    fprintf(outputFile, "%d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
		    L, sysN, k, block.x, grid.x,
		    redMean*1.0e-3, errorRed*1.0e-3, blueMean*1.0e-3, errorBlue*1.0e-3);
	    fflush(outputFile);
	    fclose(outputFile);
	    
	    // KERNEL: Sliced
	    redMean = blueMean = redMean2 = blueMean2 = 0.;
	    for(int sample = 0; sample < initSamples; sample++){
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepReds_Sliced_PR<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
							sys->d_reds2, sys->d_reds3,
							rng->d_rngSeeds,
							ip0, ip1, ip2, ip3, spins);
		ip0 = (ip0 + 4*k)&PRWHEELMASK;
		ip1 = (ip1 + 4*k)&PRWHEELMASK;
		ip2 = (ip2 + 4*k)&PRWHEELMASK;
		ip2 = (ip2 + 4*k)&PRWHEELMASK;
							
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      redMean += (swap/(initSteps)); 
	      redMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepBlues_Sliced_PR<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
							 sys->d_blues2, sys->d_blues3,
							 rng->d_rngSeeds,
							 ip0, ip1, ip2, ip3, spins);
		ip0 = (ip0 + 4*k)&PRWHEELMASK;
		ip1 = (ip1 + 4*k)&PRWHEELMASK;
		ip2 = (ip2 + 4*k)&PRWHEELMASK;
		ip2 = (ip2 + 4*k)&PRWHEELMASK;
		
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      blueMean += (swap/(initSteps)); 
	      blueMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	    }
	    
	    redMean /= initSamples;  redMean2 /= initSamples;      
	    blueMean /= initSamples; blueMean2 /= initSamples;
	    
	    errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
	    errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
	    
	    snprintf(fileName, sizeof(fileName), "GPU%d_slicedKernel_PR_smooth", sys->cardN);
	    outputFile = fopen(fileName, "a+");
	    fprintf(outputFile, "%d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
		    L, sysN, k, block.x, grid.x,
		    redMean*1.0e-3, errorRed*1.0e-3, blueMean*1.0e-3, errorBlue*1.0e-3);
	    fflush(outputFile);
	    fclose(outputFile);
	  	  
	    PR_RNGArrayReset(rng);
	    free(rng);
	  }
	}
      }

      // CLEANING UP
      //CRNGArrayFree(&rng);
      if(dMemFree(sys)){
	fprintf(stderr,"\nError in dMemFree\n");
	return 1;
      }
      if(hMemFree(sys)){
	fprintf(stderr,"\nError in hMemFree\n");
	return 1;
      }
    }

    initSysN = 0;
  }
  // allocazione
  // deallocazione
  // linearizzazione nel numero dei sistemi con limite la memoria
  return 0;
}

extern int benchmarkGrid_PR(int argc, char **argv){
  if(argc != 8){
    fprintf(stderr, "usage: %s <gpu> <L> <sysN> <block> <samples> <steps> <memory>\n\n", argv[0]);
    fprintf(stderr, "Systems' GPUs are:\n");
    int numberOfGPUs = 0;
    MYCUDA_ERROR( cudaGetDeviceCount(&numberOfGPUs) );
    for(int gpu = 0; gpu < numberOfGPUs; gpu++){
      cudaDeviceProp prop;
      MYCUDA_ERROR( cudaGetDeviceProperties(&prop, gpu) );
      printf("GPU %d: %s\n", gpu, prop.name);
    }
    return 1;
  }

  char fileName[1024];
  //struct timespec start, stop;

  sysEA3D_t *sys = (sysEA3D_t *)malloc(sizeof(sysEA3D_t));
  int initL, initSysN, initSamples, initSteps, initBlock;
  float memory;
  sscanf(argv[1], "%d", &sys->cardN);
  sscanf(argv[2], "%d", &initL);
  sscanf(argv[3], "%d", &initSysN);
  sscanf(argv[4], "%d", &initBlock);
  sscanf(argv[5], "%d", &initSamples);
  sscanf(argv[6], "%d", &initSteps);
  sscanf(argv[7], "%f", &memory);

  if(initL == 0) initL = 8;
  if(initSamples == 0) initSamples = 100;
  if(initSteps == 0) initSteps = 50;

  sys->beta = 1.101;

  int numberOfGPUs = 0;
  cudaGetDeviceCount(&numberOfGPUs);
  if(sys->cardN >= numberOfGPUs){
    fprintf(stderr, "gpu %d >= gpuN %d\n", sys->cardN, numberOfGPUs);
    return 1;
  }

  MYCUDA_ERROR( cudaSetDevice(sys->cardN) );
  cudaDeviceProp prop;
  MYCUDA_ERROR( cudaGetDeviceProperties(&prop, sys->cardN) );
  printf("GPU %d: %s Compute Capability: %d.%d\n", sys->cardN, prop.name, prop.major, prop.minor);

  cudaEvent_t startEvent, stopEvent;
  MYCUDA_ERROR( cudaEventCreate(&startEvent) );
  MYCUDA_ERROR( cudaEventCreate(&stopEvent) );

  int maxThreadsPerBlock = 0;
  if(prop.major == 2) maxThreadsPerBlock = 512;
  if(prop.major == 3) maxThreadsPerBlock = 1024;

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_StandardGrid_PR, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_StandardGrid_PR, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_SlicedGrid_PR, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_SlicedGrid_PR, cudaFuncCachePreferL1) );


  for(int L = initL; L <= 256; L += 2){

    int hV = L*L*L/2;
    size_t NJ_Size = 6*hV*sizeof(MSC), NS_Size = 4*2*hV*sizeof(MSC), RNG_Size = PRWHEEL*hV*sizeof(RNGT);
    size_t devFree = 0, devTotal = 0;
    MYCUDA_ERROR( cudaMemGetInfo(&devFree, &devTotal) );
    // here to decide the percentage of device memory to use. 
    int MaxSysN = (int)(memory*(devFree/(NJ_Size + NS_Size + RNG_Size)));
    printf("\n\nL = %d, Free memory on device: %lu, MaxSysN: %d\n", L, devFree, MaxSysN);

    if(initSysN == 0) initSysN = 1;

    for(int sysN = initSysN; sysN <= MaxSysN; sysN *= 2){
      sys->L = L;
      sys->V = sys->L*sys->L*sys->L;
      sys->sysN = sysN;
      sys->NJ = sys->sysN*sys->V/2;
      sys->NS = sys->sysN*sys->V/2;
      
      if(masksInit(sys)){
	fprintf(stderr,"\nError in masksInit\n");
	return 1;
      }
      if(MCWeightsInit(sys)){
	fprintf(stderr,"\nError in MCWeightsInit\n");
	return 1;
      }
      if(dMemInit(sys)){
	fprintf(stderr,"\nError in dMemInit\n");
	return 1;
      }
      if(hMemInit2(sys)){
	fprintf(stderr,"\nError in hMemInit\n");
	return 1;
      }
      if(H2DSysCopy(sys)){
	fprintf(stderr,"\nError in H2DSysCopy\n");
	return 1;
      }

      MYCUDA_ERROR( cudaMemGetInfo(&devFree, &devTotal) );
      printf("\n\n---\tL = %d, sysN = %d, free memory after allocation: %lu\t---\n\n",
	     L, sysN, devFree);

      ////////////////////////////////////////
      // GRID KERNELS
      ///////////////////////////////////////
	  
      int A = sys->L*sys->L;
      int hA = A/2;

      int count3 = 0, test3 = 1;

      for(sys->k = 1; test3; ){

	float redMean = 0., redMean2 = 0., swap = 0.;
	float blueMean = 0., blueMean2 = 0.;
	float errorRed = 0., errorBlue = 0.;

	if(count3 > 0) do{sys->k++;}while(sys->L%sys->k != 0 && sys->k <= sys->L);
	count3++;

	dim3 blockS(sys->L/2, sys->k, 1);
	dim3 gridS(hA/(blockS.x*blockS.y), sys->L, sys->sysN);
	dim3 blockG(sys->L, sys->k, 1);
	dim3 gridG(A/(blockG.x*blockG.y), sys->L/2, sys->sysN);

	int threads = blockS.x*blockS.y*blockS.z*gridS.x*gridS.y*gridS.z;
	PR_RNGArray_t *rng;

	if(threads > 0) rng = PR_RNGArrayInit(threads, 0);
	int ip0 = 61, ip1 = (61 - 24), ip2 = (61 - 57), ip3 = 0;  	    
	int spins = 4;


	if(sys->k*sys->L <= maxThreadsPerBlock && sys->k <= sys->L && sys->k*sys->L >= 32){	  
	  printf("Measuring Times Grid    L:%d lines:%d\n",
		 L, sys->k);
	}


	if(sys->k*sys->L/2 <= maxThreadsPerBlock && sys->k <= sys->L && sys->k*sys->L/2 >= 32){
	  // KERNEL: Standard
	  	  	  
	  int h_block_xStrideS = blockS.x*blockS.y;
	  int h_block_yStrideS = gridS.x*h_block_xStrideS;
	  
	  MYCUDA_ERROR( cudaMemcpyToSymbol(d_block_xStrideS, &h_block_xStrideS, sizeof(int)) );
	  MYCUDA_ERROR( cudaMemcpyToSymbol(d_block_yStrideS, &h_block_yStrideS, sizeof(int)) );	  
	  
	  for(int sample = 0; sample < initSamples; sample++){
	    
	    MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	    
	    for(int steps = 0; steps < initSteps; steps++){	  
	      MCStepReds_StandardGrid_PR<<< gridS, blockS >>>(sys->d_reds0, sys->d_reds1,
							      sys->d_reds2, sys->d_reds3,
							      rng->d_rngSeeds,
							      ip0, ip1, ip2, ip3, spins);
	      ip0 = (ip0 + 4)&PRWHEELMASK;
	      ip1 = (ip1 + 4)&PRWHEELMASK;
	      ip2 = (ip2 + 4)&PRWHEELMASK;
	      ip2 = (ip2 + 4)&PRWHEELMASK;
	      
	      MYCUDA_ERROR( cudaDeviceSynchronize() );
	    }
	    
	    MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	    MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	    MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	    redMean += (swap/(initSteps)); 
	    redMean2 += (swap/(initSteps))*(swap/(initSteps));
	    swap = 0.;
	    
	    MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	    
	    for(int steps = 0; steps < initSteps; steps++){	  
	      MCStepBlues_StandardGrid_PR<<< gridS, blockS >>>(sys->d_blues0, sys->d_blues1,
							       sys->d_blues2, sys->d_blues3,
							       rng->d_rngSeeds,
							       ip0, ip1, ip2, ip3, spins);
	      ip0 = (ip0 + 4)&PRWHEELMASK;
	      ip1 = (ip1 + 4)&PRWHEELMASK;
	      ip2 = (ip2 + 4)&PRWHEELMASK;
	      ip2 = (ip2 + 4)&PRWHEELMASK;

	      MYCUDA_ERROR( cudaDeviceSynchronize() );
	    }
	    
	    MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	    MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	    MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	    blueMean += (swap/(initSteps)); 
	    blueMean2 += (swap/(initSteps))*(swap/(initSteps));
	    swap = 0.;
	  }
	  
	  redMean /= initSamples;  redMean2 /= initSamples;      
	  blueMean /= initSamples; blueMean2 /= initSamples;
	  
	  errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
	  errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
	  
	  snprintf(fileName, sizeof(fileName), "GPU%d_standardGridKernel_PR", sys->cardN);
	  FILE *outputFile = fopen(fileName, "a+");
	  fprintf(outputFile, "%d %d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
		  L, sysN, sys->k, blockS.x, blockS.y, gridS.x,
		  redMean*1.0e-3, errorRed*1.0e-3, blueMean*1.0e-3, errorBlue*1.0e-3);
	  fflush(outputFile);
	  fclose(outputFile);
	}

	if(sys->k*sys->L <= maxThreadsPerBlock && sys->k <= sys->L){	  
	  if(sys->k*sys->L >= 32){
	    // KERNEL: Sliced
	    redMean = blueMean = redMean2 = blueMean2 = 0.;
	    
	    int h_block_xStride = blockG.x*blockG.y;
	    int h_block_yStride = gridG.x*h_block_xStride;
	    
	    MYCUDA_ERROR( cudaMemcpyToSymbol(d_block_xStride, &h_block_xStride, sizeof(int)) );
	    MYCUDA_ERROR( cudaMemcpyToSymbol(d_block_yStride, &h_block_yStride, sizeof(int)) );
	    
	    for(int sample = 0; sample < initSamples; sample++){
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	    
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepReds_SlicedGrid_PR<<< gridG, blockG >>>(sys->d_reds0, sys->d_reds1,
							      sys->d_reds2, sys->d_reds3,
							      rng->d_rngSeeds,
							      ip0, ip1, ip2, ip3, spins);
		ip0 = (ip0 + 4)&PRWHEELMASK;
		ip1 = (ip1 + 4)&PRWHEELMASK;
		ip2 = (ip2 + 4)&PRWHEELMASK;
		ip2 = (ip2 + 4)&PRWHEELMASK;

		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      redMean += (swap/(initSteps)); 
	      redMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepBlues_SlicedGrid_PR<<< gridG, blockG >>>(sys->d_blues0, sys->d_blues1,
							       sys->d_blues2, sys->d_blues3,
							       rng->d_rngSeeds,
							       ip0, ip1, ip2, ip3, spins);
		ip0 = (ip0 + 4)&PRWHEELMASK;
		ip1 = (ip1 + 4)&PRWHEELMASK;
		ip2 = (ip2 + 4)&PRWHEELMASK;
		ip2 = (ip2 + 4)&PRWHEELMASK;

		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      blueMean += (swap/(initSteps)); 
	      blueMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	    }
	    
	    redMean /= initSamples;  redMean2 /= initSamples;      
	    blueMean /= initSamples; blueMean2 /= initSamples;
	    
	    errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
	    errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
	    
	    snprintf(fileName, sizeof(fileName), "GPU%d_slicedGridKernel_PR", sys->cardN);
	    FILE *outputFile = fopen(fileName, "a+");
	    fprintf(outputFile, "%d %d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
		    L, sysN, sys->k, blockG.x, blockG.y, gridG.x,
		    redMean*1.0e-3, errorRed*1.0e-3, blueMean*1.0e-3, errorBlue*1.0e-3);
	    fflush(outputFile);
	    fclose(outputFile);
	  }
	} else test3 = 0;
	
	if(threads > 0){
	  PR_RNGArrayReset(rng);
	  free(rng);
	}
      }
      // CLEANING UP
      //CRNGArrayFree(&rng);
      if(dMemFree(sys)){
	fprintf(stderr,"\nError in dMemFree\n");
	return 1;
      }
      if(hMemFree(sys)){
	fprintf(stderr,"\nError in hMemFree\n");
	return 1;
      }
    }

    initSysN = 0;
  }

  return 0;
}

extern int benchmarkSmooth_MT19937(int argc, char **argv){
  if(argc != 9){
    fprintf(stderr, "usage: %s <gpu> <L> <sysN> <block> <samples> <steps> <memory> <spinsPerThread>\n\n", argv[0]);
    fprintf(stderr, "Systems' GPUs are:\n");
    int numberOfGPUs = 0;
    MYCUDA_ERROR( cudaGetDeviceCount(&numberOfGPUs) );
    for(int gpu = 0; gpu < numberOfGPUs; gpu++){
      cudaDeviceProp prop;
      MYCUDA_ERROR( cudaGetDeviceProperties(&prop, gpu) );
      printf("GPU %d: %s\n", gpu, prop.name);
    }
    return 1;
  }

  char fileName[1024];
  //struct timespec start, stop;

  sysEA3D_t *sys = (sysEA3D_t *)malloc(sizeof(sysEA3D_t));
  int initL, initSysN, initSamples, initSteps, initBlock, spinsPerThread;
  float memory;
  sscanf(argv[1], "%d", &sys->cardN);
  sscanf(argv[2], "%d", &initL);
  sscanf(argv[3], "%d", &initSysN);
  sscanf(argv[4], "%d", &initBlock);
  sscanf(argv[5], "%d", &initSamples);
  sscanf(argv[6], "%d", &initSteps);
  sscanf(argv[7], "%f", &memory);
  sscanf(argv[8], "%d", &spinsPerThread);

  if(initL == 0) initL = 8;
  if(initSamples == 0) initSamples = 100;
  if(initSteps == 0) initSteps = 50;

  sys->beta = 1.101;

  int numberOfGPUs = 0;
  cudaGetDeviceCount(&numberOfGPUs);
  if(sys->cardN >= numberOfGPUs){
    fprintf(stderr, "gpu %d >= gpuN %d\n", sys->cardN, numberOfGPUs);
    return 1;
  }

  MYCUDA_ERROR( cudaSetDevice(sys->cardN) );
  cudaDeviceProp prop;
  MYCUDA_ERROR( cudaGetDeviceProperties(&prop, sys->cardN) );
  printf("GPU %d: %s Compute Capability: %d.%d\n", sys->cardN, prop.name, prop.major, prop.minor);

  cudaEvent_t startEvent, stopEvent;
  MYCUDA_ERROR( cudaEventCreate(&startEvent) );
  MYCUDA_ERROR( cudaEventCreate(&stopEvent) );

  int maxThreadsPerBlock = 0;
  if(prop.major == 2) maxThreadsPerBlock = 512;
  if(prop.major == 3) maxThreadsPerBlock = 1024;

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Bitwise_MT19937, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Bitwise_MT19937, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Standard_MT19937, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Standard_MT19937, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Sliced_MT19937, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Sliced_MT19937, cudaFuncCachePreferL1) );


  for(int L = initL; L <= 256; L += 2){

    int hV = L*L*L/2;
    size_t NJ_Size = 6*hV*sizeof(MSC), NS_Size = 4*2*hV*sizeof(MSC), RNG_Size = N_MT*hV*sizeof(RNGT);
    size_t devFree = 0, devTotal = 0;
    MYCUDA_ERROR( cudaMemGetInfo(&devFree, &devTotal) );
    // here to decide the percentage of device memory to use. 
    int MaxSysN = (int)(memory*(devFree/(NJ_Size + NS_Size + RNG_Size)));
    printf("\n\nL = %d, Free memory on device: %lu, MaxSysN: %d\n", L, devFree, MaxSysN);

    if(initSysN == 0) initSysN = 1;

    for(int sysN = initSysN; sysN <= MaxSysN; sysN *= 2){
      sys->L = L;
      sys->V = sys->L*sys->L*sys->L;
      sys->sysN = sysN;
      sys->NJ = sys->sysN*sys->V/2;
      sys->NS = sys->sysN*sys->V/2;
      
      if(masksInit(sys)){
	fprintf(stderr,"\nError in masksInit\n");
	return 1;
      }
      if(MCWeightsInit(sys)){
	fprintf(stderr,"\nError in MCWeightsInit\n");
	return 1;
      }
      if(dMemInit(sys)){
	fprintf(stderr,"\nError in dMemInit\n");
	return 1;
      }
      if(hMemInit2(sys)){
	fprintf(stderr,"\nError in hMemInit\n");
	return 1;
      }
      if(H2DSysCopy(sys)){
	fprintf(stderr,"\nError in H2DSysCopy\n");
	return 1;
      }

      MYCUDA_ERROR( cudaMemGetInfo(&devFree, &devTotal) );
      printf("\n\n---\tL = %d, sysN = %d, free memory after allocation: %lu\t---\n\n",
	     L, sysN, devFree);

      ////////////////////////////////////////
      // NO-GRID KERNELS
      ///////////////////////////////////////

      int blockInit = 0;

      if(initBlock == 0){
	blockInit = maxThreadsPerBlock;
	while(blockInit > sys->V/2) blockInit /= 2;
      }


      for(sys->blockSize = blockInit; sys->blockSize >= 32; sys->blockSize /= 2){

	//printf("blockSize: %d\n", sys->blockSize);

	if(sys->blockSize >= 32){
	  int kMax = sys->V/sys->blockSize/2;
	  if(kMax > spinsPerThread && spinsPerThread != 0) kMax = spinsPerThread;
	  //printf("blockSizeIF: %d\n", sys->blockSize);

	  for(int k = 0; k < kMax ;){
	    
	    float redMean = 0., redMean2 = 0., swap = 0.;
	    float blueMean = 0., blueMean2 = 0.;
	    float errorRed = 0., errorBlue = 0.;

	    do{k++;}while(kMax%k != 0);	    
	    
	    dim3 block(sys->blockSize,1,1);

	    int fitGrid = (sys->V/2/k + sys->blockSize - 1)/sys->blockSize;
	    dim3 grid(fitGrid, sys->sysN, 1);
	    int h_loopStride = block.x*grid.x;
	    MYCUDA_ERROR( cudaMemcpyToSymbol(d_loopStride, &h_loopStride, sizeof(int)) );
	    
	    printf("Measuring Times No-Grid L:%d block.x:%d grid.x:%d spins per thread:%d\n",
		   L, block.x, grid.x, k);
	    
	    int threads = block.x*block.y*block.z*grid.x*grid.y*grid.z;
	    MYCUDA_ERROR( cudaMemcpyToSymbol(d_threads, &threads, sizeof(int)) );
	    
	    MT19937_RNGArray_t *rng = MT19937_RNGArrayInit(threads*N_MT);
	    int k0 = 0, k1 = 1, k2 = M_MT;  	    
	    int spins = 4*k;

	    // KERNEL: Bitwise; Only L = 2^k are examined
	    if( ((1U<<findLog2(sys->L))^sys->L) == 0 ){
	      for(int sample = 0; sample < initSamples; sample++){
		
		MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
		
		for(int steps = 0; steps < initSteps; steps++){	  
		  MCStepReds_Bitwise_MT19937<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
								sys->d_reds2, sys->d_reds3,
								rng->d_rngSeeds,
								k0, k1, k2, spins);
		  k0 = (k0 + 4*k)%N_MT;
		  k1 = (k1 + 4*k)%N_MT;
		  k2 = (k2 + 4*k)%N_MT;
							   
		  MYCUDA_ERROR( cudaDeviceSynchronize() );
		}
		
		MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
		MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
		MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
		redMean += (swap/(initSteps)); 
		redMean2 += (swap/(initSteps))*(swap/(initSteps));
		swap = 0.;
		
		MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
		
		for(int steps = 0; steps < initSteps; steps++){	  
		  MCStepBlues_Bitwise_MT19937<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
								 sys->d_blues2, sys->d_blues3,
								 rng->d_rngSeeds,
								 k0, k1, k2, spins);

		  k0 = (k0 + 4*k)%N_MT;
		  k1 = (k1 + 4*k)%N_MT;
		  k2 = (k2 + 4*k)%N_MT;
							    
		  MYCUDA_ERROR( cudaDeviceSynchronize() );
		}
		
		MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
		MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
		MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
		blueMean += (swap/(initSteps)); 
		blueMean2 += (swap/(initSteps))*(swap/(initSteps));
		swap = 0.;
	      }
	      
	      redMean /= initSamples;  redMean2 /= initSamples;      
	      blueMean /= initSamples; blueMean2 /= initSamples;
	      
	      errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
	      errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
	      
	      snprintf(fileName, sizeof(fileName), "GPU%d_bitwiseKernel_PR_smooth", sys->cardN);
	      FILE *outputFile = fopen(fileName, "a+");
	      fprintf(outputFile, "%d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
		      L, sysN, k, block.x, grid.x,
		      redMean*1.0e-3, errorRed*1.0e-3, blueMean*1.0e-3, errorBlue*1.0e-3);
	      fflush(outputFile);
	      fclose(outputFile);
	    }
	    
	    // KERNEL: Standard
	    redMean = blueMean = redMean2 = blueMean2 = 0.;
	    for(int sample = 0; sample < initSamples; sample++){
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepReds_Standard_MT19937<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
							       sys->d_reds2, sys->d_reds3,
							       rng->d_rngSeeds,
							       k0, k1, k2, spins);
		k0 = (k0 + 4*k)%N_MT;
		k1 = (k1 + 4*k)%N_MT;
		k2 = (k2 + 4*k)%N_MT;

		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      redMean += (swap/(initSteps)); 
	      redMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepBlues_Standard_MT19937<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
								sys->d_blues2, sys->d_blues3,
								rng->d_rngSeeds,
								k0, k1, k2, spins);
		k0 = (k0 + 4*k)%N_MT;
		k1 = (k1 + 4*k)%N_MT;
		k2 = (k2 + 4*k)%N_MT;
		
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      blueMean += (swap/(initSteps)); 
	      blueMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	    }
	    
	    redMean /= initSamples;  redMean2 /= initSamples;      
	    blueMean /= initSamples; blueMean2 /= initSamples;
	    
	    errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
	    errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
	    
	    snprintf(fileName, sizeof(fileName), "GPU%d_standardKernel_PR_smooth", sys->cardN);
	    FILE *outputFile = fopen(fileName, "a+");
	    fprintf(outputFile, "%d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
		    L, sysN, k, block.x, grid.x,
		    redMean*1.0e-3, errorRed*1.0e-3, blueMean*1.0e-3, errorBlue*1.0e-3);
	    fflush(outputFile);
	    fclose(outputFile);
	    
	    // KERNEL: Sliced
	    redMean = blueMean = redMean2 = blueMean2 = 0.;
	    for(int sample = 0; sample < initSamples; sample++){
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepReds_Sliced_MT19937<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
							     sys->d_reds2, sys->d_reds3,
							     rng->d_rngSeeds,
							     k0, k1, k2, spins);
		k0 = (k0 + 4*k)%N_MT;
		k1 = (k1 + 4*k)%N_MT;
		k2 = (k2 + 4*k)%N_MT;
							
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      redMean += (swap/(initSteps)); 
	      redMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepBlues_Sliced_MT19937<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
							      sys->d_blues2, sys->d_blues3,
							      rng->d_rngSeeds,
							      k0, k1, k2, spins);
		k0 = (k0 + 4*k)%N_MT;
		k1 = (k1 + 4*k)%N_MT;
		k2 = (k2 + 4*k)%N_MT;
		
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      blueMean += (swap/(initSteps)); 
	      blueMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	    }
	    
	    redMean /= initSamples;  redMean2 /= initSamples;      
	    blueMean /= initSamples; blueMean2 /= initSamples;
	    
	    errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
	    errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
	    
	    snprintf(fileName, sizeof(fileName), "GPU%d_slicedKernel_PR_smooth", sys->cardN);
	    outputFile = fopen(fileName, "a+");
	    fprintf(outputFile, "%d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
		    L, sysN, k, block.x, grid.x,
		    redMean*1.0e-3, errorRed*1.0e-3, blueMean*1.0e-3, errorBlue*1.0e-3);
	    fflush(outputFile);
	    fclose(outputFile);
	  	  
	    MT19937_RNGArrayReset(rng);
	    free(rng);
	  }
	}
      }

      // CLEANING UP
      //CRNGArrayFree(&rng);
      if(dMemFree(sys)){
	fprintf(stderr,"\nError in dMemFree\n");
	return 1;
      }
      if(hMemFree(sys)){
	fprintf(stderr,"\nError in hMemFree\n");
	return 1;
      }
    }

    initSysN = 0;
  }
  // allocazione
  // deallocazione
  // linearizzazione nel numero dei sistemi con limite la memoria
  return 0;
}


__global__ void setupXORWOW_kernel(curandState *state)
{
  int id = threadIdx.x + blockIdx.x * blockDim.x;
  /* Each thread gets same seed, a different sequence 
     number, no offset */
  curand_init(1234, id, 0, &state[id]);
}


extern int benchmarkSmooth_XORWOW(int argc, char **argv){
  if(argc != 9){
    fprintf(stderr, "usage: %s <gpu> <L> <sysN> <block> <samples> <steps> <memory> <spinsPerThread>\n\n", argv[0]);
    fprintf(stderr, "Systems' GPUs are:\n");
    int numberOfGPUs = 0;
    MYCUDA_ERROR( cudaGetDeviceCount(&numberOfGPUs) );
    for(int gpu = 0; gpu < numberOfGPUs; gpu++){
      cudaDeviceProp prop;
      MYCUDA_ERROR( cudaGetDeviceProperties(&prop, gpu) );
      printf("GPU %d: %s\n", gpu, prop.name);
    }
    return 1;
  }

  char fileName[1024];
  //struct timespec start, stop;

  sysEA3D_t *sys = (sysEA3D_t *)malloc(sizeof(sysEA3D_t));
  int initL, initSysN, initSamples, initSteps, initBlock, spinsPerThread;
  float memory;
  sscanf(argv[1], "%d", &sys->cardN);
  sscanf(argv[2], "%d", &initL);
  sscanf(argv[3], "%d", &initSysN);
  sscanf(argv[4], "%d", &initBlock);
  sscanf(argv[5], "%d", &initSamples);
  sscanf(argv[6], "%d", &initSteps);
  sscanf(argv[7], "%f", &memory);
  sscanf(argv[8], "%d", &spinsPerThread);

  if(initL == 0) initL = 8;
  if(initSamples == 0) initSamples = 100;
  if(initSteps == 0) initSteps = 50;

  sys->beta = 1.101;

  int numberOfGPUs = 0;
  cudaGetDeviceCount(&numberOfGPUs);
  if(sys->cardN >= numberOfGPUs){
    fprintf(stderr, "gpu %d >= gpuN %d\n", sys->cardN, numberOfGPUs);
    return 1;
  }

  MYCUDA_ERROR( cudaSetDevice(sys->cardN) );
  cudaDeviceProp prop;
  MYCUDA_ERROR( cudaGetDeviceProperties(&prop, sys->cardN) );
  printf("GPU %d: %s Compute Capability: %d.%d\n", sys->cardN, prop.name, prop.major, prop.minor);

  cudaEvent_t startEvent, stopEvent;
  MYCUDA_ERROR( cudaEventCreate(&startEvent) );
  MYCUDA_ERROR( cudaEventCreate(&stopEvent) );

  int maxThreadsPerBlock = 0;
  if(prop.major == 2) maxThreadsPerBlock = 512;
  if(prop.major == 3) maxThreadsPerBlock = 1024;

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Bitwise_XORWOW, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Bitwise_XORWOW, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Standard_XORWOW, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Standard_XORWOW, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Sliced_XORWOW, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Sliced_XORWOW, cudaFuncCachePreferL1) );


  for(int L = initL; L <= 256; L += 2){

    int hV = L*L*L/2;
    size_t NJ_Size = 6*hV*sizeof(MSC), NS_Size = 4*2*hV*sizeof(MSC), RNG_Size = hV*sizeof(curandState);
    size_t devFree = 0, devTotal = 0;
    MYCUDA_ERROR( cudaMemGetInfo(&devFree, &devTotal) );
    // here to decide the percentage of device memory to use. 
    int MaxSysN = (int)(memory*(devFree/(NJ_Size + NS_Size + RNG_Size)));
    printf("\n\nL = %d, Free memory on device: %lu, MaxSysN: %d\n", L, devFree, MaxSysN);

    if(initSysN == 0) initSysN = 1;

    for(int sysN = initSysN; sysN <= MaxSysN; sysN *= 2){
      sys->L = L;
      sys->V = sys->L*sys->L*sys->L;
      sys->sysN = sysN;
      sys->NJ = sys->sysN*sys->V/2;
      sys->NS = sys->sysN*sys->V/2;
      
      if(masksInit(sys)){
	fprintf(stderr,"\nError in masksInit\n");
	return 1;
      }
      if(MCWeightsInit(sys)){
	fprintf(stderr,"\nError in MCWeightsInit\n");
	return 1;
      }
      if(dMemInit(sys)){
	fprintf(stderr,"\nError in dMemInit\n");
	return 1;
      }
      if(hMemInit2(sys)){
	fprintf(stderr,"\nError in hMemInit\n");
	return 1;
      }
      if(H2DSysCopy(sys)){
	fprintf(stderr,"\nError in H2DSysCopy\n");
	return 1;
      }

      MYCUDA_ERROR( cudaMemGetInfo(&devFree, &devTotal) );
      printf("\n\n---\tL = %d, sysN = %d, free memory after allocation: %lu\t---\n\n",
	     L, sysN, devFree);

      ////////////////////////////////////////
      // NO-GRID KERNELS
      ///////////////////////////////////////

      int blockInit = 0;

      if(initBlock == 0){
	blockInit = maxThreadsPerBlock;
	while(blockInit > sys->V/2) blockInit /= 2;
      }


      for(sys->blockSize = blockInit; sys->blockSize >= 32; sys->blockSize /= 2){

	//printf("blockSize: %d\n", sys->blockSize);

	if(sys->blockSize >= 32){
	  int kMax = sys->V/sys->blockSize/2;
	  if(kMax > spinsPerThread && spinsPerThread != 0) kMax = spinsPerThread;
	  //printf("blockSizeIF: %d\n", sys->blockSize);

	  for(int k = 0; k < kMax ;){
	    
	    float redMean = 0., redMean2 = 0., swap = 0.;
	    float blueMean = 0., blueMean2 = 0.;
	    float errorRed = 0., errorBlue = 0.;

	    do{k++;}while(kMax%k != 0);	    
	    
	    dim3 block(sys->blockSize,1,1);

	    int fitGrid = (sys->V/2/k + sys->blockSize - 1)/sys->blockSize;
	    dim3 grid(fitGrid, sys->sysN, 1);
	    int h_loopStride = block.x*grid.x;
	    MYCUDA_ERROR( cudaMemcpyToSymbol(d_loopStride, &h_loopStride, sizeof(int)) );
	    
	    printf("Measuring Times No-Grid L:%d block.x:%d grid.x:%d spins per thread:%d\n",
		   L, block.x, grid.x, k);
	    
	    int threads = block.x*block.y*block.z*grid.x*grid.y*grid.z;
	    MYCUDA_ERROR( cudaMemcpyToSymbol(d_threads, &threads, sizeof(int)) );
	    
	    curandState *rng = NULL;
	    MYCUDA_ERROR(cudaMalloc((void **)&rng, threads*sizeof(curandState)));
	    setupXORWOW_kernel<<< grid, block>>>(rng);
	    MYCUDA_ERROR( cudaDeviceSynchronize() );

	    // KERNEL: Bitwise; Only L = 2^k are examined
	    if( ((1U<<findLog2(sys->L))^sys->L) == 0 ){
	      for(int sample = 0; sample < initSamples; sample++){
		
		MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
		
		for(int steps = 0; steps < initSteps; steps++){	  
		  MCStepReds_Bitwise_XORWOW<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
							       sys->d_reds2, sys->d_reds3,
							       rng);
							   
		  MYCUDA_ERROR( cudaDeviceSynchronize() );
		}
		
		MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
		MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
		MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
		redMean += (swap/(initSteps)); 
		redMean2 += (swap/(initSteps))*(swap/(initSteps));
		swap = 0.;
		
		MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
		
		for(int steps = 0; steps < initSteps; steps++){	  
		  MCStepBlues_Bitwise_XORWOW<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
								 sys->d_blues2, sys->d_blues3,
								 rng);
							    
		  MYCUDA_ERROR( cudaDeviceSynchronize() );
		}
		
		MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
		MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
		MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
		blueMean += (swap/(initSteps)); 
		blueMean2 += (swap/(initSteps))*(swap/(initSteps));
		swap = 0.;
	      }
	      
	      redMean /= initSamples;  redMean2 /= initSamples;      
	      blueMean /= initSamples; blueMean2 /= initSamples;
	      
	      errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
	      errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
	      
	      snprintf(fileName, sizeof(fileName), "GPU%d_bitwiseKernel_PR_smooth", sys->cardN);
	      FILE *outputFile = fopen(fileName, "a+");
	      fprintf(outputFile, "%d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
		      L, sysN, k, block.x, grid.x,
		      redMean*1.0e-3, errorRed*1.0e-3, blueMean*1.0e-3, errorBlue*1.0e-3);
	      fflush(outputFile);
	      fclose(outputFile);
	    }
	    
	    // KERNEL: Standard
	    redMean = blueMean = redMean2 = blueMean2 = 0.;
	    for(int sample = 0; sample < initSamples; sample++){
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepReds_Standard_XORWOW<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
							      sys->d_reds2, sys->d_reds3,
							      rng);
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      redMean += (swap/(initSteps)); 
	      redMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepBlues_Standard_XORWOW<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
							       sys->d_blues2, sys->d_blues3,
							       rng);
		
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      blueMean += (swap/(initSteps)); 
	      blueMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	    }
	    
	    redMean /= initSamples;  redMean2 /= initSamples;      
	    blueMean /= initSamples; blueMean2 /= initSamples;
	    
	    errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
	    errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
	    
	    snprintf(fileName, sizeof(fileName), "GPU%d_standardKernel_PR_smooth", sys->cardN);
	    FILE *outputFile = fopen(fileName, "a+");
	    fprintf(outputFile, "%d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
		    L, sysN, k, block.x, grid.x,
		    redMean*1.0e-3, errorRed*1.0e-3, blueMean*1.0e-3, errorBlue*1.0e-3);
	    fflush(outputFile);
	    fclose(outputFile);
	    
	    // KERNEL: Sliced
	    redMean = blueMean = redMean2 = blueMean2 = 0.;
	    for(int sample = 0; sample < initSamples; sample++){
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepReds_Sliced_XORWOW<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
							    sys->d_reds2, sys->d_reds3,
							    rng);
							
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      redMean += (swap/(initSteps)); 
	      redMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepBlues_Sliced_XORWOW<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
							     sys->d_blues2, sys->d_blues3,
							     rng);
		
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      blueMean += (swap/(initSteps)); 
	      blueMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	    }
	    
	    redMean /= initSamples;  redMean2 /= initSamples;      
	    blueMean /= initSamples; blueMean2 /= initSamples;
	    
	    errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
	    errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
	    
	    snprintf(fileName, sizeof(fileName), "GPU%d_slicedKernel_PR_smooth", sys->cardN);
	    outputFile = fopen(fileName, "a+");
	    fprintf(outputFile, "%d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
		    L, sysN, k, block.x, grid.x,
		    redMean*1.0e-3, errorRed*1.0e-3, blueMean*1.0e-3, errorBlue*1.0e-3);
	    fflush(outputFile);
	    fclose(outputFile);
	  	  
	    MYCUDA_ERROR(cudaFree(rng));
	  }
	}
      }

      // CLEANING UP
      //CRNGArrayFree(&rng);
      if(dMemFree(sys)){
	fprintf(stderr,"\nError in dMemFree\n");
	return 1;
      }
      if(hMemFree(sys)){
	fprintf(stderr,"\nError in hMemFree\n");
	return 1;
      }
    }

    initSysN = 0;
  }
  // allocazione
  // deallocazione
  // linearizzazione nel numero dei sistemi con limite la memoria
  return 0;
}



extern int benchmarkSmooth_MTGP32(int argc, char **argv){
  if(argc != 9){
    fprintf(stderr, "usage: %s <gpu> <L> <sysN> <block> <samples> <steps> <memory> <spinsPerThread>\n\n", argv[0]);
    fprintf(stderr, "Systems' GPUs are:\n");
    int numberOfGPUs = 0;
    MYCUDA_ERROR( cudaGetDeviceCount(&numberOfGPUs) );
    for(int gpu = 0; gpu < numberOfGPUs; gpu++){
      cudaDeviceProp prop;
      MYCUDA_ERROR( cudaGetDeviceProperties(&prop, gpu) );
      printf("GPU %d: %s\n", gpu, prop.name);
    }
    return 1;
  }

  char fileName[1024];
  //struct timespec start, stop;

  sysEA3D_t *sys = (sysEA3D_t *)malloc(sizeof(sysEA3D_t));
  int initL, initSysN, initSamples, initSteps, initBlock, spinsPerThread;
  float memory;
  sscanf(argv[1], "%d", &sys->cardN);
  sscanf(argv[2], "%d", &initL);
  sscanf(argv[3], "%d", &initSysN);
  sscanf(argv[4], "%d", &initBlock);
  sscanf(argv[5], "%d", &initSamples);
  sscanf(argv[6], "%d", &initSteps);
  sscanf(argv[7], "%f", &memory);
  sscanf(argv[8], "%d", &spinsPerThread);

  if(initL == 0) initL = 8;
  if(initSamples == 0) initSamples = 100;
  if(initSteps == 0) initSteps = 50;

  sys->beta = 1.101;

  int numberOfGPUs = 0;
  cudaGetDeviceCount(&numberOfGPUs);
  if(sys->cardN >= numberOfGPUs){
    fprintf(stderr, "gpu %d >= gpuN %d\n", sys->cardN, numberOfGPUs);
    return 1;
  }

  MYCUDA_ERROR( cudaSetDevice(sys->cardN) );
  cudaDeviceProp prop;
  MYCUDA_ERROR( cudaGetDeviceProperties(&prop, sys->cardN) );
  printf("GPU %d: %s Compute Capability: %d.%d\n", sys->cardN, prop.name, prop.major, prop.minor);

  cudaEvent_t startEvent, stopEvent;
  MYCUDA_ERROR( cudaEventCreate(&startEvent) );
  MYCUDA_ERROR( cudaEventCreate(&stopEvent) );

  int maxThreadsPerBlock = 0;
  if(prop.major == 2) maxThreadsPerBlock = 512;
  if(prop.major == 3) maxThreadsPerBlock = 1024;

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Bitwise_MTGP32, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Bitwise_MTGP32, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Standard_MTGP32, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Standard_MTGP32, cudaFuncCachePreferL1) );

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Sliced_MTGP32, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Sliced_MTGP32, cudaFuncCachePreferL1) );


  for(int L = initL; L <= 256; L += 2){

    int hV = L*L*L/2;
    size_t NJ_Size = 6*hV*sizeof(MSC), NS_Size = 4*2*hV*sizeof(MSC), RNG_Size = (hV/32)*sizeof(curandStateMtgp32);
    size_t devFree = 0, devTotal = 0;
    MYCUDA_ERROR( cudaMemGetInfo(&devFree, &devTotal) );
    // here to decide the percentage of device memory to use. 
    int MaxSysN = (int)(memory*(devFree/(NJ_Size + NS_Size + RNG_Size)));
    printf("\n\nL = %d, Free memory on device: %lu, MaxSysN: %d\n", L, devFree, MaxSysN);

    if(initSysN == 0) initSysN = 1;

    for(int sysN = initSysN; sysN <= MaxSysN; sysN *= 2){
      sys->L = L;
      sys->V = sys->L*sys->L*sys->L;
      sys->sysN = sysN;
      sys->NJ = sys->sysN*sys->V/2;
      sys->NS = sys->sysN*sys->V/2;
      
      if(masksInit(sys)){
	fprintf(stderr,"\nError in masksInit\n");
	return 1;
      }
      if(MCWeightsInit(sys)){
	fprintf(stderr,"\nError in MCWeightsInit\n");
	return 1;
      }
      if(dMemInit(sys)){
	fprintf(stderr,"\nError in dMemInit\n");
	return 1;
      }
      if(hMemInit2(sys)){
	fprintf(stderr,"\nError in hMemInit\n");
	return 1;
      }
      if(H2DSysCopy(sys)){
	fprintf(stderr,"\nError in H2DSysCopy\n");
	return 1;
      }

      MYCUDA_ERROR( cudaMemGetInfo(&devFree, &devTotal) );
      printf("\n\n---\tL = %d, sysN = %d, free memory after allocation: %lu\t---\n\n",
	     L, sysN, devFree);

      ////////////////////////////////////////
      // NO-GRID KERNELS
      ///////////////////////////////////////

      int blockInit = 0;

      if(initBlock == 0){
	blockInit = maxThreadsPerBlock;
	while(blockInit > sys->V/2) blockInit /= 2;
      }


      for(sys->blockSize = blockInit; sys->blockSize >= 32; sys->blockSize /= 2){

	//printf("blockSize: %d\n", sys->blockSize);

	if(sys->blockSize >= 32){
	  int kMax = sys->V/sys->blockSize/2;
	  if(kMax > spinsPerThread && spinsPerThread != 0) kMax = spinsPerThread;
	  //printf("blockSizeIF: %d\n", sys->blockSize);

	  for(int k = 0; k < kMax ;){
	    
	    float redMean = 0., redMean2 = 0., swap = 0.;
	    float blueMean = 0., blueMean2 = 0.;
	    float errorRed = 0., errorBlue = 0.;

	    do{k++;}while(kMax%k != 0);	    
	    
	    dim3 block(sys->blockSize,1,1);

	    int fitGrid = (sys->V/2/k + sys->blockSize - 1)/sys->blockSize;
	    dim3 grid(fitGrid, sys->sysN, 1);
	    int h_loopStride = block.x*grid.x;
	    MYCUDA_ERROR( cudaMemcpyToSymbol(d_loopStride, &h_loopStride, sizeof(int)) );
	    
	    printf("Measuring Times No-Grid L:%d block.x:%d grid.x:%d spins per thread:%d\n",
		   L, block.x, grid.x, k);
	    
	    int threads = block.x*block.y*block.z*grid.x*grid.y*grid.z;
	    MYCUDA_ERROR( cudaMemcpyToSymbol(d_threads, &threads, sizeof(int)) );
	    
	    if(grid.x*grid.y*grid.z <= 200){

	    int spins = 4*k;
	    curandStateMtgp32 *rng = NULL;
	    mtgp32_kernel_params *devKernelParams = NULL;
	    MYCUDA_ERROR( cudaMalloc((void **)&rng, grid.x*grid.y*grid.z*sizeof(curandStateMtgp32)) );
	    MYCUDA_ERROR( cudaMalloc((void**)&devKernelParams, sizeof(mtgp32_kernel_params)) );
	    MYCURAND_ERROR( curandMakeMTGP32Constants(mtgp32dc_params_fast_11213, devKernelParams) );
	    MYCURAND_ERROR( curandMakeMTGP32KernelState(rng, mtgp32dc_params_fast_11213, 
							devKernelParams, grid.x*grid.y*grid.z, 1234) );
	    

	    // KERNEL: Bitwise; Only L = 2^k are examined
	    if( ((1U<<findLog2(sys->L))^sys->L) == 0 ){
	      for(int sample = 0; sample < initSamples; sample++){
		
		MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
		
		for(int steps = 0; steps < initSteps; steps++){	  
		  MCStepReds_Bitwise_MTGP32<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
							       sys->d_reds2, sys->d_reds3,
							       rng, spins);
		  
		  MYCUDA_ERROR( cudaDeviceSynchronize() );
		}
		
		MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
		MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
		MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
		redMean += (swap/(initSteps)); 
		redMean2 += (swap/(initSteps))*(swap/(initSteps));
		swap = 0.;
		
		MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
		
		for(int steps = 0; steps < initSteps; steps++){	  
		  MCStepBlues_Bitwise_MTGP32<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
								sys->d_blues2, sys->d_blues3,
								rng, spins);
							    
		  MYCUDA_ERROR( cudaDeviceSynchronize() );
		}
		
		MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
		MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
		MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
		blueMean += (swap/(initSteps)); 
		blueMean2 += (swap/(initSteps))*(swap/(initSteps));
		swap = 0.;
	      }
	      
	      redMean /= initSamples;  redMean2 /= initSamples;      
	      blueMean /= initSamples; blueMean2 /= initSamples;
	      
	      errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
	      errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
	      
	      snprintf(fileName, sizeof(fileName), "GPU%d_bitwiseKernel_PR_smooth", sys->cardN);
	      FILE *outputFile = fopen(fileName, "a+");
	      fprintf(outputFile, "%d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
		      L, sysN, k, block.x, grid.x,
		      redMean*1.0e-3, errorRed*1.0e-3, blueMean*1.0e-3, errorBlue*1.0e-3);
	      fflush(outputFile);
	      fclose(outputFile);
	    }
	    
	    // KERNEL: Standard
	    redMean = blueMean = redMean2 = blueMean2 = 0.;
	    for(int sample = 0; sample < initSamples; sample++){
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepReds_Standard_MTGP32<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
							      sys->d_reds2, sys->d_reds3,
							      rng, spins);
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      redMean += (swap/(initSteps)); 
	      redMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepBlues_Standard_MTGP32<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
							       sys->d_blues2, sys->d_blues3,
							       rng, spins);
		
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      blueMean += (swap/(initSteps)); 
	      blueMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	    }
	    
	    redMean /= initSamples;  redMean2 /= initSamples;      
	    blueMean /= initSamples; blueMean2 /= initSamples;
	    
	    errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
	    errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
	    
	    snprintf(fileName, sizeof(fileName), "GPU%d_standardKernel_PR_smooth", sys->cardN);
	    FILE *outputFile = fopen(fileName, "a+");
	    fprintf(outputFile, "%d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
		    L, sysN, k, block.x, grid.x,
		    redMean*1.0e-3, errorRed*1.0e-3, blueMean*1.0e-3, errorBlue*1.0e-3);
	    fflush(outputFile);
	    fclose(outputFile);
	    
	    // KERNEL: Sliced
	    redMean = blueMean = redMean2 = blueMean2 = 0.;
	    for(int sample = 0; sample < initSamples; sample++){
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepReds_Sliced_MTGP32<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
							    sys->d_reds2, sys->d_reds3,
							    rng, spins);
		
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      redMean += (swap/(initSteps)); 
	      redMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	      
	      MYCUDA_ERROR( cudaEventRecord(startEvent, 0) );
	      
	      for(int steps = 0; steps < initSteps; steps++){	  
		MCStepBlues_Sliced_MTGP32<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
							     sys->d_blues2, sys->d_blues3,
							     rng, spins);
		
		MYCUDA_ERROR( cudaDeviceSynchronize() );
	      }
	      
	      MYCUDA_ERROR( cudaEventRecord(stopEvent, 0) );
	      MYCUDA_ERROR( cudaEventSynchronize(stopEvent) );
	      MYCUDA_ERROR( cudaEventElapsedTime(&swap, startEvent, stopEvent) );
	      blueMean += (swap/(initSteps)); 
	      blueMean2 += (swap/(initSteps))*(swap/(initSteps));
	      swap = 0.;
	    }
	    
	    redMean /= initSamples;  redMean2 /= initSamples;      
	    blueMean /= initSamples; blueMean2 /= initSamples;
	    
	    errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
	    errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
	    
	    snprintf(fileName, sizeof(fileName), "GPU%d_slicedKernel_PR_smooth", sys->cardN);
	    outputFile = fopen(fileName, "a+");
	    fprintf(outputFile, "%d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
		    L, sysN, k, block.x, grid.x,
		    redMean*1.0e-3, errorRed*1.0e-3, blueMean*1.0e-3, errorBlue*1.0e-3);
	    fflush(outputFile);
	    fclose(outputFile);
	  	  
	    MYCUDA_ERROR(cudaFree(rng));
	    MYCUDA_ERROR(cudaFree(devKernelParams));
	    }
	  }
	}
      }

      // CLEANING UP
      //CRNGArrayFree(&rng);
      if(dMemFree(sys)){
	fprintf(stderr,"\nError in dMemFree\n");
	return 1;
      }
      if(hMemFree(sys)){
	fprintf(stderr,"\nError in hMemFree\n");
	return 1;
      }
    }

    initSysN = 0;
  }
  // allocazione
  // deallocazione
  // linearizzazione nel numero dei sistemi con limite la memoria
  return 0;
}

extern int benchmarkSliced_MTGP(int argc, char **argv){
  if(argc != 8){
    fprintf(stderr, "usage: %s <gpu> <L> <sysN> <spinsPerThread> <block> <samples> <steps>\n\n", argv[0]);
    fprintf(stderr, "Systems' GPUs are:\n");
    int numberOfGPUs = 0;
    MYCUDA_ERROR( cudaGetDeviceCount(&numberOfGPUs) );
    for(int gpu = 0; gpu < numberOfGPUs; gpu++){
      cudaDeviceProp prop;
      MYCUDA_ERROR( cudaGetDeviceProperties(&prop, gpu) );
      printf("GPU %d: %s\n", gpu, prop.name);
    }
    return 1;
  }

  char fileName[1024];
  struct timespec start, stop;
  double redMean = 0., redMean2 = 0., swap = 0.;
  double blueMean = 0., blueMean2 = 0.;
  double errorRed = 0., errorBlue = 0.;


  sysEA3D_t *sys = (sysEA3D_t *)malloc(sizeof(sysEA3D_t));
  int initL, initSysN, initSamples, initSteps, initBlock, spinsPerThread;

  sscanf(argv[1], "%d", &sys->cardN);
  sscanf(argv[2], "%d", &initL);
  sscanf(argv[3], "%d", &initSysN);
  sscanf(argv[4], "%d", &spinsPerThread);
  sscanf(argv[5], "%d", &initBlock);
  sscanf(argv[6], "%d", &initSamples);
  sscanf(argv[7], "%d", &initSteps);

  sys->beta = 1.101;

  int numberOfGPUs = 0;
  cudaGetDeviceCount(&numberOfGPUs);
  if(sys->cardN >= numberOfGPUs){
    fprintf(stderr, "gpu %d >= gpuN %d\n", sys->cardN, numberOfGPUs);
    return 1;
  }

  MYCUDA_ERROR( cudaSetDevice(sys->cardN) );
  cudaDeviceProp prop;
  MYCUDA_ERROR( cudaGetDeviceProperties(&prop, sys->cardN) );
  printf("GPU %d: %s Compute Capability: %d.%d\n", sys->cardN, prop.name, prop.major, prop.minor);

  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepReds_Sliced_MT19937, cudaFuncCachePreferL1) );
  MYCUDA_ERROR( cudaFuncSetCacheConfig(MCStepBlues_Sliced_MT19937, cudaFuncCachePreferL1) );

  int L = initL;

  int sysN = initSysN;

  sys->L = L;
  sys->V = sys->L*sys->L*sys->L;
  sys->sysN = sysN;
  sys->NJ = sys->sysN*sys->V/2;
  sys->NS = sys->sysN*sys->V/2;
      
  if(masksInit(sys)){
    fprintf(stderr,"\nError in masksInit\n");
    return 1;
  }
  if(MCWeightsInit(sys)){
    fprintf(stderr,"\nError in MCWeightsInit\n");
    return 1;
  }
  if(dMemInit(sys)){
    fprintf(stderr,"\nError in dMemInit\n");
    return 1;
  }
  if(hMemInit2(sys)){
    fprintf(stderr,"\nError in hMemInit\n");
    return 1;
  }
  if(H2DSysCopy(sys)){
    fprintf(stderr,"\nError in H2DSysCopy\n");
    return 1;
  }

  ////////////////////////////////////////
  // NO-GRID KERNELS
  ///////////////////////////////////////
  
  int blockInit = initBlock;
  sys->blockSize = blockInit;

  int k = spinsPerThread;
	    	    
  dim3 block(sys->blockSize,1,1);

  int fitGrid = (sys->V/2/k + sys->blockSize - 1)/sys->blockSize;
  dim3 grid(fitGrid, sys->sysN, 1);
  int h_loopStride = block.x*grid.x;
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_loopStride, &h_loopStride, sizeof(int)) );
	    	    
  int threads = block.x*block.y*block.z*grid.x*grid.y*grid.z;
  MYCUDA_ERROR( cudaMemcpyToSymbol(d_threads, &threads, sizeof(int)) );
	    
  // KERNEL: Sliced

  curandStateMtgp32 *devMTGPStates;
  mtgp32_kernel_params *devKernelParams;

  /* Allocate space for prng states on device */
  MYCUDA_ERROR( cudaMalloc((void **)&devMTGPStates, grid.x*grid.y*sizeof(curandStateMtgp32)) );
    
  /* Setup MTGP prng states */
    
  /* Allocate space for MTGP kernel parameters */
  MYCUDA_ERROR( cudaMalloc((void**)&devKernelParams, sizeof(mtgp32_kernel_params)) );
    
  /* Reformat from predefined parameter sets to kernel format, */
  /* and copy kernel parameters to device memory               */
  MYCURAND_ERROR( curandMakeMTGP32Constants(mtgp32dc_params_fast_11213, devKernelParams) );
    
  /* Initialize one state per thread block */
  MYCURAND_ERROR( curandMakeMTGP32KernelState(devMTGPStates, mtgp32dc_params_fast_11213, 
					      devKernelParams, grid.x*grid.y, 1234) );


  for(int sample = 0; sample < initSamples; sample++){
  
    clock_gettime(CLOCK_MONOTONIC, &start);
  
    for(int steps = 0; steps < initSteps; steps++){	  
      /*MCStepReds_Sliced_MTGP32<<< grid, block >>>(sys->d_reds0, sys->d_reds1,
						  sys->d_reds2, sys->d_reds3,
						  devMTGPStates); */

      MYCUDA_ERROR( cudaDeviceSynchronize() );
    }

    clock_gettime(CLOCK_MONOTONIC, &stop);

    swap = ((stop.tv_sec - start.tv_sec) + (stop.tv_nsec - start.tv_nsec)/1.0e9);
    redMean += (swap/(initSteps)); 
    redMean2 += (swap/(initSteps))*(swap/(initSteps));
    swap = 0.;

    clock_gettime(CLOCK_MONOTONIC, &start);
	      	      
    for(int steps = 0; steps < initSteps; steps++){	  
      /*      MCStepBlues_Sliced_MTGP32<<< grid, block >>>(sys->d_blues0, sys->d_blues1,
						   sys->d_blues2, sys->d_blues3,
						   devMTGPStates);*/
      
      MYCUDA_ERROR( cudaDeviceSynchronize() );
    }

    clock_gettime(CLOCK_MONOTONIC, &stop);

    swap = ((stop.tv_sec - start.tv_sec) + (stop.tv_nsec - start.tv_nsec)/1.0e9); 	    
    blueMean += (swap/(initSteps)); 
    blueMean2 += (swap/(initSteps))*(swap/(initSteps));
    swap = 0.;
	      
  }
	    
  redMean /= initSamples;  redMean2 /= initSamples;      
  blueMean /= initSamples; blueMean2 /= initSamples;
  
  errorRed = sqrt((redMean2 - redMean*redMean)/(initSteps - 1.));
  errorBlue = sqrt((blueMean2 - blueMean*blueMean)/(initSteps - 1.));
  
  snprintf(fileName, sizeof(fileName), "GPU%d_slicedKernel_PR_smooth", sys->cardN);

  FILE *outputFile = fopen(fileName, "a+");
  fprintf(outputFile, "%d %d %d %d %d %21.15e %21.15e %21.15e %21.15e\n", 
	  L, sysN, k, block.x, grid.x,
	  redMean, errorRed, blueMean, errorBlue);
  fflush(outputFile);
  fclose(outputFile);

  if(dMemFree(sys)){
    fprintf(stderr,"\nError in dMemFree\n");
    return 1;
  }
  if(hMemFree(sys)){
    fprintf(stderr,"\nError in hMemFree\n");
    return 1;
  }
  
  return 0;
}


